package org.cn3.morpho.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.NeuronProcess;
import org.cn3.morpho.domain.NodePair;
import org.cn3.morpho.domain.Soma;
import org.cn3.morpho.domain.SwcDataNode;
import org.cn3.morpho.domain.SwcSecondaryData;
import org.cn3.morpho.domain.SwcTreeNode;
import org.cn3.morpho.domain.TreeRoot;
import org.cn3.morpho.resources.MorphoConstants;
import org.cn3.morpho.utils.SwcDataUtils;


/**
 * 
 * @author gillette
 * Imports .swc files, generating tree structures made of TreeNode objects and/or their subclasses
 * Conversion based on method of division (full cell, arborizations, trees).
 * 
 */
public class SWCReader {
	
	//private static final int TREE_TYPE_CELL = 1;
	public static final int TREE_TYPE_AXON = SwcDataNode.TYPE_AXON;
	public static final int TREE_TYPE_BASAL_DENDRITE = SwcDataNode.TYPE_DENDRITE;
	public static final int TREE_TYPE_APICAL_DENDRITE = SwcDataNode.TYPE_APICAL;
	
	/*
	private static final int TREE_TYPE_GENERAL_DENDRITE = 7;
	private static final int TREE_TYPE_BDTREE = 5;
	private static final int TREE_TYPE_ADTREE = 6;
	*/

	// For swc->TreeNodes
	public static final int INDEX_AXON = 0;
	public static final int INDEX_APICAL = 1;
	public static final int INDEX_DENDRITE = 2;
	
	private static final int DEFAULT_MIN_SIZE = 0;
	private static final boolean DEFAULT_WRITE_ARCHIVES = false;
	private static final boolean DEFAULT_WRITE_CELLTYPES = false;
	
	public static final Pattern swcDataPattern = Pattern.compile("[ \\t]*(\\d+)[ \\t]+(\\d+)[ \\t]+(\\-?[\\.\\d]+)[ \\t]+(\\-?[\\.\\d]+)[ \\t]+(\\-?[\\.\\d]+)[ \\t]+([\\.\\d]+)[ \\t]+(\\-?\\d+)[ \\t]*");
	public static final int SWC_PATTERN_NODE_ID = 1;
	public static final int SWC_PATTERN_TYPE = 2;
	public static final int SWC_PATTERN_X = 3;
	public static final int SWC_PATTERN_Y = 4;
	public static final int SWC_PATTERN_Z = 5;
	public static final int SWC_PATTERN_RADIUS = 6;
	public static final int SWC_PATTERN_PARENT = 7;
	private static boolean debug = false;
	
	private String swcDir;
	private String dataset;
	private String classification;
	private int minSize;
	private boolean writeArchives;
	private boolean writeCellTypes;
	
	public boolean isWriteArchives() {
		return writeArchives;
	}

	public void setWriteArchives(boolean writeArchives) {
		this.writeArchives = writeArchives;
	}

	public boolean isWriteCellTypes() {
		return writeCellTypes;
	}

	public void setWriteCellTypes(boolean writeCellTypes) {
		this.writeCellTypes = writeCellTypes;
	}

	public int getMinSize() {
		return minSize;
	}

	public void setMinSize(int minSize) {
		this.minSize = minSize;
	}

	public String getClassification() {
		return classification;
	}

	public void setClassification(String classification) {
		this.classification = classification;
	}

	public String getDataset() {
		return dataset;
	}

	public void setDataset(String dataset) {
		this.dataset = dataset;
	}

	public String getSwcDir() {
		return swcDir;
	}

	public void setSwcDir(String swcDir) {
		this.swcDir = swcDir;
	}

	public SWCReader(String swcDir){
		setDefaults();
		this.swcDir = swcDir+"/"+dataset;
	}
	
	public void setDefaults(){
		dataset = MorphoConstants.DEFAULT_DATASET;
		minSize = DEFAULT_MIN_SIZE;
		writeArchives = DEFAULT_WRITE_ARCHIVES;
		writeCellTypes = DEFAULT_WRITE_CELLTYPES;
		debug = false;
	}
	
	/**
	 * 
	 * @param inputFile
	 * @param classification
	 * @param minSize
	 * @param calcPathDist
	 * @return List<SwcTreeNode> containing SwcTreeNode: axon, List of SwcTreeNode: apical dendrites, List of SwcTreeNode: basal dendrites
	 * @NOTE: returned trees are now still connected to parent branches if the parent is a soma node
	 * @throws Exception
	 */
	public static List<SwcTreeNode> convertSwcToTreeNodes(
			String inputFile) throws Exception{
		return convertSwcToTreeNodes(new File(inputFile));
	}
	public static List<SwcTreeNode> convertSwcToTreeNodes(
			File inputFile) throws IOException{

		BufferedReader reader 
    		= new BufferedReader(new FileReader(inputFile));

		String dataString = "";
		Map<Integer,SwcTreeNode> nodeMap = new HashMap<Integer,SwcTreeNode>();
		Map<SwcTreeNode,Integer> addNodeSideMap = new HashMap<SwcTreeNode,Integer>();
		SwcTreeNode node, parent, axon = null, somaRoot = null;
		Matcher m;
		int type, nodeId, parentId;
		
		List<SwcTreeNode> apicalDendriteTrees = new ArrayList<SwcTreeNode>();
		List<SwcTreeNode> basalDendriteTrees = new ArrayList<SwcTreeNode>();
		
		// 1. Assembles axon, basal trees, and apical trees (non-binary).
		while ( (dataString=reader.readLine()) != null ) { 
	        m = swcDataPattern.matcher(dataString);
	        //System.out.println(dataString);
	        
	        if (m.matches()){
	        	node = new SwcTreeNode();
	        	nodeId = Integer.parseInt(m.group(1));
	        	node.setNodeId(nodeId);
	        	type = Integer.parseInt(m.group(2));
	        	node.setType(type);
	        	node.setX(Double.parseDouble(m.group(3)));
	        	node.setY(Double.parseDouble(m.group(4)));
	        	node.setZ(Double.parseDouble(m.group(5)));
	        	node.setRadius(Double.parseDouble(m.group(6)));
	        	parentId = Integer.parseInt(m.group(7));
	        	node.setParentId(parentId);

	        	// Types: 1=soma, 2=axon, 3=basal dendrite, 4=apical dendrite
	        	if (type == 1 || type == 2 || type == 3 || type == 4){
	        		// Get parent from nodeMap
		        	parent = null;
		        	if (nodeMap.containsKey(parentId)){
		        		parent = nodeMap.get(parentId);
		        	}

	        		switch (type){
	        		case SwcDataNode.TYPE_SOMA:
	        			if (somaRoot == null){
	        				somaRoot = node;
	        				//parent = null;
	        			}
	        			addNode(node,parent,nodeMap,addNodeSideMap);
	        			break;
	        		case SwcDataNode.TYPE_AXON:
	        			// First data for axon, create new tree
	        			if (axon == null){
	        				axon = node;
	        				if (parent != null && parent.getType() != SwcDataNode.TYPE_AXON && parent.getType() != SwcDataNode.TYPE_SOMA){
	        					parent = null;
	        				}
	        			}
	        			addNode(node,parent,nodeMap,addNodeSideMap);
	        			break;
	        		case SwcDataNode.TYPE_DENDRITE:
	        			// if true then this is a new tree (used to only do this when classification is "tree")
        				if (parent == null || parent.getType() != TREE_TYPE_BASAL_DENDRITE){
        					basalDendriteTrees.add(node);
	        				if (parent != null && parent.getType() != SwcDataNode.TYPE_SOMA){
	        					parent = null;
	        				}
        				}
	        			addNode(node,parent,nodeMap,addNodeSideMap);
	        			break;
	        		case SwcDataNode.TYPE_APICAL:
	        			// if true then this is a new tree
        				if (parent == null || parent.getType() != TREE_TYPE_APICAL_DENDRITE){
        					apicalDendriteTrees.add(node);
        					if (parent != null && parent.getType() != SwcDataNode.TYPE_SOMA){
        						parent = null;
        					}
        				}
	        			addNode(node,parent,nodeMap,addNodeSideMap);
	        			break;
	        		default:
	        			node = null;
	        		}

	        	}
	        }
	        m = null;
	    }

		List<SwcTreeNode> trees = new ArrayList<SwcTreeNode>();
		if (somaRoot != null) trees.add(somaRoot);
		if (axon != null) trees.add(axon);
		trees.addAll(apicalDendriteTrees);
		trees.addAll(basalDendriteTrees);
		if (debug) System.out.println("Done assembling trees from swc");
		return trees;
	}
	
	/**
	 * @param inputFile
	 * @param classification
	 * @param traversal
	 * @return
	 * @throws Exception
	 * 
	 * Functionality:
	 *    1. Assembles axon, basal trees, and apical trees (non-binary).
	 *    2. Removes continuations.
	 *    3. Transforms each into binary tree since all tree info is now known (provides stability in transformation).
	 *    4. Changes data to structure data (A - termination, C - bifurcation) 
	 *    5. Builds basal dendrite arborization and apical dendrite arborization
	 */
	/*
	public static Soma<SwcTreeNode> convertSwcToSoma(File inputFile) throws Exception{
		return convertSwcToSoma(inputFile, NeuronTreeClassification.SUB_ARBORIZATION);
	}
	public static Soma<SwcTreeNode> convertSwcToSoma(
			File inputFile, boolean calcPathDist) throws Exception{
		return convertSwcToSoma(inputFile, NeuronTreeClassification.SUB_ARBORIZATION);
	}*/

	public static Soma<BinaryTreeNode> convertSwcToBinarySoma(File inputFile) throws Exception{
		return convertSwcToBinarySoma(inputFile, NeuronTreeClassification.SUB_ARBORIZATION, 0, false);
	}
	public static Soma<BinaryTreeNode> convertSwcToBinarySoma(
			File inputFile, boolean calcPathDist) throws Exception{
		return convertSwcToBinarySoma(inputFile, NeuronTreeClassification.SUB_ARBORIZATION, 0, calcPathDist);
	}
	public static Soma<BinaryTreeNode> convertSwcToBinarySoma(
			File inputFile, NeuronTreeClassification classification) throws Exception{
		return convertSwcToBinarySoma(inputFile, classification, 0, false);
	}

	public static Soma<BinaryTreeNode> convertSwcToBinarySoma(
			File inputFile, NeuronTreeClassification classification, boolean calcPathDist) throws Exception{
		return convertSwcToBinarySoma(inputFile, classification, 0, calcPathDist);
	}

	public static Soma<BinaryTreeNode> convertSwcToBinarySoma(
			File inputFile, NeuronTreeClassification classification, int minSize) throws Exception{
		return convertSwcToBinarySoma(inputFile, classification, minSize, false);
	}

	public static Soma<BinaryTreeNode> convertSwcToBinarySoma(
			File inputFile, NeuronTreeClassification classification, int minSize, boolean calcPathDist) throws Exception{
		
		Soma<SwcTreeNode> swcSoma = convertSwcToSoma(inputFile);
		Soma<BinaryTreeNode> binarySoma = new Soma<BinaryTreeNode>();
		binarySoma.setNumDendriteStems(swcSoma.getNumDendriteStems());
		
		// 2. Removes continuations.
		// 3. Builds basal dendrite arborization and apical dendrite arborization
		// 4. Transforms each into binary tree since all tree info is now known (provides stability in transformation).
		// 5. Changes data to structure data (A - termination, C - bifurcation)
		SwcTreeNode treeHead;
		List<SwcTreeNode> apicalTrees = swcSoma.getApicalDendriteTrees(), basalTrees = swcSoma.getBasalDendriteTrees(), 
				dendriteTrees = new ArrayList<SwcTreeNode>(), neuronTrees = new ArrayList<SwcTreeNode>();
		BinaryTreeNode neurite, binaryTree;
		
		if (swcSoma.getAxonalTree() != null){
			SwcTreeNode axon = swcSoma.getAxonalTree().getTreeHead();
			switch(classification){
			case NEURON:
				neuronTrees.add(axon);
				break;
			default:
				axon = removeContinuations(axon, calcPathDist);
				SwcTreeNode.updateStatisticsDown(axon);
				binaryTree = convertTreeToBinary(axon);
				if (debug) System.out.println("Axon size: "+binaryTree.getTreeSize());
				if (binaryTree.getTreeSize() >= minSize){
					binarySoma.setAxonalTree(new NeuronProcess<BinaryTreeNode>());
					binarySoma.getAxonalTree().setTreeHead(binaryTree);
				}
			}
		}
		if (apicalTrees != null && apicalTrees.size() > 0){
			switch(classification){
			case TREE:
				binarySoma.setApicalDendriteTrees(new ArrayList<BinaryTreeNode>());
				for (int i = 0; i < apicalTrees.size(); i++){
					treeHead = apicalTrees.get(i);
					treeHead = removeContinuations(treeHead, calcPathDist);
					SwcTreeNode.updateStatisticsDown(treeHead);
					binaryTree = convertTreeToBinary(treeHead);
					if (binaryTree.getTreeSize() >= minSize){
						System.out.println("Apical tree does not have enough branches; has "+binaryTree.getTreeSize());
						binarySoma.getApicalDendriteTrees().add(binaryTree);
					}
				}
				break;
			case SUB_ARBORIZATION:
				neurite = assembleProcesses(apicalTrees, calcPathDist);
				if (neurite.getTreeSize() >= minSize){
					binarySoma.setApicalDendriteTree(new NeuronProcess<BinaryTreeNode>());
					binarySoma.getApicalDendriteTree().setTreeHead(neurite);
				}
				break;
			case ARBORIZATION:
				dendriteTrees.addAll(apicalTrees);
				break;
			case NEURON:
				neuronTrees.addAll(apicalTrees);
				break;
			}
		}
		if (basalTrees != null && basalTrees.size() > 0){
			switch(classification){
			case TREE:
				binarySoma.setBasalDendriteTrees(new ArrayList<BinaryTreeNode>());
				for (int i = 0; i < basalTrees.size(); i++){
					treeHead = basalTrees.get(i);
					treeHead = removeContinuations(treeHead, calcPathDist);
					SwcTreeNode.updateStatisticsDown(treeHead);
					binaryTree = convertTreeToBinary(treeHead);
					if (binaryTree.getTreeSize() >= minSize){
						//System.out.println("basal tree not large enough");
						binarySoma.getBasalDendriteTrees().add(binaryTree);
					}
				}
				break;
			case SUB_ARBORIZATION:
				neurite = assembleProcesses(basalTrees, calcPathDist);
				if (neurite.getTreeSize() >= minSize){
					binarySoma.setBasalDendriteTree(new NeuronProcess<BinaryTreeNode>());
					binarySoma.getBasalDendriteTree().setTreeHead(neurite);
				}
				break;
			case ARBORIZATION:
				dendriteTrees.addAll(basalTrees);
				break;
			case NEURON:
				neuronTrees.addAll(basalTrees);
				break;
			}
		}
		
		switch(classification){
		case ARBORIZATION:
			neurite = assembleProcesses(dendriteTrees, calcPathDist);
			if (neurite.getTreeSize() >= minSize){
				binarySoma.setBasalDendriteTree(new NeuronProcess<BinaryTreeNode>());
				binarySoma.getBasalDendriteTree().setTreeHead(neurite);
			}
			break;
		case NEURON:
			neurite = assembleProcesses(neuronTrees, calcPathDist);
			if (neurite.getTreeSize() >= minSize){
				binarySoma.setSomaRoot(new NeuronProcess<BinaryTreeNode>());
				binarySoma.getSomaRoot().setTreeHead(neurite);
			}
			break;
		default:
			break;
		}
		
		return binarySoma;
	}


	public static Soma<SwcTreeNode> convertSwcToSoma(
			File inputFile) throws Exception{
	    
		List<SwcTreeNode> trees = convertSwcToTreeNodes(inputFile);
		SwcTreeNode axon = null, somaRoot = null;
		List<SwcTreeNode> apicalTrees = new ArrayList<SwcTreeNode>(), 
			basalTrees = new ArrayList<SwcTreeNode>();
		for (int i = 0; i < trees.size(); i++){
			if (trees.get(i) != null){
				switch(trees.get(i).getType()){
				case SwcDataNode.TYPE_SOMA:
					somaRoot = trees.get(i);
					break;
				case SwcDataNode.TYPE_AXON:
					axon = trees.get(i);
					break;
				case SwcDataNode.TYPE_APICAL:
					apicalTrees.add(trees.get(i));
					break;
				case SwcDataNode.TYPE_DENDRITE:
					basalTrees.add(trees.get(i));
					break;
				default:
					throw new Exception("Unexpected arbor type index: "+trees.get(i).getType());
				}
			}
		}
		
		Soma<SwcTreeNode> soma = new Soma<SwcTreeNode>();
		
		// Leave as SwcTreeNode, but load into Soma structure
		if (somaRoot != null){
			soma.setSomaRoot(new NeuronProcess<SwcTreeNode>(somaRoot));
		}
		if (axon != null){
			SwcTreeNode.updateStatisticsDown(axon);
			soma.setAxonalTree(new NeuronProcess<SwcTreeNode>(axon));
		}
		if (apicalTrees.size() > 0){
			soma.setApicalDendriteTrees(apicalTrees);
		}
		if (debug){
			if (soma.getBasalDendriteTrees() != null){
				System.out.println("BasalDendriteTrees: "+soma.getBasalDendriteTrees().size());
			}
			else{
				System.out.println("BasalDendriteTrees: "+null);
			}
		}
		if (basalTrees.size() > 0){
			soma.setBasalDendriteTrees(basalTrees);
			soma.setNumDendriteStems(basalTrees.size());
		}
		return soma;
	}

	/**
	 * 
	 * @param inputFile
	 * @param classification
	 * @param minSize
	 * @param calcPathDist
	 * @return
	 * List of ParentedBinaryTree. 
	 * Assumes any number of binary trees branching from a single initial point, 
	 * regardless of type.
	 * @throws Exception
	 */
	public static List<BinaryTreeNode> convertSwcToBinaryTreeList(
			File inputFile) throws Exception{
		return convertSwcToBinaryTreeList(inputFile,true);
	}
	public static List<BinaryTreeNode> convertSwcToBinaryTreeList(
			File inputFile, boolean calcPathDist) throws Exception{
	    
		List<BinaryTreeNode> binaryTrees = new ArrayList<BinaryTreeNode>();
		
		// 1. Builds individual trees
		// 2. Removes continuations.
		// 3. Transforms each into binary tree since all tree info is now known (provides stability in transformation).

		BufferedReader reader = new BufferedReader(new FileReader(inputFile));

		String dataString = "";
		Map<Integer,SwcTreeNode> nodeMap = new HashMap<Integer,SwcTreeNode>();
		Map<SwcTreeNode,Integer> addNodeSideMap = new HashMap<SwcTreeNode,Integer>();
		SwcTreeNode node, parent, root = null;
		Matcher m;
		int type, nodeId, parentId;
		
		// 1. Assembles individual trees
		while ( (dataString=reader.readLine()) != null ) { 
	        m = swcDataPattern.matcher(dataString);
	        //System.out.println(dataString);
	        
	        if (m.matches()){
	        	node = new SwcTreeNode();
	        	nodeId = Integer.parseInt(m.group(1));
	        	node.setNodeId(nodeId);
	        	type = Integer.parseInt(m.group(2));
	        	node.setType(type);
	        	// Node types will all be set to 1
	        	//node.setType(1);
	        	node.setX(Double.parseDouble(m.group(3)));
	        	node.setY(Double.parseDouble(m.group(4)));
	        	node.setZ(Double.parseDouble(m.group(5)));
	        	node.setRadius(Double.parseDouble(m.group(6)));
	        	parentId = Integer.parseInt(m.group(7));
	        	node.setParentId(parentId);
	        	//System.out.println(nodeId + " " + type + " " + parentId);
	        	if (parentId == -1){
	        		root = node;
	        	}
	
        		// Get parent from nodeMap
	        	parent = null;
	        	if (nodeMap.containsKey(parentId)){
	        		//parent = (ParentedBinaryTreeNode)nodeMap.get(new Integer(parentId));
	        		parent = (SwcTreeNode)nodeMap.get(parentId);
	        	}

    			addNodeTypeless(node,parent,nodeMap,addNodeSideMap);

        	}
	        m = null;
	    }

		// Assuming only multifurcation is root
		SwcTreeNode tree;
		for (int i = 0; root != null && i < root.getChildren().size(); i++){
			//System.out.println("filtering "+i);
			tree = removeContinuations(root.getChildren().get(i),calcPathDist);
			binaryTrees.add(convertTreeToBinary(tree));
		}
		
		System.out.println("Done assembling trees from swc");
		
		return binaryTrees;
	}

	/**
	 * 
	 * @param inputFile
	 * @param classification
	 * @param minSize
	 * @param calcPathDist
	 * @return
	 * Map of NeuronProcess Lists (keyed by type), each containing a typeId and BinaryTree. 
	 * Assumes any number of binary trees branching from various locations, where the parent is of a different type
	 * @throws Exception
	 */
	public static Map<Integer,List<NeuronProcess<BinaryTreeNode>>> convertSwcToNeuronProcesses(File inputFile) throws Exception{
		return convertSwcToNeuronProcesses(inputFile, new TreeSet<Integer>());
	}
	public static Map<Integer,List<NeuronProcess<BinaryTreeNode>>> convertSwcToNeuronProcesses(File inputFile, Set<Integer> ignoreTypes) throws Exception{
	    
		Map<Integer,List<NeuronProcess<BinaryTreeNode>>> neuronProcesses = new TreeMap<Integer,List<NeuronProcess<BinaryTreeNode>>>();
		
		// 1. Builds individual trees
		// 2. Removes continuations.
		// 3. Transforms each into binary tree since all tree info is now known (provides stability in transformation).

		BufferedReader reader = new BufferedReader(new FileReader(inputFile));

		String dataString = "";
		Map<Integer,SwcTreeNode> nodeMap = new HashMap<Integer,SwcTreeNode>();
		Map<SwcTreeNode,Integer> addNodeSideMap = new HashMap<SwcTreeNode,Integer>();
		SwcTreeNode node, parent;
		Matcher m;
		int type, nodeId, parentId;
		List<SwcTreeNode> treeRoots = new ArrayList<SwcTreeNode>();
		NeuronProcess<BinaryTreeNode> neuronProcess;
		
		// 1. Assembles individual trees
		while ( (dataString=reader.readLine()) != null ) { 
	        m = swcDataPattern.matcher(dataString);
	        //System.out.println(dataString);
	        
	        if (m.matches()){
	        	node = new SwcTreeNode();
	        	nodeId = Integer.parseInt(m.group(1));
	        	node.setNodeId(nodeId);
	        	type = Integer.parseInt(m.group(2));
	        	node.setType(type);
	        	node.setX(Double.parseDouble(m.group(3)));
	        	node.setY(Double.parseDouble(m.group(4)));
	        	node.setZ(Double.parseDouble(m.group(5)));
	        	node.setRadius(Double.parseDouble(m.group(6)));
	        	parentId = Integer.parseInt(m.group(7));
	        	node.setParentId(parentId);
	        	//System.out.println(nodeId + " " + type + " " + parentId);
	
        		// Get parent from nodeMap
	        	parent = null;
	        	if (nodeMap.containsKey(parentId)){
	        		//parent = (ParentedBinaryTreeNode)nodeMap.get(new Integer(parentId));
	        		parent = nodeMap.get(parentId);
	        	}
	        	if (parentId == -1 || parent.getType() != node.getType()){
	        		treeRoots.add(node);
	        		parent = null;
	        	}

        		addNode(node,parent,nodeMap,addNodeSideMap);

        	}
	        m = null;
	    }

		SwcTreeNode tree;
		BinaryTreeNode binaryTree;
		List<NeuronProcess<BinaryTreeNode>> processList;
		for (int i = 0; i < treeRoots.size(); i++){
			//System.out.println("filtering "+i);
			tree = treeRoots.get(i);
			if (!ignoreTypes.contains(tree.getType())){
				/* 4/21/14 Changing to remove continuations after converting to binary so that extra binary nodes are created
				and positioned based on local child segment position and not based on remote child bifurcation position. */
				//tree = removeContinuations(tree, true);
				//neuronProcess = new NeuronProcess<BinaryTreeNode>(convertTreeToBinary(tree));
				binaryTree = convertTreeToBinary(tree);
				neuronProcess = new NeuronProcess<BinaryTreeNode>(removeContinuations(binaryTree, true));
				neuronProcess.setTypeId(tree.getType());
				if (!neuronProcesses.containsKey(tree.getType())){
					neuronProcesses.put(tree.getType(), new ArrayList<NeuronProcess<BinaryTreeNode>>());
				}
				processList = neuronProcesses.get(tree.getType());
				processList.add(neuronProcess);
			}
		}
		
		System.out.println("Done assembling trees from swc");
		
		return neuronProcesses;
	}

	public static BinaryTreeNode convertToBinaryTreeNode(List<SwcTreeNode> roots) throws Exception{
		return assembleProcesses(roots, false);
	}
	public static BinaryTreeNode convertToBinaryTreeNode(List<SwcTreeNode> roots, boolean calcPathDist)
	throws Exception{
		return assembleProcesses(roots, calcPathDist);
	}
	public static BinaryTreeNode convertToBinaryTreeNode(SwcTreeNode root)
	throws Exception{
		return convertToBinaryTreeNode(root, false);
	}
	public static BinaryTreeNode convertToBinaryTreeNode(SwcTreeNode root, boolean calcPathDist)
	throws Exception{
		SwcTreeNode copy = root.copy();
		removeContinuations(copy,calcPathDist);
		return convertTreeToBinary(copy);
	}
	
	public static SwcTreeNode removeContinuations(SwcTreeNode node){
		return removeContinuations(node, false);
	}
	public static void calculatePathLength(SwcTreeNode root){
    	Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
    	SwcTreeNode node;
    	List<SwcTreeNode> children;
    	
    	root.setSecondaryData(new SwcSecondaryData());
    	if (!root.hasChildren()){
    		return;
    	}
    	if (root.getChildren().size() == 1){
    		SwcTreeNode child = root.getChildren().get(0);
			calcSegmentProperties(root, child);
    		stack.push(child);
    	}
    	else{
    		stack.push(root);
    	}

    	while (!stack.isEmpty()){
    		node = stack.pop();
    		if (node.hasChildren()){
    			children = node.getChildren();
    			for (SwcTreeNode child : node.getChildren()){
    				stack.push(child);
   					calcSegmentProperties(node, child);
    			}
    			// Accumulate node data in child since node will be removed
    			if (children.size() == 1){
    				SwcTreeNode child = children.get(0);

					child.getSecondaryData().setPathLength(
							child.getSecondaryData().getPathLength()
							+ node.getSecondaryData().getPathLength());
					child.getSecondaryData().setSurfaceArea(
							child.getSecondaryData().getPathLength()
							* Math.PI * child.getDiameter());
					child.getSecondaryData().setVolume(
							child.getSecondaryData().getPathLength()
							* Math.PI * child.getRadiusSquared());
    			}
    		}
    	}
	}
	
	public static SwcTreeNode removeContinuations(SwcTreeNode root, boolean calcPathDist){
    	Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
    	SwcTreeNode node;
    	List<SwcTreeNode> children;
    	
    	root.setSecondaryData(new SwcSecondaryData());
    	if (!root.hasChildren()){
    		return root;
    	}
    	if (root.getChildren().size() == 1){
    		SwcTreeNode child = root.getChildren().get(0);
			if (calcPathDist){
				calcSegmentProperties(root, child);
			}
    		stack.push(child);
    	}
    	else{
    		stack.push(root);
    	}

    	while (!stack.isEmpty()){
    		node = stack.pop();
    		if (node.hasChildren()){
    			children = node.getChildren();
    			for (SwcTreeNode child : node.getChildren()){
    				stack.push(child);
    				if (calcPathDist){
    					calcSegmentProperties(node, child);
    				}
    			}
    			// Accumulate node data in child since node will be removed
    			if (children.size() == 1){
    				SwcTreeNode child = children.get(0);
    				if (calcPathDist && node.getSecondaryData() != null){
    					child.getSecondaryData().setPathLength(
    							child.getSecondaryData().getPathLength()
    							+ node.getSecondaryData().getPathLength());
    					child.getSecondaryData().setSurfaceArea(
    							child.getSecondaryData().getPathLength()
    							* Math.PI * child.getDiameter());
    					child.getSecondaryData().setVolume(
    							child.getSecondaryData().getPathLength()
    							* Math.PI * child.getRadiusSquared());
    				}
    				
					// add child to parent, remove self
					node.getParent().addChildNode(child);
					node.detach();
    			}
    		}
    	}
    	return root;
	}

	public static BinaryTreeNode removeContinuations(BinaryTreeNode root, boolean calcPathDist){
    	Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>();
    	BinaryTreeNode node;
    	SwcDataNode data, childData;
    	List<BinaryTreeNode> children;
    	
    	root.getSwcData().setSecondaryData(new SwcSecondaryData());
    	if (!root.hasChildren()){
    		return root;
    	}
    	if (root.getChildren().size() == 1){
    		BinaryTreeNode child = root.getChildren().get(0);
			if (calcPathDist){
				calcSegmentProperties(root.getSwcData(), child.getSwcData());
			}
    		stack.push(child);
    	}
    	else{
    		stack.push(root);
    	}

    	while (!stack.isEmpty()){
    		node = stack.pop();
    		if (node.hasChildren()){
    			children = node.getChildren();
    			for (BinaryTreeNode child : node.getChildren()){
    				stack.push(child);
    				if (calcPathDist){
    					calcSegmentProperties(node.getSwcData(), child.getSwcData());
    				}
    			}
    			// Accumulate node data in child since node will be removed
    			if (children.size() == 1){
    				BinaryTreeNode child = children.get(0);
    				data = node.getSwcData();
    				if (calcPathDist && data.getSecondaryData() != null){
    					childData = child.getSwcData(); 
    					childData.getSecondaryData().setPathLength(
    							childData.getSecondaryData().getPathLength()
    							+ data.getSecondaryData().getPathLength());
    					childData.getSecondaryData().setSurfaceArea(
    							childData.getSecondaryData().getPathLength()
    							* Math.PI * childData.getDiameter());
    					childData.getSecondaryData().setVolume(
    							childData.getSecondaryData().getPathLength()
    							* Math.PI * childData.getRadiusSquared());
    				}
    				
					// add child to parent, remove self
    				if (node.isLeft()){
    					node.getParent().setLeft(child);
    				}else{
    					node.getParent().setRight(child);
    				}
    			}
    		}
    	}
    	return root;
	}

	private static void calcSegmentProperties(SwcTreeNode parent, SwcTreeNode node){
		if (node.getSecondaryData() == null){
			node.setSecondaryData(new SwcSecondaryData());
		}
		if (node.getSecondaryData().getPathLength() == 0){
			node.getSecondaryData().setPathLength(
					SwcDataUtils.getDistance(parent.getData(), node.getData()));
			node.getSecondaryData().setSurfaceArea(
					node.getSecondaryData().getPathLength()
					* Math.PI * node.getDiameter());
			node.getSecondaryData().setVolume(
					node.getSecondaryData().getPathLength()
					* Math.PI * node.getRadiusSquared());
		}
	}

	private static void calcSegmentProperties(SwcDataNode parent, SwcDataNode node){
		if (node.getSecondaryData() == null){
			node.setSecondaryData(new SwcSecondaryData());
		}
		if (node.getSecondaryData().getPathLength() == 0){
			node.getSecondaryData().setPathLength(
					SwcDataUtils.getDistance(parent, node));
			node.getSecondaryData().setSurfaceArea(
					node.getSecondaryData().getPathLength()
					* Math.PI * node.getDiameter());
			node.getSecondaryData().setVolume(
					node.getSecondaryData().getPathLength()
					* Math.PI * node.getRadiusSquared());
		}
	}

	public static BinaryTreeNode assembleProcesses(List<SwcTreeNode> trees, boolean calcPathDist)throws Exception{
		SwcTreeNode tmpHead, node;
		if (trees.size() == 1){
			tmpHead = trees.get(0);
		}
		else{
			tmpHead = new SwcTreeNode();
			for (int i = 0; i < trees.size(); i++){
				node = trees.get(i);
				node.setType(0);
				tmpHead.addChildNode(node);
			}
		}
		tmpHead = removeContinuations(tmpHead, calcPathDist);
		BinaryTreeNode topNode = convertTreeToBinary(tmpHead);
		if (debug){
			System.out.println("Tree size: "+topNode.getTreeSize());
		}
		return topNode;
	}
	
	private static void addNode(
			SwcTreeNode node,
			SwcTreeNode parent,
			Map<Integer,SwcTreeNode> nodeMap,
			Map<SwcTreeNode,Integer> addNodeSideMap){

		if (node != null){
			//System.out.println("Parent: "+parent);
    		// Put new node in map
    		nodeMap.put(node.getNodeId(), node);
    		// Set node in meta data map (used for nodes with > 2 children)
    		addNodeSideMap.put(node, 0);
    		// If parent is null, then this is just the start of a new tree
    		/* If parent is of a different node type, DO NOT ATTACH
    		if (parent != null && 
    				parent.getType() == node.getType()){
    			parent.addChildNode(node);
    		}*/
    		/* Now attaching even if parent is of a different type. This should have no major impact as the node
    		 * can be set as the root of a tree in any case.
    		 */
    		if (parent != null){
    			parent.addChildNode(node);
    		}
    	}
	}

	private static void addNodeTypeless(
			SwcTreeNode node,
			SwcTreeNode parent,
			Map<Integer,SwcTreeNode> nodeMap,
			Map<SwcTreeNode,Integer> addNodeSideMap){

		if (node != null){
			//System.out.println("Parent: "+parent);
    		// Put new node in map
    		nodeMap.put(node.getNodeId(), node);
    		// Set node in meta data map (used for nodes with > 2 children)
    		addNodeSideMap.put(node, 0);
    		// If parent is null, then this is just the start of a new tree
    		// If parent is of a different node type, DO NOT ATTACH
    		if (parent != null){
    			parent.addChildNode(node);
    		}
    	}
	}

	private static void addNodeTypeless(
			SwcTreeNode node,
			SwcTreeNode parent,
			Map<Integer,SwcTreeNode> nodeMap){

		if (node != null){
			//System.out.println("Parent: "+parent);
    		// Put new node in map
    		nodeMap.put(node.getNodeId(), node);
    		// If parent is null, then this is just the start of a new tree
    		// If parent is of a different node type, DO NOT ATTACH
    		if (parent != null){
    			parent.addChildNode(node);
    		}
    	}
	}

	// Assuming continuations have been removed
	public static BinaryTreeNode convertTreeToBinary(SwcTreeNode head)throws Exception{
		if (head == null){
			return null;
		}
		if (!head.hasChildren()){
			return new BinaryTreeNode(head.getData());
		}
		BinaryTreeNode rootBifurcation, node, tmp;
		SwcTreeNode treeNode;
		Stack<BinaryTreeNode> stack = new Stack<BinaryTreeNode>(),
			recentTreeStack = new Stack<BinaryTreeNode>();
		
		if (head.getChildren().size() == 1){
			rootBifurcation = new BinaryTreeNode(head.getChildren().get(0));
			rootBifurcation.setRoot(new TreeRoot<BinaryTreeNode>(rootBifurcation, head.getData()));
		}
		else{
			rootBifurcation = new BinaryTreeNode(head);
		}
		stack.push(rootBifurcation);
		//System.out.println("Converting Tree to Binary");
		while (!stack.isEmpty()){
			node = stack.pop();
			treeNode = (SwcTreeNode)node.getData();
			//System.out.println("Node: "+treeNode.getData());
			// Copy tree node data to binary tree node data
			node.setData(treeNode.getData());
			//System.out.println("Node data: "+node.getData());
			if (treeNode.hasChildren() && treeNode.getChildren().size() > 0){
				tmp = convertTreeNodeToBinaryTreeNode(
						node.getSwcData(),treeNode.getChildren());
				node.setLeft(tmp.getLeft());
				node.setRight(tmp.getRight());
				// Find new bottom nodes
				if (tmp.getLeft() != null) recentTreeStack.push(tmp.getLeft());
				if (tmp.getRight() != null) recentTreeStack.push(tmp.getRight());
				while (!recentTreeStack.isEmpty()){
					node = recentTreeStack.pop();
					if (node.hasLeft()){
						recentTreeStack.push(node.getLeft());
						recentTreeStack.push(node.getRight());
					}
					else{
						stack.push(node);
					}
				}
			}
		}
		
		return rootBifurcation;
	}

	private static BinaryTreeNode convertTreeNodeToBinaryTreeNode(
			SwcDataNode data, Collection<SwcTreeNode> children) {
		BinaryTreeNode newNode = null;
		SwcDataNode swcData1, swcData2, newSwcDataNode;
		NodePair nodePair, bestNodePair = new NodePair();
		int best1, best2;
		Object object;
		List<SwcDataNode> dataList;
		Map<SwcDataNode,Object> childMap = new HashMap<SwcDataNode,Object>();
		for (SwcTreeNode child : children){
			childMap.put(child.getData(), child);
		}
		if (childMap.size() == 1){
			return new BinaryTreeNode(childMap.get(0));
		}
		
		// Keep going until there is only one node left on top, it will be parent to a binary tree
		while (childMap.size() > 1){
			best1 = 0; best2 = 1;
			bestNodePair = new NodePair();
			bestNodePair.setDistanceSq(-1);

			// Loop through multiple children, find closest two nodes
			dataList = new ArrayList<SwcDataNode>(childMap.keySet());
			for (int i = 0; i < childMap.size()-1; i++){
				swcData1 = dataList.get(i);

				// Loop through the rest, comparing current child to all others
				for (int j = i+1; j < childMap.size(); j++){
					swcData2 = dataList.get(j);

					// create a node pair for determining closest two nodes
					nodePair = new NodePair(swcData1,swcData2);
					if (bestNodePair.getDistanceSq() == -1 
							|| nodePair.getDistanceSq() < bestNodePair.getDistanceSq()){
						bestNodePair = nodePair;
						best1 = i;
						best2 = j;
					}
				}
			}
			
			// Insert new node with coordinates between 2 closest nodes
			newNode = new BinaryTreeNode();
			newSwcDataNode = new SwcDataNode();
			newSwcDataNode.setSecondaryData(new SwcSecondaryData());
			newNode.setData(newSwcDataNode);
			if (data.getType() == 0){
				// Set new data node type to first child node if both are real nodes
				// Keep type as 0 (default) if either child is a "root" node
				if (bestNodePair.getSwcDataNode1().getType() != 0 
						&& bestNodePair.getSwcDataNode2().getType() != 0){
					newSwcDataNode.setType(bestNodePair.getSwcDataNode1().getType());
				}
				newSwcDataNode.setX(
						(bestNodePair.getSwcDataNode1().getX() + bestNodePair.getSwcDataNode2().getX())/2);
				newSwcDataNode.setY(
						(bestNodePair.getSwcDataNode1().getY() + bestNodePair.getSwcDataNode2().getY())/2);
				newSwcDataNode.setZ(
						(bestNodePair.getSwcDataNode1().getZ() + bestNodePair.getSwcDataNode2().getZ())/2);
			}
			else{
				newSwcDataNode.setNodeId(data.getNodeId());
				newSwcDataNode.setX(data.getX());
				newSwcDataNode.setY(data.getY());
				newSwcDataNode.setZ(data.getZ());
				newSwcDataNode.setType(data.getType());
				newSwcDataNode.setSecondaryData(new SwcSecondaryData());
			}
			// Make closest nodes children of the new node
			object = childMap.remove(dataList.get(best1));
			if (object.getClass().equals(SwcTreeNode.class)){
				newNode.setLeft(new BinaryTreeNode());
				newNode.getLeft().setData(object);
			}
			else{
				newNode.setLeft((BinaryTreeNode)object);
			}
			//object = childList.remove(best2-1); // best1 has already been removed, and best2 > best1
			object = childMap.remove(dataList.get(best2));
			if (object.getClass().equals(SwcTreeNode.class)){
				newNode.setRight(new BinaryTreeNode());
				newNode.getRight().setData(object);
			}
			else{
				newNode.setRight((BinaryTreeNode)object);
			}
			childMap.put(newNode.getSwcData(),newNode);
		}
		

		return newNode;
	}

	/**
	 * 
	 * @param inputFile
	 * @return String containing all comments in the file
	 * @throws Exception
	 */
	public static String readSwcComments(File inputFile) throws IOException{
	    
		StringBuffer comments = new StringBuffer();
		BufferedReader reader = new BufferedReader(new FileReader(inputFile));

		String dataString = "";
		
		// 1. Assembles individual trees
		while ( (dataString=reader.readLine()) != null ) {
			dataString = dataString.trim();
	        if (dataString.length() > 0 && dataString.charAt(0)=='#'){
	        	comments.append(dataString+"\n");
	        }
		}
		
		return comments.toString();
	}
	
	public static List<SwcTreeNode> convertSwcToTreeNodesSpecial(
			File inputFile) throws IOException{

		BufferedReader reader 
    		= new BufferedReader(new FileReader(inputFile));

		String dataString = "";
		Map<Integer,SwcTreeNode> nodeMap = new HashMap<Integer,SwcTreeNode>();
		SwcTreeNode node, parent;
		Matcher m;
		int type, nodeId, parentId;
		
		List<SwcTreeNode> trees = new ArrayList<SwcTreeNode>();
		
		Map<Integer,List<SwcTreeNode>> unattachedNodesByParentId = new HashMap<Integer,List<SwcTreeNode>>();
		List<SwcTreeNode> unattachedChildren;
		
		// 1. Assembles axon, basal trees, and apical trees (non-binary).
		while ( (dataString=reader.readLine()) != null ) { 
	        m = swcDataPattern.matcher(dataString);
	        //System.out.println(dataString);
	        
	        if (m.matches()){
	        	node = new SwcTreeNode();
	        	nodeId = Integer.parseInt(m.group(1));
	        	node.setNodeId(nodeId);
	        	type = Integer.parseInt(m.group(2));
	        	node.setType(type);
	        	node.setX(Double.parseDouble(m.group(3)));
	        	node.setY(Double.parseDouble(m.group(4)));
	        	node.setZ(Double.parseDouble(m.group(5)));
	        	node.setRadius(Double.parseDouble(m.group(6)));
	        	parentId = Integer.parseInt(m.group(7));
	        	node.setParentId(parentId);

        		// Get parent from nodeMap
	        	parent = null;
	        	if (parentId == -1){ // Root node
	        		nodeMap.put(node.getNodeId(), node);
	        		trees.add(node);
	        		//System.out.println("New root: "+node.getNodeId());
	        	}
	        	// Node with a parent already created
	        	else if (nodeMap.containsKey(parentId)){
	        		parent = nodeMap.get(parentId);
    				addNodeTypeless(node,parent,nodeMap);
	        	}
	        	// Node with a parent not yet created
	        	else{
	        		nodeMap.put(node.getNodeId(), node);
	        		if (!unattachedNodesByParentId.containsKey(parentId)){
	        			unattachedNodesByParentId.put(parentId,new LinkedList<SwcTreeNode>());
	        		}
	        		unattachedNodesByParentId.get(parentId).add(node);
	        	}
	        	
	        	// Assign children that were created before their parent (current node)
	        	if (unattachedNodesByParentId.containsKey(nodeId)){
	        		unattachedChildren = unattachedNodesByParentId.remove(nodeId);
	        		for (SwcTreeNode child : unattachedChildren){
	        			node.addChildNode(child);
	        		}
	        	}

	        }
	        m = null;
	    }

		if (debug) System.out.println("Done assembling trees from swc");
		return trees;
	}

	public static int getArborTypeId(String arborType){
		if (arborType.equalsIgnoreCase("Axon")){
			return 2;
		}
		else if (arborType.equalsIgnoreCase("ApicalDendrite")){
			return 4;
		}
		else if (arborType.equalsIgnoreCase("Dendrite") || arborType.equalsIgnoreCase("BasaDendrite")){
			return 3;
		}
		return 0;
	}

	
}
