package org.cn3.morpho.data;

import java.io.File;
import java.text.NumberFormat;
import java.util.List;

import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.morpho.domain.Soma;
import org.cn3.morpho.domain.SwcTreeNode;
import org.cn3.morpho.domain.SwcDataNode;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;

/**
 * 
 * @author gillette
 *
 * Converts an SWC file into a representation of nodes with each's distance to its
 * parent, in the Newick file format.
 * 
 * E.g. ((:6.2,:2.1):3.3,((:2.8,:3.4):0.3,(:1.7,:6.3):2.3):3.8)
 * 
 * java org.cn3.morpho.data.SwcToNewick -i "C:/Documents and Settings/gillette/My Documents/research/MultipleAlignment and Domains/ri06-ApicalDendrite-STLLocalAlignmentVs-Domain Sqrt 1.5.swc" -o "C:/Documents and Settings/gillette/My Documents/research/MultipleAlignment and Domains/" -a 4 -L ACT -b 
 * java org.cn3.morpho.data.SwcToNewick -i "C:/Documents and Settings/gillette/My Documents/research/MultipleAlignment and Domains/ri06-ApicalDendrite-STLLocalAlignmentVs-Domain Sqrt 1.5.swc" -o "C:/Documents and Settings/gillette/My Documents/research/MultipleAlignment and Domains/" -a 4 -L NONE -b
 * 
 * java org.cn3.morpho.data.SwcToNewick -i "C:/Documents and Settings/gillette/My Documents/research/papers & presentations/PASTA Paper/Paper2 - AlignmentClusteringDomains/Figures/Workspace/pc2b-ApicalDendrite-STLAlignmentVs-Apical Cluster 1 50% Consensus.swc" -a 0 -L NODE_TYPE -b
 * 
 * java org.cn3.morpho.data.SwcToNewick -i "C:/Documents and Settings/gillette/My Documents/NMO/wrl&swc/ca3b-N2-l51-spliced.swc" -o "C:/Documents and Settings/gillette/My Documents/research/papers & presentations/PASTA Paper/Paper1 - Motifs/figures/" -a 2 -L NONE -b
 * java org.cn3.morpho.data.SwcToNewick -i "C:/Documents and Settings/gillette/My Documents/NMO/wrl&swc/ca3b-N2-l51-spliced.swc" -o "C:/Documents and Settings/gillette/My Documents/research/papers & presentations/PASTA Paper/Paper1 - Motifs/figures/" -a 3 -L NONE -b
 * java org.cn3.morpho.data.SwcToNewick -i "C:/Documents and Settings/gillette/My Documents/NMO/wrl&swc/ca3b-N2-l51-spliced.swc" -o "C:/Documents and Settings/gillette/My Documents/research/papers & presentations/PASTA Paper/Paper1 - Motifs/figures/" -a 4 -L NONE -b
 *
 */
public class SwcToNewick {

	private static final int DEFAULT_MAX_FRACTION_DIGITS = 1;
	public enum NewickLabel {NONE, ACT, NODE_TYPE, ORDER_AND_TYPE, DATA};
	public enum Order {STL, LTS, ALTERNATING, RANDOM};
	
	private static interface EdgeLabeler{
		public String label(BinaryTreeNode node) throws Exception;
	}
	private static class BlankLabeler implements EdgeLabeler{
		public String label(BinaryTreeNode node) throws Exception{
			return " ";
		}
	}
	private static class TypeLabeler implements EdgeLabeler{
		public String label(BinaryTreeNode node) throws Exception{
			if (node.getData() == null || !node.getData().getClass().equals(SwcDataNode.class)){
				throw new Exception("Node does not contain SwcDataNode data object");
			}
			return ""+node.getSwcData().getType();
		}
	}
	private static class OrderAndTypeLabeler implements EdgeLabeler{
		private int count = 0;
		public String label(BinaryTreeNode node) throws Exception{
			return (++count)+"-"+((SwcDataNode)(node.getData())).getType();
		}
	}
	private static class ACTLabeler implements EdgeLabeler{
		public String label(BinaryTreeNode node) throws Exception{
			if (node.hasBothChildren()){
				if (node.getRight().hasBothChildren() && node.getLeft().hasBothChildren()){
					return "A";
				}
				else if (node.getRight().hasBothChildren() || node.getLeft().hasBothChildren()){
					return "C";
				}
				else{
					return "T";
				}
			}else{
				return "x";
			}
		}
	}
	private static class DataLabeler implements EdgeLabeler{
		public String label(BinaryTreeNode node) throws Exception{
			if (node.getData() == null){
				return "";
			}
			return ""+node.getData();
		}
	}
	
	public static String getNewickString(File swcFile) throws Exception{
		return getNewickString(swcFile,DEFAULT_MAX_FRACTION_DIGITS);
	}
	public static String getNewickString(File swcFile, int maxFractionDigits) throws Exception{
		List<SwcTreeNode> treeList = SWCReader.convertSwcToTreeNodes(swcFile);
		String newickString = "(";
		SwcTreeNode node;
		for (int i = 0; i < treeList.size(); i++){
			node = (SwcTreeNode)treeList.get(i);
			if (i > 0){
				newickString += ",";
			}
			newickString += getNewickString(node, maxFractionDigits);
		}
		newickString += ")";
		return newickString;
	}

	public static String getBinaryTreeNewickString(File swcFile) throws Exception{
		return getBinaryTreeNewickString(swcFile, DEFAULT_MAX_FRACTION_DIGITS, NewickLabel.NONE);
	}
	public static String getBinaryTreeNewickString(File swcFile, NewickLabel labelType) throws Exception{
		return getBinaryTreeNewickString(swcFile, DEFAULT_MAX_FRACTION_DIGITS, labelType, Order.STL);
	}
	public static String getBinaryTreeNewickString(File swcFile, NewickLabel labelType, Order order) throws Exception{
		return getBinaryTreeNewickString(swcFile, DEFAULT_MAX_FRACTION_DIGITS, labelType, order);
	}
	public static String getBinaryTreeNewickString(File swcFile, int maxFractionDigits, NewickLabel labelType) throws Exception{
		return getBinaryTreeNewickString(swcFile, maxFractionDigits, labelType, Order.STL);
	}
	public static String getBinaryTreeNewickString(File swcFile, int maxFractionDigits, NewickLabel labelType, Order order) throws Exception{
		List<BinaryTreeNode> treeList = SWCReader.convertSwcToBinaryTreeList(swcFile,true);
		
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(maxFractionDigits);
		nf.setGroupingUsed(false);

		//String newickString = "(";
		String newickString = "";
		BinaryTreeNode node;
		for (int i = 0; i < treeList.size(); i++){
			node = treeList.get(i);
			if (i > 0){
				newickString += ",";
			}
			newickString += getNewickString(node,nf,labelType,order);
		}
		//newickString += ")";
		return newickString;
	}

	public static String getSpecificBinaryTreeNewickString(File swcFile, int arborType) throws Exception{
		return getSpecificBinaryTreeNewickString(swcFile, arborType, DEFAULT_MAX_FRACTION_DIGITS, NewickLabel.NONE);
	}
	public static String getSpecificBinaryTreeNewickString(File swcFile, int arborType, NewickLabel labelType) throws Exception{
		return getSpecificBinaryTreeNewickString(swcFile, arborType, DEFAULT_MAX_FRACTION_DIGITS, labelType);
	}
	public static String getSpecificBinaryTreeNewickString(File swcFile, int arborType, NewickLabel labelType, Order order) throws Exception{
		return getSpecificBinaryTreeNewickString(swcFile, arborType, DEFAULT_MAX_FRACTION_DIGITS, labelType, order);
	}
	public static String getSpecificBinaryTreeNewickString(File swcFile, int arborType, 
			int maxFractionDigits, NewickLabel labelType) throws Exception{
		return getSpecificBinaryTreeNewickString(swcFile, arborType, maxFractionDigits, labelType, Order.STL);
	}
	public static String getSpecificBinaryTreeNewickString(File swcFile, int arborType, 
			int maxFractionDigits, NewickLabel labelType, Order order) throws Exception{
		Soma<BinaryTreeNode> soma = SWCReader.convertSwcToBinarySoma(swcFile,true);
		
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(DEFAULT_MAX_FRACTION_DIGITS);
		nf.setGroupingUsed(false);
		
		switch (arborType){
		case NeuronMetadata.ARBOR_TYPE_BASAL:
		case NeuronMetadata.ARBOR_TYPE_DENDRITE:
			if (soma.hasBasalDendriteTree()){
				return(getNewickString(soma.getBasalDendriteTree().getTreeHead(),nf,labelType, order));
			}else{
				throw new Exception("File "+swcFile.getAbsolutePath()+" contains no dendritic (type 3) arbor");
			}
		case NeuronMetadata.ARBOR_TYPE_APICAL:
			if (soma.hasApicalDendriteTree()){
				return(getNewickString(soma.getApicalDendriteTree().getTreeHead(),nf,labelType, order));
			}else{
				throw new Exception("File "+swcFile.getAbsolutePath()+" contains no apical dendritic (type 4) arbor");
			}
		case NeuronMetadata.ARBOR_TYPE_AXON:
			if (soma.hasAxonalTree()){
				return(getNewickString(soma.getAxonalTree().getTreeHead(),nf,labelType, order));
			}else{
				throw new Exception("File "+swcFile.getAbsolutePath()+" contains no axonal (type 2) arbor");
			}
		default:
			throw new Exception("Unknown arbor type");
		}
	}

	public static String getNewickString(BinaryTreeNode tree) throws Exception{
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(DEFAULT_MAX_FRACTION_DIGITS);
		nf.setGroupingUsed(false);
		return getNewickString(tree,nf);
	}
	public static String getNewickString(BinaryTreeNode tree, int maxFractionDigits) throws Exception{
		NumberFormat nf = null;
		if (maxFractionDigits >= 0){
			nf = NumberFormat.getInstance();
			nf.setMaximumFractionDigits(maxFractionDigits);
			nf.setGroupingUsed(false);
		}
		return getNewickString(tree,nf);
	}
	public static String getNewickString(BinaryTreeNode tree, NumberFormat numberFormat) throws Exception{
		return getNewickString(tree,numberFormat,new SwcToNewick.BlankLabeler());
	}
	public static String getOrderAndTypeLabeledNewickString(BinaryTreeNode tree, NumberFormat numberFormat) throws Exception{
		return getNewickString(tree,numberFormat,new SwcToNewick.OrderAndTypeLabeler());
	}
	public static String getTypeLabeledNewickString(BinaryTreeNode tree, NumberFormat numberFormat) throws Exception{
		return getNewickString(tree,numberFormat,new SwcToNewick.TypeLabeler());
	}
	public static String getACTNewickString(BinaryTreeNode tree, NumberFormat numberFormat) throws Exception{
		return getNewickString(tree,numberFormat,new SwcToNewick.ACTLabeler());
	}
	public static String getNewickString(BinaryTreeNode tree, NewickLabel labelType)
	throws Exception{
		return getNewickString(tree, null, labelType, Order.STL);
	}
	public static String getNewickString(BinaryTreeNode tree, NumberFormat numberFormat, NewickLabel labelType)
	throws Exception{
		return getNewickString(tree, numberFormat, labelType, Order.STL);
	}
	public static String getNewickString(BinaryTreeNode tree, NumberFormat numberFormat, NewickLabel labelType, Order order)
	throws Exception{
		switch(labelType){
		case NONE:
			return getNewickSubtreeString(tree,numberFormat,new SwcToNewick.BlankLabeler(), order)+";";
		case ACT:
			return getNewickSubtreeString(tree,numberFormat,new SwcToNewick.ACTLabeler(), order)+";";
		case NODE_TYPE:
			return getNewickSubtreeString(tree,numberFormat,new SwcToNewick.TypeLabeler(), order)+";";
		case ORDER_AND_TYPE:
			return getNewickSubtreeString(tree,numberFormat,new SwcToNewick.OrderAndTypeLabeler(), order)+";";
		case DATA:
			return getNewickSubtreeString(tree,numberFormat,new SwcToNewick.DataLabeler(), order)+";";
		default:
			throw new Exception("Unknown label type ("+labelType+").");
		}
	}
	public static String getNewickString(BinaryTreeNode tree, NumberFormat numberFormat, EdgeLabeler labeler)
	throws Exception{
		return getNewickSubtreeString(tree,numberFormat,labeler)+";";
	}
	public static String getNewickSubtreeString(BinaryTreeNode tree, NumberFormat numberFormat, EdgeLabeler labeler)
	throws Exception{
		return getNewickSubtreeString(tree, numberFormat, labeler, Order.STL);
	}
	public static String getNewickSubtreeString(BinaryTreeNode tree, NumberFormat numberFormat, EdgeLabeler labeler, Order order)
	throws Exception{
		if (order == Order.ALTERNATING){
			return getNewickSubtreeStringAlternating(tree, numberFormat, labeler, true);
		}
		String newickString = "";
		if (tree.hasChildren()){
			newickString = "(";
			if (tree.hasBothChildren()){
				if (order == Order.RANDOM || 
						(tree.getLeft().compareTo(tree.getRight()) > 0 ^ order == Order.LTS)){
					newickString += getNewickSubtreeString(tree.getLeft(), numberFormat, labeler);
					newickString += ","+getNewickSubtreeString(tree.getRight(), numberFormat, labeler);
				}else{
					newickString += getNewickSubtreeString(tree.getRight(), numberFormat, labeler);
					newickString += ","+getNewickSubtreeString(tree.getLeft(), numberFormat, labeler);
				}
			}else if (tree.hasLeft()){
				newickString += getNewickSubtreeString(tree.getLeft(), numberFormat, labeler);
			}else if (tree.hasRight()){
				if (tree.hasLeft()){
					newickString += ",";
				}
				newickString += getNewickSubtreeString(tree.getRight(), numberFormat, labeler);
			}
			newickString += ")";
			newickString += labeler.label(tree);
/*
			if (encodeACT && tree.hasBothChildren()){
				if (tree.getRight().hasBothChildren() && tree.getLeft().hasBothChildren()){
					newickString += "A";
				}
				else if (tree.getRight().hasBothChildren() || tree.getLeft().hasBothChildren()){
					newickString += "C";
				}
				else{
					newickString += "T";
				}
			}*/
		}
		else{
			//newickString += "x";
			newickString += labeler.label(tree);
		}
		
		if (numberFormat != null){
			if (tree.getSwcData() != null && tree.getSwcData().getSecondaryData() != null){
				newickString = newickString + ":" + numberFormat.format(tree.getSwcData().getSecondaryData().getPathLength());
			}
			else{
				newickString = newickString + ":0";
			}
		}
		else{
			newickString = newickString + ":1";
		}
		return(newickString);
	}
	public static String getNewickSubtreeStringAlternating(BinaryTreeNode tree, NumberFormat numberFormat, EdgeLabeler labeler, boolean largerFirst)
	throws Exception{
		String newickString = "";
		if (tree.hasChildren()){
			newickString = "(";
			if (tree.hasBothChildren()){
				if (tree.getLeft().compareTo(tree.getRight()) > 0 ^ largerFirst){
					newickString += getNewickSubtreeStringAlternating(tree.getLeft(), numberFormat, labeler, !largerFirst);
					newickString += ","+getNewickSubtreeStringAlternating(tree.getRight(), numberFormat, labeler, !largerFirst);
				}else{
					newickString += getNewickSubtreeStringAlternating(tree.getRight(), numberFormat, labeler, !largerFirst);
					newickString += ","+getNewickSubtreeStringAlternating(tree.getLeft(), numberFormat, labeler, !largerFirst);
				}
			}else if (tree.hasLeft()){
				newickString += getNewickSubtreeStringAlternating(tree.getLeft(), numberFormat, labeler, !largerFirst);
			}else if (tree.hasRight()){
				if (tree.hasLeft()){
					newickString += ",";
				}
				newickString += getNewickSubtreeStringAlternating(tree.getRight(), numberFormat, labeler, !largerFirst);
			}
			newickString += ")";
			newickString += labeler.label(tree);
		}
		else{
			//newickString += "x";
			newickString += labeler.label(tree);
		}
		
		if (numberFormat != null){
			if (tree.getSwcData() != null && tree.getSwcData().getSecondaryData() != null){
				newickString = newickString + ":" + numberFormat.format(tree.getSwcData().getSecondaryData().getPathLength());
			}
			else{
				newickString = newickString + ":0";
			}
		}
		else{
			newickString = newickString + ":1";
		}
		return(newickString);
	}

	public static String getNewickString(SwcTreeNode tree){
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(DEFAULT_MAX_FRACTION_DIGITS);
		nf.setGroupingUsed(false);
		return getNewickString(tree,nf);
	}
	public static String getNewickString(SwcTreeNode tree, int maxFractionDigits){
		NumberFormat nf = null;
		if (maxFractionDigits >= 0){
			nf = NumberFormat.getInstance();
			nf.setMaximumFractionDigits(maxFractionDigits);
			nf.setGroupingUsed(false);
		}
		return getNewickString(tree,nf);
	}
	public static String getNewickString(SwcTreeNode tree, NumberFormat numberFormat){
		return getNewickSubtreeString(tree,numberFormat)+";";
	}
	public static String getNewickSubtreeString(SwcTreeNode tree, NumberFormat numberFormat){
		String newickString = "";
		if (tree.hasChildren()){
			newickString = "(";
			SwcTreeNode node;
			for (int i = 0; i < tree.getChildren().size(); i++){
				node = tree.getChild(i);
				if (i > 0){
					newickString += ",";
				}
				newickString += getNewickSubtreeString(node,numberFormat);
			}
			newickString += ")";
		}
		
		if (numberFormat != null && tree.getData() != null && tree.getData().getSecondaryData() != null){
			newickString = newickString + ":" + numberFormat.format(tree.getData().getSecondaryData().getPathLength());
		}
		return(newickString);
	}
	
	public static void createNewickFile(BinaryTreeNode tree, NewickLabel labelType, File outputfile) throws Exception{
		String newickString = getNewickString(tree,labelType);
		DataWriter.writeStringToFile(outputfile, newickString);
	}

	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);

			File swcFile = new File(config.getString(PARAM_INPUT_FILE));
			int arborType = config.getInt(PARAM_ARBOR_TYPE);
			boolean binaryTree = config.getBoolean(PARAM_BINARY_TREE);
			String labelTypeVal = config.getString(PARAM_LABEL_TYPE);
			NewickLabel labelType = NewickLabel.valueOf(labelTypeVal);
			Order order = Order.valueOf(config.getString(PARAM_TRAVERSAL_ORDER));
			
			if (!swcFile.getName().endsWith(".swc") && !swcFile.getName().endsWith(".swc")){
				throw new Exception("Requires an SWC file as input");
			}
			
			// Get base filename and create newick filename
			String newickFilename;
			if (config.userSpecified(PARAM_OUTPUT_DIR)){
				newickFilename = config.getString(PARAM_OUTPUT_DIR);
			}
			else{
				newickFilename = swcFile.getParent();
			}
			newickFilename += "/"+swcFile.getName().substring(0, swcFile.getName().length()-4);
			
			String newickString;
			if (binaryTree){
				if (arborType > 0){
					switch (arborType){
					case NeuronMetadata.ARBOR_TYPE_DENDRITE:
					case NeuronMetadata.ARBOR_TYPE_BASAL:
						newickFilename += "-Dendrite";
						break;
					case NeuronMetadata.ARBOR_TYPE_APICAL:
						newickFilename += "-Apical";
						break;
					case NeuronMetadata.ARBOR_TYPE_AXON:
						newickFilename += "-Axon";
						break;
					}
					newickString = getSpecificBinaryTreeNewickString(swcFile, arborType, labelType, order);
				}
				else{
					newickFilename += "-Binary";
					newickString = getBinaryTreeNewickString(swcFile, labelType, order);
				}
				switch(labelType){
				case ACT:
					newickFilename += "LabelACT";
					break;
				case NODE_TYPE:
					newickFilename += "LabelType";
					break;
				case ORDER_AND_TYPE:
					newickFilename += "LabelOrderType";
					break;
				case DATA:
					newickFilename += "LabelData";
					break;
				default:
				}
			}
			else{
				newickString = getNewickString(swcFile);
			}
			newickFilename += ".newick";
			
			DataWriter.writeStringToFile(newickFilename, newickString);
			System.out.println("Wrote newick string of "+swcFile.getAbsolutePath()+" to "+newickFilename);
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	private static final String PARAM_INPUT_FILE = "inputFile";
	private static final String PARAM_OUTPUT_DIR = "outputDir";
	private static final String PARAM_ARBOR_TYPE = "arborType";
	private static final String PARAM_BINARY_TREE = "binaryTree";
	private static final String PARAM_LABEL_TYPE = "labelType";
	private static final String PARAM_TRAVERSAL_ORDER = "traversal";
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
		FlaggedOption opt;

    	opt = new FlaggedOption(PARAM_INPUT_FILE)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true) 
		.setShortFlag('i') 
		.setLongFlag(PARAM_INPUT_FILE);
        jsap.registerParameter(opt);

    	opt = new FlaggedOption(PARAM_OUTPUT_DIR)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(false)
		.setShortFlag('o') 
		.setLongFlag(PARAM_OUTPUT_DIR);
        jsap.registerParameter(opt);

    	opt = new FlaggedOption(PARAM_ARBOR_TYPE)
		.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault("0")
		.setShortFlag('a') 
		.setLongFlag(PARAM_ARBOR_TYPE);
    	opt.setHelp("0 - All\n1 - Axon\n2 - Dendrite\n3 - Basal Dendrite\n4 - Apical Dendrite\n");
        jsap.registerParameter(opt);

    	opt = new FlaggedOption(PARAM_LABEL_TYPE)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(false)
		.setDefault("None")
		.setShortFlag('L') 
		.setLongFlag(PARAM_LABEL_TYPE);
    	opt.setHelp("NONE\nACT\nNODE_TYPE\nORDER_AND_TYPE\n");
        jsap.registerParameter(opt);
        
        Switch switchOpt = new Switch(PARAM_BINARY_TREE)
		.setShortFlag('b') 
		.setLongFlag(PARAM_BINARY_TREE);
        jsap.registerParameter(switchOpt);

    	opt = new FlaggedOption(PARAM_TRAVERSAL_ORDER)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(false)
		.setDefault("STL")
		.setShortFlag('O') 
		.setLongFlag(PARAM_TRAVERSAL_ORDER);
    	opt.setHelp("STL\nLTS\nRANDOM\nALTERNATING\n");
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + SwcToNewick.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        return config;
	}

	
}
