package org.cn3.morpho.domain;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.cn3.morpho.utils.BinaryTreeUtils;
import org.cn3.morpho.utils.SwcDataUtils;

/**
 * 
 * @author gillette
 *
 * The workhorse of the project, BinaryTreeNode contains various node metadata
 * and all functions for tree manipulation, including connecting to and
 * disconnecting from parent and child nodes.
 * 
 *
 */
public class BinaryTreeNode extends AbstractTreeNode<BinaryTreeNode,Object> implements Comparable<BinaryTreeNode>{
//implements TreeNode<BinaryTreeNode,Object>, Comparable<BinaryTreeNode>{
    
	private static final int RIGHT = 0;
    private static final int LEFT = 1;
    
    protected Object data;
    protected BinaryTreeNode parent;
    private BinaryTreeNode left,right;
    protected List<BinaryTreeNode> children = new ArrayList<BinaryTreeNode>(2);
    protected int side; // left or right of parent
    private TreeRoot<BinaryTreeNode> root;

    protected int treeSize;
    protected NodeMetaData metaData;
    private Map<Integer,Asymmetry> asymmetryMap;

    public BinaryTreeNode(){
        data = null;
        left = right = parent = null;
        treeSize = 1;
        children = new ArrayList<BinaryTreeNode>(2);
    }
    public BinaryTreeNode(Object d){
        data = d;
        left = right = parent = null;
        treeSize = 1;
        children = new ArrayList<BinaryTreeNode>(2);
    }
    public void setParent(BinaryTreeNode p){
    	setParent(p,true);
    }
    public void setParent(BinaryTreeNode p, boolean updateStats){
        parent = p;
        if (p != null && updateStats){
        	updateBranchStatisticsUp(p);
        	updateBranchOrderDown();
        }
    }
    
    public TreeRoot<BinaryTreeNode> getRoot() {
		return root;
	}
	public void setRoot(TreeRoot<BinaryTreeNode> root) {
		this.root = root;
	}
	public BinaryTreeNode getParent(){
    	return parent;
    }

    public void setLeft(BinaryTreeNode l){
        setLeft(l,true);
    }
    public void setLeft(BinaryTreeNode l, boolean updateStats){
        //super.setLeft(l,false);
    	children.remove(left);
    	children.add(l);
        if (left != null){
        	left.setParent(null,updateStats);
        }
        left = l;
        setShapeNum(null);
        if (l != null){
	        left.setParent(this,false);
	        left.side = LEFT;
        }
        if (updateStats){
        	updateBranchStatisticsUp(l);
        	if (left != null) left.updateBranchOrderDown();
        }else if (left != null) left.setBranchOrder(this.branchOrder+1);
    }
    
    public void setRight(BinaryTreeNode r){
    	setRight(r,true);
    }
    public void setRight(BinaryTreeNode r, boolean updateStats){
    	//super.setRight(r,false);
    	children.remove(right);
    	children.add(r);
        if (right != null){
           right.setParent(null,updateStats);
        }
        right = r;
        setShapeNum(null);
        if (r != null){
	        right.setParent(this,false);
	        right.side = RIGHT;
        }
        if (updateStats){
        	updateBranchStatisticsUp(r);
        	if (right != null) right.updateBranchOrderDown();
        }else if (right != null) right.setBranchOrder(this.branchOrder+1);
    	children = new ArrayList<BinaryTreeNode>();
    	children.add(r);
    	if (left != null){
    		children.add(left);
    	}
    }

    public void updateStatistics(){
    	int treeSize, maxHeight, totalHeight, leaves, leftSize, rightSize, asym;
    	treeSize = 1;
		leaves = totalHeight = maxHeight = leftSize = rightSize = asym = 0;
		if (this.getParent() != null){
			this.setBranchOrder(this.getParent().getBranchOrder()+1);
		}else{
			this.setBranchOrder(0);
		}
		if (this.getLeft() != null){
			leftSize = this.getLeft().getTreeSize();
			treeSize += this.getLeft().getTreeSize();
			leaves += this.getLeft().getLeaves();
			maxHeight = this.getLeft().getHeight() + 1;
			// totalLength is the total length of the children plus 1 for each child below (treeSize)
			totalHeight += this.getLeft().getTotalHeight()
				+ this.getLeft().getLeaves();
			asym += this.getLeft().getTotalAsymmetry();
		}
		else if (this.getRight() == null){
			leaves++;
		}
		if (this.getRight() != null){
			rightSize = this.getRight().getTreeSize();
			treeSize += this.getRight().getTreeSize();
			leaves += this.getRight().getLeaves();
			if (this.getRight().getHeight() >= maxHeight){
				maxHeight = this.getRight().getHeight() + 1;
			}
			totalHeight += this.getRight().getTotalHeight()
				+ this.getRight().getLeaves();
			asym += this.getRight().getTotalAsymmetry();
		}
		this.setTreeSize(treeSize);
		this.setLeaves(leaves);
		this.setHeight(maxHeight);
		this.setTotalHeight(totalHeight);
		// treeSize-1 so this node is not included
		this.setAverageHeight((float)totalHeight/leaves);
		if (treeSize > 2){
			this.setTotalAsymmetry(asym
				+ (double)(treeSize - 2 * Math.min(leftSize, rightSize)) /
					(treeSize - 2));
		}
		else{
			this.setTotalAsymmetry(0);
		}
    }
    
    public void setData(Object d){
        data = d;
    }
    public BinaryTreeNode getLeft(){
        return left;
    }
    public BinaryTreeNode getRight(){
        return right;
    }
    public List<BinaryTreeNode> getChildren(){
        return new ArrayList<BinaryTreeNode>(children);
    }
    public SwcDataNode getSwcData(){
    	if (data != null && data.getClass().equals(SwcDataNode.class)){
    		return (SwcDataNode)data;
    	}
        return null;
    }
    public Object getData(){
        return data;
    }
    public String toString(){
        return ""+hashCode()+" "+data;
    }
    public int getTreeSize(){
    	return treeSize;
    }
    
    public void setTreeSize(int treeSize) {
		this.treeSize = treeSize;
	}
    
	public int getTreeSizeRecursively(){
    	int size = 1;
    	if (left != null){
    		size += left.getTreeSizeRecursively();
    	}
    	if (right != null){
    		size += right.getTreeSizeRecursively();
    	}
    	return size;
    }


    public static void updateTreeSizeDown(BinaryTreeNode node){
    	BinaryTreeNode current;
    	Stack<BinaryTreeNode> assembleStack = new Stack<BinaryTreeNode>();
    	Stack<BinaryTreeNode> branchStack = new Stack<BinaryTreeNode>();
    	int treeSize, height;
    	assembleStack.push(node);
    	branchStack.push(node);
    	while (!assembleStack.isEmpty()){
    		current = assembleStack.pop();
    		if (current.hasLeft()){
    			assembleStack.push(current.getLeft());
    			branchStack.push(current.getLeft());
    		}
    		if (current.hasRight()){
    			assembleStack.push(current.getRight());
    			branchStack.push(current.getRight());
    		}
    	}
    	assembleStack = null;
    	while (!branchStack.isEmpty()){
    		current = branchStack.pop();
    		treeSize = 1;
    		height = 0;
    		if (node.getLeft() != null){
    			treeSize += node.getLeft().getTreeSize();
    			height = node.getLeft().getHeight();
    		}
    		if (node.getRight() != null){
    			treeSize += node.getRight().getTreeSize();
    			if (node.getRight().getHeight() > height){
    				height = node.getRight().getHeight();
    			}
    		}
    		node.setTreeSize(treeSize);
    		node.setHeight(height+1);
    	}
    	branchStack = null;
    }

    public void updateBranchOrderDown(){
    	BinaryTreeNode current;
    	Stack<BinaryTreeNode> assembleStack = new Stack<BinaryTreeNode>();
    	assembleStack.push(this);
    	while (!assembleStack.isEmpty()){
    		current = assembleStack.pop();
    		if (current.hasParent()){
    			current.setBranchOrder(current.getParent().getBranchOrder()+1);
    		}else{
    			current.setBranchOrder(0);
    		}
    		if (current.hasLeft()){
    			assembleStack.push(current.getLeft());
    		}
    		if (current.hasRight()){
    			assembleStack.push(current.getRight());
    		}
    	}
    }

    public static void updateStatisticsDown(BinaryTreeNode node){
    	BinaryTreeNode current;
    	Stack<BinaryTreeNode> assembleStack = new Stack<BinaryTreeNode>();
    	Stack<BinaryTreeNode> branchStack = new Stack<BinaryTreeNode>();
    	int treeSize, maxHeight, totalHeight, leaves, leftSize, rightSize, leftDeg, rightDeg;
    	assembleStack.push(node);
    	branchStack.push(node);
    	while (!assembleStack.isEmpty()){
    		current = assembleStack.pop();
    		if (current.hasLeft()){
    			assembleStack.push(current.getLeft());
    			branchStack.push(current.getLeft());
    		}
    		if (current.hasRight()){
    			assembleStack.push(current.getRight());
    			branchStack.push(current.getRight());
    		}
    	}
    	assembleStack = null;
    	while (!branchStack.isEmpty()){
    		current = branchStack.pop();
    		treeSize = 1;
    		leaves = totalHeight = maxHeight = leftSize = rightSize = 0;
    		if (current.getLeft() != null){
    			leftSize = current.getLeft().getTreeSize();
    			treeSize += current.getLeft().getTreeSize();
    			leaves += current.getLeft().getLeaves();
    			maxHeight = current.getLeft().getHeight() + 1;
    			// totalLength is the total length of the children plus 1 for each child below (treeSize)
    			totalHeight += current.getLeft().getTotalHeight()
    				+ current.getLeft().getLeaves();
    		}
    		else if (current.getRight() == null){
    			leaves++;
    		}
    		if (current.getRight() != null){
    			rightSize = current.getRight().getTreeSize();
    			treeSize += current.getRight().getTreeSize();
    			leaves += current.getRight().getLeaves();
    			if (current.getRight().getHeight() >= maxHeight){
    				maxHeight = current.getRight().getHeight() + 1;
    			}
    			totalHeight += current.getRight().getTotalHeight()
    				+ current.getRight().getLeaves();
    		}
    		current.setTreeSize(treeSize);
    		current.setLeaves(leaves);
    		current.setHeight(maxHeight);
    		current.setTotalHeight(totalHeight);
    		rightDeg = rightSize/2 + 1;
    		leftDeg = leftSize/2 + 1;
    		// treeSize-1 so this node is not included
    		current.setAverageHeight((float)totalHeight/leaves);
    		if (treeSize > 3){
    			// Ap defined as Ap(r,m-r) = (m-2r)/(m-2) where m>2, r<=m-r
    			// or, r=min(leftDeg,rightDeg), m=rightDeg+leftDeg, 
    			// m-2r = rightDeg+leftDeg-2*min(leftDeg,rightDeg) = abs(leftDeg=rightDeg)
    			current.setPartitionAsymmetry((double)(Math.abs(rightDeg - leftDeg)) /
    					(rightDeg + leftDeg - 2));
    			if (current.hasLeft() && current.hasRight()){
    				current.setTotalAsymmetry(
    					(double)current.getPartitionAsymmetry()
    					+ current.getRight().getTotalAsymmetry()
    					+ current.getLeft().getTotalAsymmetry());
    			}
    		}
    		else{
    			current.setPartitionAsymmetry(0);
    			current.setTotalAsymmetry(0);
    		}
    	}
    	branchStack = null;
    }

    public boolean hasChildren(){
        if (left == null && right == null){
           return false;
        }
        else{
           return true;
        }
     }
    
    public boolean hasLeft(){
    	//System.out.println("hasLeft: "+left);
    	if (left == null){
    		//System.out.println("FALSE");
    		return false;
    	}
    	else{
    	//	System.out.println("TRUE");
    		return true;
    	}
    }

    public boolean hasRight(){
       if (right == null){
          return false;
       }
       else{
          return true;
       }
    }

    public BinaryTreeNode removeOneChildNodes(){
    	BinaryTreeNode node = this;
    	if (hasLeft() && !hasRight()){
    		node = left;
    		if (node != null){
    			node.removeOneChildNodes();
    		}
    	}
    	else if (hasRight() && !hasLeft()){
    		node = right;
    		if (node != null){
    			node.removeOneChildNodes();
    		}
    	}
    	else{
    		left.removeOneChildNodes();
    		right.removeOneChildNodes();
    	}
    	return node;
    }
    
    public boolean isLeaf(){
    	/*System.out.println("isLeaf... hasLeft: "+hasLeft() + "; hasRight: "+hasRight()+
    			"; isLeaf: "+(!hasLeft() && !hasRight()));*/
    	return !hasLeft() && !hasRight();
    }
    
	public float getAverageHeight() {
		if (metaData != null){
			return metaData.getAverageHeight();
		}
		return 0;
	}
	public void setAverageHeight(float averageBranchHeight) {
		if (metaData == null){
			metaData = new NodeMetaData();
		}
		metaData.setAverageHeight(averageBranchHeight);
	}
	public int getLeaves() {
		if (metaData != null){
			return metaData.getLeaves();
		}
		return 0;
	}
	public void setLeaves(int leaves) {
		if (metaData == null){
			metaData = new NodeMetaData();
		}
		metaData.setLeaves(leaves);
	}
	
	public int getTotalHeight() {
		if (metaData != null){
			return metaData.getTotalHeight();
		}
		return 0;
	}
	public void setTotalHeight(int totalHeight) {
		if (metaData == null){
			metaData = new NodeMetaData();
		}
		metaData.setTotalHeight(totalHeight);
	}
	public double getTreeAsymmetry() {
		if (metaData != null){
			return metaData.getTotalAsymmetry()/((treeSize-1)/2);
		}
		return 0;
	}
	
	public double getTotalAsymmetry() {
		if (metaData != null){
			return metaData.getTotalAsymmetry();
		}
		return 0;
	}
	public void setTotalAsymmetry(double totalAsymmetry) {
		if (metaData == null){
			metaData = new NodeMetaData();
		}
		metaData.setTotalAsymmetry(totalAsymmetry);
	}

	public float getBifurcationAsymmetry() {
		if (treeSize <= 2){
			return 0;
		}
		else if (treeSize == 3){
			return 1;
		}
		return (float)(treeSize - 2 * Math.min(left.getTreeSize(),right.getTreeSize()))
			/ (treeSize - 2);
	}
	
    public int compareTo(BinaryTreeNode node){
    	try{
    		if (treeSize != node.getTreeSize()){
	    		return treeSize - node.getTreeSize();	
	    	}
	    	else {
	    		try{
	    			// Smaller shapeNum is makes the tree "larger" in terms of order
	    			return -getShapeNum().compareTo(node.getShapeNum());
	    		} catch (Exception e){
	    			e.printStackTrace();
	    			return 0;
	    		}
	    	}
    	} catch (ClassCastException cce){
    		//System.out.println("ClassCaseException!");
    		return 1;
    	}
    }

    public static final int ASYMMETRY_WEIGHT_UNIFORM = 0;
    public static final int ASYMMETRY_WEIGHT_DEGREE_ABOVE_3 = 1;
    public static final int ASYMMETRY_WEIGHT_SIZE = 2;
    public static final int ASYMMETRY_WEIGHT_DF = 3;
    public static final int ASYMMETRY_WEIGHT_GLOBAL = 4;
    
    /* Traditional asymmetry calculation */
    public double getTreeAsymmetry(int weightMode) throws Exception{
    	if (!this.hasChildren()){
    		return 0;
    	}
		double totalWeightedAsymmetry = 0;
		double totalWeights = 0, weight;
		int degree;
    	BinaryTreeNode current;
    	Stack<BinaryTreeNode> calcStack = new Stack<BinaryTreeNode>();
    	calcStack.push(this);
    	while (!calcStack.isEmpty()){
    		current = calcStack.pop();
    		if (current.getLeft().hasChildren()){
    			calcStack.push(current.getLeft());
    		}
    		if (current.getRight().hasChildren()){
    			calcStack.push(current.getRight());
    		}
    		degree = (current.getTreeSize() / 2) + 1;
    		switch (weightMode){
    		case ASYMMETRY_WEIGHT_UNIFORM:
    			weight = 1;
    			break;
    		case ASYMMETRY_WEIGHT_DEGREE_ABOVE_3:
    			if (degree > 3){
    				weight = 1;
    			}
    			else{
    				weight = 0;
    			}
    			break;
    		case ASYMMETRY_WEIGHT_SIZE:
    			if (degree > 3){
    				weight = degree - 2;
    			}
    			else{
    				weight = 0;
    			}
    			break;
    		case ASYMMETRY_WEIGHT_DF:
    			if (degree > 3){
    				weight = degree - 3;
    			}
    			else{
    				weight = 0;
    			}
    			break;
    		case ASYMMETRY_WEIGHT_GLOBAL:
   				weight = degree;
    			break;
    		default:
    			weight = 1;
    		}
    		totalWeights += weight;
    		totalWeightedAsymmetry += weight*current.getPartitionAsymmetry();
    	}
		return totalWeightedAsymmetry/totalWeights;
    }
    
    /* Asymmetry calculations for non-nodal measures */
	public double getTreeAsymmetry(int weightMode, int metric, int sizeMode)
	throws Exception{
		if (this.isLeaf()){
			return 0;
		}
		double totalWeightedAsymmetry = 0;
		double totalWeights = 0, weight;
		if (getAsymmetry(metric,sizeMode) == null){
			calculateAsymmetry(metric);
		}
		int degree;
    	BinaryTreeNode current;
    	Stack<BinaryTreeNode> calcStack = new Stack<BinaryTreeNode>();
    	calcStack.push(this);
    	Asymmetry asym;
    	while (!calcStack.isEmpty()){
    		current = calcStack.pop();
    		asym = current.getAsymmetry(metric, sizeMode);
    		if (current.getLeft().hasChildren()){
    			calcStack.push(current.getLeft());
    		}
    		if (current.getRight().hasChildren()){
    			calcStack.push(current.getRight());
    		}
    		degree = (current.getTreeSize() / 2) + 1;
    		switch (weightMode){
    		case ASYMMETRY_WEIGHT_UNIFORM:
    			weight = 1;
    			break;
    		case ASYMMETRY_WEIGHT_DEGREE_ABOVE_3:
    			if (degree > 3){
    				weight = 1;
    			}
    			else{
    				weight = 0;
    			}
    			break;
    		case ASYMMETRY_WEIGHT_DF:
       			if (degree > 3){
       				weight = degree-3;
       			}
    			else{
    				weight = 0;
    			}
    			break;
    		case ASYMMETRY_WEIGHT_SIZE:
       			if (degree > 3){
       				weight = degree-2;
       			}
    			else{
    				weight = 0;
    			}
    			break;
    		case ASYMMETRY_WEIGHT_GLOBAL:
   				weight = asym.getTotalSize();
    			break;
    		default:
    			weight = 1;
    		}
    		totalWeights += weight;
    		totalWeightedAsymmetry += weight*current.getPartitionAsymmetry(metric,sizeMode);
    	}
		
		return totalWeightedAsymmetry/totalWeights;
	}
	
    public static final int LOPSIDEDNESS_BRANCHES = 0;
    public static final int LOPSIDEDNESS_DEGREE = 1;
    public double getLopsidedness() {
    	return getLopsidedness(LOPSIDEDNESS_BRANCHES);
    }
	public double getLopsidedness(int mode) {
		double sumDiff = 0, sumSum = 0;
		int leftDeg, rightDeg, leftBranch, rightBranch;
    	BinaryTreeNode current;
    	Stack<BinaryTreeNode> calcStack = new Stack<BinaryTreeNode>();
    	calcStack.push(this);
    	while (!calcStack.isEmpty()){
    		current = calcStack.pop();
			leftDeg = rightDeg = 0;
			leftBranch = rightBranch = 0;
    		if (current.hasLeft()){
    			calcStack.push(current.getLeft());
    			leftDeg = current.getLeft().getDegree();
    			 //+1 connects tree to curr node
    			leftBranch = current.getLeft().getTreeSize()+1;
    		}
    		if (current.hasRight()){
    			calcStack.push(current.getRight());
    			rightDeg = current.getRight().getDegree();
    			rightBranch = current.getRight().getTreeSize()+1;
    		}
    		
    		switch (mode){
    		case LOPSIDEDNESS_DEGREE:
    			sumDiff += (Math.max(rightDeg, leftDeg) - Math.min(rightDeg, leftDeg));
    			sumSum += rightDeg + leftDeg;
    			break;
    		case LOPSIDEDNESS_BRANCHES:
    			sumDiff += (Math.max(rightBranch, leftBranch) - Math.min(rightBranch, leftBranch));
    			sumSum += rightBranch + leftBranch;
    			break;
    		}
    	}
		
		return sumDiff/sumSum;
	}
	
	private class NodeMetaData{
	    protected int totalHeight;
	    protected float averageBranchHeight;
	    protected int leaves;
	    protected double partitionAsymmetry;
	    protected double totalAsymmetry;
	    protected BigInteger shapeNum;
	    
		public float getAverageHeight() {
			return averageBranchHeight;
		}
		public void setAverageHeight(float averageBranchHeight) {
			this.averageBranchHeight = averageBranchHeight;
		}
		public int getLeaves() {
			return leaves;
		}
		public void setLeaves(int leaves) {
			this.leaves = leaves;
		}
		public int getTotalHeight() {
			return totalHeight;
		}
		public void setTotalHeight(int totalHeight) {
			this.totalHeight = totalHeight;
		}
		
		public double getTotalAsymmetry() {
			return totalAsymmetry;
		}
		public void setTotalAsymmetry(double totalAsymmetry) {
			this.totalAsymmetry = totalAsymmetry;
		}
		/**
		 * 
		 * @return partition asymmetry calculated as tree was built; based on degree, not size
		 */
		public double getPartitionAsymmetry() {
			return partitionAsymmetry;
		}
		public void setPartitionAsymmetry(double partitionAsymmetry) {
			this.partitionAsymmetry = partitionAsymmetry;
		}
		/*
		public float getBifurcationAsymmetry() {
			if (treeSize <= 2){
				return 0;
			}
			else if (treeSize == 3){
				return 1;
			}
			return (float)(treeSize - 2 * Math.min(left.getTreeSize(),right.getTreeSize()))
				/ (treeSize - 2);
		}*/
		public BigInteger getShapeNum() {
			return shapeNum;
		}
		public void setShapeNum(BigInteger shapeNum) {
			this.shapeNum = shapeNum;
		}
		
	}

	public void setPartitionAsymmetry(double partitionAsymmetry) {
		if (metaData == null){
			metaData = new NodeMetaData();
		}
		metaData.setPartitionAsymmetry(partitionAsymmetry);
	}

	public Asymmetry getAsymmetry()throws Exception{
		return getAsymmetry(BinaryTreeUtils.METRIC_NODES, Asymmetry.SIZE_MODE_SUM);
	}
	/**
	 * Traditional partition asymmetry 
	 * @return partition asymmetry calculated as tree was built; based on degree, not size
	 */

	public double getPartitionAsymmetry(){
		if (metaData != null){
			return metaData.getPartitionAsymmetry();
		}
		return 0;
	}

	/** 
	 * 
	 * @param metric
	 * @return Partition asymmetry based on size, not degree
	 * @throws Exception
	 */
	public double getPartitionAsymmetry(int metric)throws Exception{
		return getPartitionAsymmetry(metric, Asymmetry.SIZE_MODE_SUM);
	}
	
	public double getPartitionAsymmetry(int metric, int sizeMode)
	throws Exception{
		Asymmetry asym = null;

		asym = getAsymmetry(metric, sizeMode);
		if (asym == null){
			calculateAsymmetry(metric, sizeMode); 
			asym = getAsymmetry(metric, sizeMode);
		}
		return asym.getPartitionAsymmetry();
	}
	
	private Integer getAsymmetryKey(int metric, int sizeMode){
		int key = 0;
		if (sizeMode == Asymmetry.SIZE_MODE_MAX){
			key = 4;
		}
		switch (metric){
		case BinaryTreeUtils.METRIC_NODES:
			break;
		case BinaryTreeUtils.METRIC_PATH_LENGTH:
			key = key + 1;
			break;
		case BinaryTreeUtils.METRIC_SURFACE_AREA:
			key = key + 2;
			break;
		case BinaryTreeUtils.METRIC_VOLUME:
			key = key + 3;
			break;
		}
		return new Integer(key);
	}
	
	public Asymmetry getAsymmetry(int metric)throws Exception{
		return getAsymmetry(metric,Asymmetry.SIZE_MODE_SUM);
	}
	public Asymmetry getAsymmetry(int metric, int sizeMode) throws Exception{
		Integer key = getAsymmetryKey(metric, sizeMode);
		if (asymmetryMap == null) asymmetryMap = new HashMap<Integer,Asymmetry>();
		if (!asymmetryMap.containsKey(key)){
			this.calculateAsymmetry(metric, sizeMode);
		}
		return asymmetryMap.get(key);
	}

	public void setAsymmetry(int metric, Asymmetry asym){
		setAsymmetry(metric,Asymmetry.SIZE_MODE_SUM,asym);
	}
	public void setAsymmetry(int metric, int sizeMode, Asymmetry asym){
		if (asymmetryMap == null) asymmetryMap = new HashMap<Integer,Asymmetry>();
		asymmetryMap.put(getAsymmetryKey(metric,sizeMode), asym);
	}

	public int getNumberOfBifurcations(){
		return treeSize/2;
	}
	public int getDegree(){
		return treeSize/2 + 1;
	}
	public BinaryTreeNode getLargerSubtree()throws Exception{
		return getLargerSubtree(BinaryTreeUtils.METRIC_NODES);
	}
	public BinaryTreeNode getLargerSubtree(int metric)throws Exception{
		return getLargerSubtree(metric, Asymmetry.SIZE_MODE_SUM);
	}
	public BinaryTreeNode getLargerSubtree(int metric, int sizeMode)throws Exception{
		Asymmetry asym = getAsymmetry(metric, sizeMode);
		return asym == null ? null : (asym.isLargerSubtreeLeft() ? left : right);
	}
	public BinaryTreeNode getSmallerSubtree()throws Exception{
		return getSmallerSubtree(BinaryTreeUtils.METRIC_NODES);
	}
	public BinaryTreeNode getSmallerSubtree(int metric)throws Exception{
		return getSmallerSubtree(metric, Asymmetry.SIZE_MODE_SUM);
	}
	public BinaryTreeNode getSmallerSubtree(int metric, int sizeMode)
	throws Exception{
		Asymmetry asym = getAsymmetry(metric, sizeMode);
		return asym == null ? null : (asym.isLargerSubtreeLeft() ? right : left);
	}
	
	private void setAsymmetryMap(Map<Integer,Asymmetry> asymmetryMap) {
		this.asymmetryMap = asymmetryMap;
	}

	public BinaryTreeNode copy(BinaryTreeNode copy){
		return copy(copy,false);
	}
	public BinaryTreeNode copy(){
		return copy(new BinaryTreeNode(),false);
	}
	
    public BinaryTreeNode copy(boolean keepData){
    	BinaryTreeNode copy = new BinaryTreeNode();
    	copy.setBranchOrder(branchOrder);
    	return copy(copy, keepData);
    }

	public BinaryTreeNode copy(BinaryTreeNode copy, boolean keepData){
		Map<Integer,Asymmetry> asymMap = new HashMap<Integer,Asymmetry>();
		copy.setAsymmetryMap(asymMap);
		Integer key;
		if (asymmetryMap!= null){
			for (Iterator<Integer> it = asymmetryMap.keySet().iterator(); it.hasNext();){
				key = it.next();
				asymMap.put(key,asymmetryMap.get(key).copy());
			}
		}
    	copy.setTreeSize(treeSize);
    	copy.setHeight(height);
    	if (metaData != null){
    		copy.setAverageHeight(metaData.getAverageHeight());
    		copy.setLeaves(metaData.getLeaves());
    		copy.setPartitionAsymmetry(metaData.getPartitionAsymmetry());
    		copy.setTotalAsymmetry(metaData.getTotalAsymmetry());
    	}
    	if (keepData){
    		copy.setData(data);
    	}
    	return copy;
    }

	public void setShapeNum(BigInteger shapeNum){
		if (metaData == null){
			metaData = new NodeMetaData();
		}
		metaData.setShapeNum(shapeNum);
	}
	public BigInteger getShapeNum() throws Exception{
		if (metaData == null){
			metaData = new NodeMetaData();
		}
		if (metaData.getShapeNum() == null){
			if (hasChildren()){
				BinaryTreeNode larger, smaller;
				BigInteger base;
				if (left.getDegree() == right.getDegree()){
					int deg = getLeft().getDegree();
					base = BinaryTreeUtils.partialSumOfNumShapes(getDegree(),deg-1);
					BigInteger shapeNumLeft = left.getShapeNum();
					BigInteger shapeNumRight = right.getShapeNum();
					BigInteger largerShapeNum, smallerShapeNum;
					if (shapeNumLeft.compareTo(shapeNumRight) > 0){
						largerShapeNum = shapeNumLeft;
						smallerShapeNum = shapeNumRight;
					}
					else{
						largerShapeNum = shapeNumRight;
						smallerShapeNum = shapeNumLeft;
					}
					BigInteger two = BigInteger.ONE.add(BigInteger.ONE);
					metaData.setShapeNum(base.add(smallerShapeNum).add(
							largerShapeNum.multiply(largerShapeNum.subtract(BigInteger.ONE))
								.divide(two)));
				}
				else{
					if (left.getDegree() > right.getDegree()){
						larger = left;
						smaller = right;
					}
					else{
						larger = right;
						smaller = left;
					}
					int smallerDeg = smaller.getDegree();
					base = BinaryTreeUtils.partialSumOfNumShapes(getDegree(),smallerDeg-1);
					metaData.setShapeNum(base.add(
							smaller.getShapeNum().add(
							larger.getShapeNum().subtract(BigInteger.ONE).multiply(
									BinaryTreeUtils.numShapes(smallerDeg)))));
				}
			}
			else{
				metaData.setShapeNum(BigInteger.ONE);
			}
		}
		return metaData.getShapeNum();
	}
	
	public void calculateAsymmetry(int metric) throws Exception{
		calculateAsymmetry(metric, Asymmetry.SIZE_MODE_SUM);
	}
	public void calculateAsymmetry(int metric, int sizeMode)
	throws Exception{
		if (metric != BinaryTreeUtils.METRIC_NODES 
				&& !getData().getClass().equals(SwcDataNode.class)){
			throw new Exception("Data member in ParentedBinaryTreeNode must be of type "+SwcDataNode.class+
					". Data member of given in put is type "+getData().getClass());
		}
		// Get all bifurcating nodes from leaf to root
		Stack<BinaryTreeNode> stack = BinaryTreeUtils.getLeavesToRootStack(this, true);

		BinaryTreeNode node, left, right;
		SwcDataNode parentData, leftData, rightData;
		Asymmetry nodeAsym, rightAsym, leftAsym;
		double leftSize, rightSize, longerSummed, leftLargerSummed, rightLargerSummed, 
			allSummed, leftAllSummed, rightAllSummed;

		if (stack.isEmpty()){
			System.out.println("No nodes in stack");
			return;
		}
		// Go through all bifurcation nodes
		while (!stack.isEmpty()){
			node = stack.pop();
			left = node.getLeft();
			right = node.getRight();
			
			// Get sizes based on given metric
			switch (metric){
			case BinaryTreeUtils.METRIC_NODES:
				leftSize = rightSize = 1;
				break;
			case BinaryTreeUtils.METRIC_EUCLIDEAN_DISTANCE:
				parentData = (SwcDataNode)node.getData();
				leftData = (SwcDataNode)left.getData();
				rightData = (SwcDataNode)right.getData();
				leftSize = SwcDataUtils.getDistance(parentData, leftData);
				rightSize = SwcDataUtils.getDistance(parentData, rightData);
				break;
			// If using branch order distance, "length" is one
			case BinaryTreeUtils.METRIC_PATH_LENGTH:
				parentData = (SwcDataNode)node.getData();
				leftData = (SwcDataNode)left.getData();
				rightData = (SwcDataNode)right.getData();
				leftSize = leftData.getSecondaryData().getPathLength();
				rightSize = rightData.getSecondaryData().getPathLength();
				break;
			case BinaryTreeUtils.METRIC_SURFACE_AREA:
				parentData = (SwcDataNode)node.getData();
				leftData = (SwcDataNode)left.getData();
				rightData = (SwcDataNode)right.getData();
				leftSize = leftData.getSecondaryData().getSurfaceArea();
				rightSize = rightData.getSecondaryData().getSurfaceArea();
				break;
			case BinaryTreeUtils.METRIC_VOLUME:
				parentData = (SwcDataNode)node.getData();
				leftData = (SwcDataNode)left.getData();
				rightData = (SwcDataNode)right.getData();
				leftSize = leftData.getSecondaryData().getVolume();
				rightSize = rightData.getSecondaryData().getVolume();
				break;
			default:
				throw new Exception("Unknown distance measure mode "+metric);
			}
			longerSummed = rightLargerSummed = leftLargerSummed = 0;
			allSummed = rightAllSummed = leftAllSummed = 0;

			// Add to total size of each sub-tree;
			if (left.hasChildren()){
				leftAsym = left.getAsymmetry(metric, sizeMode);
				if (sizeMode == Asymmetry.SIZE_MODE_SUM){
					// total size will be the longer + shorter lengths of each of left's sub-tree
					leftSize += 
						leftAsym.getLargerSize() + leftAsym.getSmallerSize();
				}
				else{
					// Max size of the left subtree will be the max size of this child + the branch from this node to its child
					leftSize += leftAsym.getLargerSize();
				}
				// keeping track of the sum of larger tree lengths for running caulescence calc.
				leftLargerSummed = leftAsym.getLargerSummedSize();
				// keeping track of the sum of all tree lengths for running caulescence calc.
				leftAllSummed = leftAsym.getAllSummedSize();
			}
			if (right.hasChildren()){
				rightAsym = right.getAsymmetry(metric, sizeMode);
				if (sizeMode == Asymmetry.SIZE_MODE_SUM){
					rightSize += 
						rightAsym.getLargerSize() + rightAsym.getSmallerSize();
				}
				else{
					rightSize += rightAsym.getLargerSize();
				}
				rightLargerSummed = rightAsym.getLargerSummedSize();
				rightAllSummed = rightAsym.getAllSummedSize();
			}
			
			// Create new Asymmetry object
			nodeAsym = new Asymmetry();
			node.setAsymmetry(metric, sizeMode, nodeAsym);
			// Add current node's total length to the summation
			allSummed += leftSize + rightSize;
			// determine larger/longer sub-tree
			if (leftSize > rightSize){
				nodeAsym.setLargerSize(leftSize);
				nodeAsym.setSmallerSize(rightSize);
				longerSummed = leftSize + leftLargerSummed;
				allSummed += leftAllSummed;
				nodeAsym.setMainPath(Asymmetry.MAIN_BRANCH_LEFT);
				nodeAsym.setLargerSubtree(Asymmetry.MAIN_BRANCH_LEFT);
			}
			else{
				nodeAsym.setLargerSize(rightSize);
				nodeAsym.setSmallerSize(leftSize);
				longerSummed = rightSize + rightLargerSummed;
				allSummed += rightAllSummed;
				nodeAsym.setMainPath(Asymmetry.MAIN_BRANCH_RIGHT);
				nodeAsym.setLargerSubtree(Asymmetry.MAIN_BRANCH_RIGHT);
			}
			nodeAsym.setLargerSummedSize(longerSummed);
			nodeAsym.setAllSummedSize(allSummed);
			
		}
	}

    public static void updateBranchStatisticsUp(BinaryTreeNode node){
    	BinaryTreeNode current = node;
    	int treeSize, maxHeight, totalHeight, leaves, leftSize, rightSize;
    	while (current != null){
    		treeSize = 1;
    		leaves = totalHeight = maxHeight = leftSize = rightSize = 0;
    		if (current.hasLeft()){
    			leftSize = current.getLeft().getTreeSize();
    			treeSize += current.getLeft().getTreeSize();
    			leaves += current.getLeft().getLeaves();
    			maxHeight = current.getLeft().getHeight() + 1;
    			// Adding total height of right subtree, plus 1 for each existing terminal path
    			totalHeight += current.getLeft().getTotalHeight()
    				+ current.getLeft().getLeaves();
    		}
    		else if (current.getRight() == null){
    			leaves++;
    		}
    		if (current.hasRight()){
    			rightSize = current.getRight().getTreeSize();
    			treeSize += current.getRight().getTreeSize();
    			leaves += current.getRight().getLeaves();
    			if (current.getRight().getHeight() >= maxHeight){
    				maxHeight = current.getRight().getHeight() + 1;
    			}
    			// Adding total height of right subtree, plus 1 for each existing terminal path
    			totalHeight += current.getRight().getTotalHeight()
    				+ current.getRight().getLeaves();
    		}
    		current.setTreeSize(treeSize);
    		current.setLeaves(leaves);
    		current.setHeight(maxHeight);
    		current.setTotalHeight(totalHeight);
    		// treeSize-1 so this node is not included
    		current.setAverageHeight((float)totalHeight/leaves);
    		if (treeSize > 3){
    			// Ap defined as Ap(r,m-r) = (m-2r)/(m-2) where m>2, r<=m-r
    			// or, r=min(leftDeg,rightDeg), m=rightDeg+leftDeg, 
    			// m-2r = rightDeg+leftDeg-2*min(leftDeg,rightDeg) = abs(leftDeg=rightDeg)
    			int rightDeg = (rightSize+1)/2;
    			int leftDeg = (leftSize+1)/2;
    			current.setPartitionAsymmetry((double)(Math.abs(rightDeg - leftDeg)) /
    					(rightDeg + leftDeg - 2));
/*
    			current.setPartitionAsymmetry((float)(rightSize + leftSize - 2 * Math.min(leftSize, rightSize)) /
    					(rightSize + leftSize - 2));*/
    			if (current.hasLeft() && current.hasRight()){
    				current.setTotalAsymmetry(
    					(double)current.getPartitionAsymmetry()
    					+ current.getRight().getTotalAsymmetry()
    					+ current.getLeft().getTotalAsymmetry());
    			}
    		}
    		else{
    			current.setPartitionAsymmetry(0);
    			current.setTotalAsymmetry(0);
    		}

    		current = current.getParent();
    	}
    }

    public boolean addLeft(BinaryTreeNode l){
       if (hasLeft()){
          return false;
       }
       else{
          left = l;
          updateBranchStatisticsUp(this);
          return true;
       }
    }

    public boolean addRight(BinaryTreeNode r){
       if (hasRight()){
          return false;
       }
       else{
          right = r;
          updateBranchStatisticsUp(this);
          return true;
       }
    }

    public void detach(){
    	detach(true);
    }

    public void detach(boolean updateParentStats){
    	BinaryTreeNode p = parent;
        if (p != null){
 	       if (side == RIGHT){
 	          p.setRight(null,false);
 	       }
 	       else{
 	          p.setLeft(null,false);
 	       }
 	       if (updateParentStats){
 	    	   updateBranchStatisticsUp(p);
 	       }
        }
        parent = null;
     }
    
    public boolean isLeft(){
    	return side == LEFT;
    }
    
    public boolean isRight(){
    	return side == RIGHT;
    }

	public boolean hasParent(){
		return parent != null;
	}
	
    public boolean isRoot(){
    	return parent == null;
    }
 
    public boolean hasBothChildren(){
    	return left != null && right != null;
    }

	/**
	 * 
	 * @param root
	 * @param threshold
	 * @return Puts any terminal branches that are within the spur parameter setting into a list of spur nodes that
	 * will not count against the score. 
	 * 
	 */
	//public BinaryTreeNode removeSpurs(double threshold){
    public void removeSpurs(double threshold){
		BinaryTreeNode sibling, gp;
		//System.out.println("Removing spurs");
		boolean parentIsLeft;
		BinaryTreeNode root = this;
		double distance;
		
		List<BinaryTreeNode> termList = BinaryTreeUtils.createTerminalNodeList(root);
		Set<BinaryTreeNode> removedSet = new HashSet<BinaryTreeNode>();

		for (BinaryTreeNode node : termList){
			// Node is termination: determine if spur
			distance = node.getSwcData().getSecondaryData().getPathLength();
			if (node.isRight()){
				// Node is right, pop left (would be next) off of stack
				//stack.pop();
				// Get sibling
				sibling = node.getParent().getLeft();
			}
			else{
				// Node is left
				sibling = node.getParent().getRight();
			}

			if (distance < threshold & !removedSet.contains(sibling)){
				// Node is a spur, remove and remove parent bifurcation
				removedSet.add(node);

				//System.out.println("Removing node: "+node.getSwcData().positionString());

				// Set path length of sibling using parent's path length
				sibling.getSwcData().getSecondaryData().setPathLength(
						sibling.getSwcData().getSecondaryData().getPathLength()+
						sibling.getParent().getSwcData().getSecondaryData().getPathLength());

				// Detach parent and attach node's sibling in its place
				gp = sibling.getParent().getParent();
				if (gp == null){ // Replace children of root (aka this)
					root.setData(sibling.getSwcData());
					root.setLeft(sibling.getLeft());
					root.setRight(sibling.getRight());
				}
				else{
					parentIsLeft = sibling.getParent().isLeft();
					node.getParent().detach(false);
					if (parentIsLeft){
						gp.setLeft(sibling);
					}else{
						gp.setRight(sibling);
					}
				}
			}
		}
		//return root;
	}
	
    public BinaryTreeNode getMaxOrderBranch(){
    	if (this.isLeaf()){
    		return this;
    	}
    	BinaryTreeNode leftMax = left.getMaxOrderBranch();
    	BinaryTreeNode rightMax = right.getMaxOrderBranch();
    	if (leftMax.getBranchOrder() > rightMax.getBranchOrder()){
    		return leftMax;
    	}
    	return rightMax;
    }
 
    public static void main(String args[]){
    	BinaryTreeNode node = new BinaryTreeNode();
    	node.setLeft(new BinaryTreeNode());
    	node.setRight(new BinaryTreeNode());
    	node.getRight().setRight(new BinaryTreeNode());
    	node.getRight().setLeft(new BinaryTreeNode());
    	node.getRight().getRight().setRight(new BinaryTreeNode());
    	node.getRight().getRight().setLeft(new BinaryTreeNode());
    	System.out.println(node.getPartitionAsymmetry());
    	node.getLeft().setRight(new BinaryTreeNode());
    	node.getLeft().setLeft(new BinaryTreeNode());
    	System.out.println(node.getPartitionAsymmetry());
    }
    
}