package org.cn3.morpho.scripts;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.text.NumberFormat;
import java.util.Stack;
import java.util.regex.Matcher;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.SWCReader;
import org.cn3.morpho.domain.SwcDataNode;
import org.cn3.morpho.domain.SwcTreeNode;
import org.cn3.morpho.utils.SwcDataUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Rotates the swc file given angles.
 * 
 * java org.cn3.morpho.scripts.SwcNeuronTransformer -s "C:/Documents and Settings/gillette/My Documents/My Dropbox/SOFALab/Virtual Model/SWC Files/.swc" -o "C:/Documents and Settings/gillette/My Documents/My Dropbox/SOFALab/Virtual Model/SWC Files/transformed/.swc" --transformationFile "C:/Documents and Settings/gillette/My Documents/My Dropbox/SOFALab/Virtual Model/Transformations/.txt" 
 * 
 * java org.cn3.morpho.scripts.SwcNeuronTransformer -s "C:/Documents and Settings/gillette/My Documents/NMO/ca3pv/l56b.CNG.swc" -o "C:/Documents and Settings/gillette/My Documents/NMO/ca3pv/l56b_rotated.swc" --transformationFile "C:/Documents and Settings/gillette/My Documents/NMO/ca3pv/l56b_rotation.csv"
 * 
 */
public class SwcNeuronTransformer {

	public static double[][] multiplyMatrix(double[][] matrix1, double[][] matrix2)
	throws Exception{
		if (matrix1[0].length != matrix2.length){
			throw new Exception("Matrix 1 must have the same number of columns as matrix 2 has rows");
		}
		double[][] resultMat = new double[matrix1.length][matrix2[0].length];
		
		for (int i = 0; i < matrix1.length; i++){
			for (int j = 0; j < matrix2[0].length; j++){
				resultMat[i][j] = 0;
				for (int k = 0; k < matrix1[0].length; k++){
					resultMat[i][j] += matrix1[i][k]*matrix2[k][j];
				}
			}
		}
		
		return resultMat;
		
	}
	
	public static double[][] getIdentityMatrix(){
		double[][] identityMatrix = new double[4][4];
		identityMatrix[0][0] = identityMatrix[1][1] = identityMatrix[2][2] = identityMatrix[3][3] = 1;
		identityMatrix[0][1] = identityMatrix[0][2] = identityMatrix[0][3] = 0;
		identityMatrix[1][0] = identityMatrix[1][2] = identityMatrix[1][3] = 0;
		identityMatrix[2][0] = identityMatrix[2][1] = identityMatrix[2][3] = 0;
		identityMatrix[3][0] = identityMatrix[3][1] = identityMatrix[3][2] = 0;
		return identityMatrix;
	}
	public static double[][] getScaleMatrix(double scale){
		return getScaleMatrix(scale,scale,scale);
	}
	public static double[][] getScaleMatrix(SwcDataNode scale){
		return getScaleMatrix(scale.getX(),scale.getY(),scale.getZ());
	}
	public static double[][] getScaleMatrix(double scaleX, double scaleY, double scaleZ){
		double[][] scaleMatrix = new double[4][4];
		
		scaleMatrix[0][0] = scaleX;
		scaleMatrix[0][1] = 0;
		scaleMatrix[0][2] = 0;
		scaleMatrix[0][3] = 0; 
		
		scaleMatrix[1][0] = 0;
		scaleMatrix[1][1] = scaleY;
		scaleMatrix[1][2] = 0;
		scaleMatrix[1][3] = 0;
		
		scaleMatrix[2][0] = 0;
		scaleMatrix[2][1] = 0;
		scaleMatrix[2][2] = scaleZ;
		scaleMatrix[2][3] = 0;
		
		scaleMatrix[3][0] = 0;
		scaleMatrix[3][1] = 0;
		scaleMatrix[3][2] = 0;
		scaleMatrix[3][3] = 1;
		
		return scaleMatrix;
	}
	public static double[][] getTranslationMatrix(SwcDataNode translate){
		return getTranslationMatrix(translate.getX(),translate.getY(),translate.getZ());
	}
	public static double[][] getTranslationMatrix(double translateX, double translateY, double translateZ){
		double[][] translationMatrix = new double[4][4];
		translationMatrix[0][0] = 1;
		translationMatrix[0][1] = 0;
		translationMatrix[0][2] = 0;
		translationMatrix[0][3] = translateX; 
		
		translationMatrix[1][0] = 0;
		translationMatrix[1][1] = 1;
		translationMatrix[1][2] = 0;
		translationMatrix[1][3] = translateY;
		
		translationMatrix[2][0] = 0;
		translationMatrix[2][1] = 0;
		translationMatrix[2][2] = 1;
		translationMatrix[2][3] = translateZ;
		
		translationMatrix[3][0] = 0;
		translationMatrix[3][1] = 0;
		translationMatrix[3][2] = 0;
		translationMatrix[3][3] = 1;
		
		return translationMatrix;
	}

	public static double[][] calculateRotationMatrix(SwcDataNode axis, double angle){
		double[][] rotationMatrix = new double[4][4];
		double cosTheta = Math.cos(angle);
		double sinTheta = Math.sin(angle);
		
		// Normalize the axis
		SwcDataUtils.normalizeVector(axis);

		rotationMatrix[0][0] = cosTheta + axis.getX()*axis.getX()*(1-cosTheta);
		rotationMatrix[0][1] = axis.getX()*axis.getY()*(1-cosTheta) - axis.getZ()*sinTheta;
		rotationMatrix[0][2] = axis.getX()*axis.getZ()*(1-cosTheta) + axis.getY()*sinTheta;
		rotationMatrix[0][3] = 0; 
		
		rotationMatrix[1][0] = axis.getX()*axis.getY()*(1-cosTheta) + axis.getZ()*sinTheta;
		rotationMatrix[1][1] = cosTheta + axis.getY()*axis.getY()*(1-cosTheta);
		rotationMatrix[1][2] = axis.getY()*axis.getZ()*(1-cosTheta) - axis.getX()*sinTheta;
		rotationMatrix[1][3] = 0;
		
		rotationMatrix[2][0] = axis.getX()*axis.getZ()*(1-cosTheta) - axis.getY()*sinTheta;
		rotationMatrix[2][1] = axis.getY()*axis.getZ()*(1-cosTheta) + axis.getX()*sinTheta;
		rotationMatrix[2][2] = cosTheta + axis.getZ()*axis.getZ()*(1-cosTheta);
		rotationMatrix[2][3] = 0;
		
		rotationMatrix[3][0] = 0;
		rotationMatrix[3][1] = 0;
		rotationMatrix[3][2] = 0;
		rotationMatrix[3][3] = 1;
		
		return rotationMatrix;
	}
	public static void rotateAndWrite(File swcFile, File outputFile, SwcDataNode axis, double angle) throws Exception{
		double[][] rotationMatrix = calculateRotationMatrix(axis,angle);
		transformAndWrite(swcFile, outputFile, rotationMatrix);
	}
	public static void rotateAndWrite(File swcFile, File outputFile, SwcDataNode axis, double angle, SwcDataNode translation) throws Exception{
		double[][] rotationMatrix = calculateRotationMatrix(axis,angle);
		double[][] translationMatrix = getTranslationMatrix(translation);
		double[][] transformMatrix = multiplyMatrix(rotationMatrix, translationMatrix);
		transformAndWrite(swcFile, outputFile, transformMatrix);
	}
	public static void transformAndWrite(File swcFile, File outputFile, double[][] transformMatrix) throws Exception{
		transformAndWrite(swcFile,outputFile,transformMatrix,1);
	}
	public static void transformAndWrite(File swcFile, File outputFile, double[][] transformMatrix, double radiusScale) throws Exception{
		BufferedReader reader = new BufferedReader(new FileReader(swcFile));

		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(2);
		nf.setGroupingUsed(false);
		
		String dataString = "";
		SwcTreeNode node;
		Matcher m;
		int type, nodeId, parentId;
		double tmpX, tmpY, tmpZ;
		
		StringBuffer sb = new StringBuffer();
		
		// 1. Assembles axon, basal trees, and apical trees (non-binary).
		while ( (dataString=reader.readLine()) != null ) { 
			if (dataString.startsWith("#")){
				sb.append(dataString+"\n");
			}
			else{
				m = SWCReader.swcDataPattern.matcher(dataString);
				//System.out.println(dataString);
				
		        if (m.matches()){
		        	node = new SwcTreeNode();
		        	nodeId = Integer.parseInt(m.group(1));
		        	node.setNodeId(nodeId);
		        	type = Integer.parseInt(m.group(2));
		        	node.setType(type);
		        	node.setX(Double.parseDouble(m.group(3)));
		        	node.setY(Double.parseDouble(m.group(4)));
		        	node.setZ(Double.parseDouble(m.group(5)));
		        	node.setRadius(radiusScale*Double.parseDouble(m.group(6)));
		        	parentId = Integer.parseInt(m.group(7));
		        	node.setParentId(parentId);
		        	
	        		tmpX = transformMatrix[0][0]*node.getX() +
		        		transformMatrix[0][1]*node.getY() +
		        		transformMatrix[0][2]*node.getZ();
	        		tmpY = transformMatrix[1][0]*node.getX() +
	        			transformMatrix[1][1]*node.getY() +
	        			transformMatrix[1][2]*node.getZ();

	    			if (transformMatrix.length > 2){
	    				tmpZ = transformMatrix[2][0]*node.getX() +
	    					transformMatrix[2][1]*node.getY() +
	    					transformMatrix[2][2]*node.getZ();
	    				
	    				// Case in which there is also translation
	    				if (transformMatrix.length == 4){
	    					tmpX += transformMatrix[0][3];
	    					tmpY += transformMatrix[1][3];
	    					tmpZ += transformMatrix[2][3];
	    				}
	    				node.setZ(tmpZ);
	    			}

	        		
	        		node.setX(tmpX);
	        		node.setY(tmpY);
		
					/*sb.append(node.getNodeId()+" "+node.getType() 
							+ " "+nf.format(node.getX())+" "+nf.format(node.getY())+" "+nf.format(node.getZ())
							+ " "+node.getRadius()+" "+node.getParentId()+"\n");
							*/
	        		sb.append(node.getNodeId()+" "+node.getType() 
							+ " "+node.getX()+" "+node.getY()+" "+node.getZ()
							+ " "+node.getRadius()+" "+node.getParentId()+"\n");
		        }
		        else{
		        	System.out.println(dataString);
		        }
	        }
		}
		DataWriter.writeStringToFile(outputFile, sb.toString());
	}
	
	public static void rotateNode(SwcDataNode vector, SwcDataNode axis, double angle)
	throws Exception{
		double[][] rotationMatrix = calculateRotationMatrix(axis,angle);
		transform(vector, rotationMatrix);
	}
	public static void rotate(SwcTreeNode root, SwcDataNode axis, double angle)
	throws Exception{
		double[][] rotationMatrix = calculateRotationMatrix(axis,angle);
		transform(root, rotationMatrix);
	}
	public static void transform(SwcDataNode data, double[][] transformMatrix)
	throws Exception{
		double tmpX, tmpY, tmpZ;

		tmpX = transformMatrix[0][0]*data.getX() +
		transformMatrix[0][1]*data.getY() +
		transformMatrix[0][2]*data.getZ();
	
		tmpY = transformMatrix[1][0]*data.getX() +
			transformMatrix[1][1]*data.getY() +
			transformMatrix[1][2]*data.getZ();
	
		// Check matrix size to ensure Z is being transformed as well
		if (transformMatrix.length > 2){
			tmpZ = transformMatrix[2][0]*data.getX() +
				transformMatrix[2][1]*data.getY() +
				transformMatrix[2][2]*data.getZ();
			
			// Case in which there is also translation
			if (transformMatrix.length == 4){
				tmpX += transformMatrix[0][3];
				tmpY += transformMatrix[1][3];
				tmpZ += transformMatrix[2][3];
			}
			data.setZ(tmpZ);
		}
		
		data.setX(tmpX);
		data.setY(tmpY);
		
	}
	public static void transform(SwcTreeNode root, double[][] transformMatrix)
	throws Exception{
		Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
		stack.push(root);
		SwcTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();

			transform(node.getData(), transformMatrix);

    		if (node.hasChildren()){
				for (int i = 0; i < node.getChildren().size(); i++){
					stack.push((SwcTreeNode)node.getChildren().get(i));
				}
			}
		}
	}

	public static void transformNode(SwcTreeNode node, double[][] transformMatrix)
	throws Exception{
		double tmpX, tmpY, tmpZ;

		tmpX = transformMatrix[0][0]*node.getX() +
			transformMatrix[0][1]*node.getY() +
			transformMatrix[0][2]*node.getZ();
		
		tmpY = transformMatrix[1][0]*node.getX() +
			transformMatrix[1][1]*node.getY() +
			transformMatrix[1][2]*node.getZ();

		// Check matrix size to ensure Z is being transformed as well
		if (transformMatrix.length > 2){
			tmpZ = transformMatrix[2][0]*node.getX() +
				transformMatrix[2][1]*node.getY() +
				transformMatrix[2][2]*node.getZ();
			
			// Case in which there is also translation
			if (transformMatrix.length == 4){
				tmpX += transformMatrix[0][3];
				tmpY += transformMatrix[1][3];
				tmpZ += transformMatrix[2][3];
			}
			node.setZ(tmpZ);
		}
		
		node.setX(tmpX);
		node.setY(tmpY);

	}

	public static void translate(SwcTreeNode root, SwcTreeNode translation){
		translate(root, translation.getX(), translation.getY(), translation.getZ());
	}

	public static void translate(SwcTreeNode root, double translateX, double translateY, double translateZ){
		Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
		stack.push(root);
		SwcTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
    		
			node.setX(node.getX()+translateX);
			node.setY(node.getY()+translateY);
			node.setZ(node.getZ()+translateZ);

    		if (node.hasChildren()){
				for (int i = 0; i < node.getChildren().size(); i++){
					stack.push((SwcTreeNode)node.getChildren().get(i));
				}
			}
		}
	}

	private static final String PARAM_TRANS_FILE = "transformationFile";
	
	private static double[][] generateTransformMatrix(File transformFile) throws Exception{
		BufferedReader reader = new BufferedReader(new FileReader(transformFile));

		double[][] transformMatrix = getIdentityMatrix(), currentTransform = null;
		
		String dataString = "";
		String[] parts;
		SwcDataNode xyz = new SwcDataNode();
		double angle;
		boolean validLine;
/*
		System.out.println("Transformation Matrix:");
		System.out.println(transformMatrix[0][0]+"   "+transformMatrix[0][1]+"   "+transformMatrix[0][2]+"    "+transformMatrix[0][3]);
		System.out.println(transformMatrix[1][0]+"   "+transformMatrix[1][1]+"   "+transformMatrix[1][2]+"    "+transformMatrix[1][3]);
		System.out.println(transformMatrix[2][0]+"   "+transformMatrix[2][1]+"   "+transformMatrix[2][2]+"    "+transformMatrix[2][3]);
		System.out.println(transformMatrix[3][0]+"   "+transformMatrix[3][1]+"   "+transformMatrix[3][2]+"    "+transformMatrix[3][3]);
*/

		// 1. Assembles axon, basal trees, and apical trees (non-binary).
		while ( (dataString=reader.readLine()) != null ) {
			parts = dataString.split(",");
			
			if (!dataString.startsWith("#") && parts.length > 1){
				
		        xyz.setX(Double.parseDouble(parts[1]));
		        xyz.setY(Double.parseDouble(parts[2]));
		        xyz.setZ(Double.parseDouble(parts[3]));
		        
		        validLine = true;
		        if (parts[0].equalsIgnoreCase("scale")){
		        	currentTransform = getScaleMatrix(xyz);
		        }
		        else if (parts[0].equalsIgnoreCase("translate")){
		        	currentTransform = getTranslationMatrix(xyz);
		        }
		        else if (parts[0].equalsIgnoreCase("rotate")){
		        	angle = Double.parseDouble(parts[4]);
		        	currentTransform = calculateRotationMatrix(xyz, angle);
		        }
		        else{
		        	validLine = false;
		        }
	
		        if (validLine){
					System.out.println("Current transform:");
					System.out.println(currentTransform[0][0]+"   "+currentTransform[0][1]+"   "+currentTransform[0][2]+"    "+currentTransform[0][3]);
					System.out.println(currentTransform[1][0]+"   "+currentTransform[1][1]+"   "+currentTransform[1][2]+"    "+currentTransform[1][3]);
					System.out.println(currentTransform[2][0]+"   "+currentTransform[2][1]+"   "+currentTransform[2][2]+"    "+currentTransform[2][3]);
					System.out.println(currentTransform[3][0]+"   "+currentTransform[3][1]+"   "+currentTransform[3][2]+"    "+currentTransform[3][3]);
					transformMatrix = 
						multiplyMatrix(currentTransform, transformMatrix);

					System.out.println("Current matrix:");
					System.out.println(transformMatrix[0][0]+"   "+transformMatrix[0][1]+"   "+transformMatrix[0][2]+"    "+transformMatrix[0][3]);
					System.out.println(transformMatrix[1][0]+"   "+transformMatrix[1][1]+"   "+transformMatrix[1][2]+"    "+transformMatrix[1][3]);
					System.out.println(transformMatrix[2][0]+"   "+transformMatrix[2][1]+"   "+transformMatrix[2][2]+"    "+transformMatrix[2][3]);
					System.out.println(transformMatrix[3][0]+"   "+transformMatrix[3][1]+"   "+transformMatrix[3][2]+"    "+transformMatrix[3][3]);

		        }
			}
		}
		return transformMatrix;
	}

	private static double getCombinedScale(File transformFile) throws Exception{
		BufferedReader reader = new BufferedReader(new FileReader(transformFile));

		double scale = 1;
		
		String dataString = "";
		String[] parts;

		// 1. Assembles axon, basal trees, and apical trees (non-binary).
		while ( (dataString=reader.readLine()) != null ) {
			parts = dataString.split(",");
			
			if (!dataString.startsWith("#") && parts.length > 1 
					&& parts[0].equalsIgnoreCase("scale")){
	        	scale *= Double.parseDouble(parts[1]);
	        }
		}
		return scale;
	}
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			String swcFilename = config.getString("swc");
			String transFilename = config.getString(PARAM_TRANS_FILE);
			String outFilename = config.getString("output");
			
			// Read transformations file and generate transformation matrix
			double[][] transformMatrix = generateTransformMatrix(new File(transFilename));
			double radiusScale = getCombinedScale(new File(transFilename));
			System.out.println("Transformation Matrix:");
			System.out.println(transformMatrix[0][0]+"   "+transformMatrix[0][1]+"   "+transformMatrix[0][2]+"    "+transformMatrix[0][3]);
			System.out.println(transformMatrix[1][0]+"   "+transformMatrix[1][1]+"   "+transformMatrix[1][2]+"    "+transformMatrix[1][3]);
			System.out.println(transformMatrix[2][0]+"   "+transformMatrix[2][1]+"   "+transformMatrix[2][2]+"    "+transformMatrix[2][3]);
			System.out.println(transformMatrix[3][0]+"   "+transformMatrix[3][1]+"   "+transformMatrix[3][2]+"    "+transformMatrix[3][3]);
			
			File swcFile = new File(swcFilename), outputFile = new File(outFilename);
			if (swcFile.isDirectory()){
				if (!outputFile.isDirectory()){
					outputFile = new File(swcFile.getAbsolutePath()+"/Transformed/");
					if (!outputFile.exists()){
						outputFile.mkdir();
					}
				}
				File[] files = swcFile.listFiles();
				for (int i = 0; i < files.length; i++){
					swcFile = files[i];
					if (!swcFile.isDirectory()&&
							(swcFile.getName().endsWith(".swc")||swcFile.getName().endsWith(".SWC"))){
						transformAndWrite(swcFile, 
								new File(outputFile.getAbsolutePath()+"/"+swcFile.getName()), 
								transformMatrix, radiusScale);
					}
				}
			}
			else{
				transformAndWrite(swcFile, outputFile, transformMatrix, radiusScale);
			}
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption("swc")
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag('s') 
        .setLongFlag("swc");
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_TRANS_FILE)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_TRANS_FILE);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption("output")
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag('o') 
        .setLongFlag("output");
	    jsap.registerParameter(opt);


        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + SwcNeuronTransformer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

	
}