package org.cn3.ns.analysis;

import java.text.NumberFormat;
import java.util.List;
import java.util.Map;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.motif.analysis.MotifFinder;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Generates raw n-mer (defined by input) frequencies for all sequences in a given dataset.
 * Earlier attempts showed low if any classification capacity of n-grams, however variations
 * of the process could turn out to be successful, and other uses may be found.
 * 
 * -d dataset
 * -m mapping mode
 * -o traversal order
 * -O output file
 * -N nmer sequence length
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.analysis.CorpusNGramGenerator -O /homes/tgillett/results/
 *
 */

public class CorpusNGramGenerator {

	private static final String PARAM_N = "ngram";
	private static final String PARAM_OUTPUT = "outputDir";
	
	private static final int DEFAULT_N = 4;

	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			String dataset = config.getString(NSConstants.PARAM_DATASET);
			int mapping = config.getInt(NSConstants.PARAM_MAPPING_MODE);
			String order = config.getString(NSConstants.PARAM_ORDER);
			String outputDir = config.getString(PARAM_OUTPUT);
			int n = config.getInt(PARAM_N);
			
			String filename = DataProperties.getInstance().getFastaFilename(dataset, mapping, order);

			// Get all ngrams
			StringBuffer sb = new StringBuffer();
			List<String> ngrams = NeuriteSequenceStringUtils.generatePossibleACTKmers(n, order);
			sb.append("Sequence");
			for (int i = 0; i < ngrams.size(); i++){
				sb.append(","+ngrams.get(i));
			}
			sb.append("\n");
			
			Map<String,Number> nmerFreqs;
			FastaReader fr = new FastaReader(filename);
			NumberFormat nf = NumberFormat.getInstance();
			nf.setMaximumFractionDigits(3);
			while (fr.next()){
				sb.append(fr.getSequenceName());
				nmerFreqs = MotifFinder.findMotifFrequencies(fr.getSequence(),n,n);
				for (int i = 0; i < ngrams.size(); i++){
					sb.append(","+nf.format(nmerFreqs.get(ngrams.get(i))));
				}
				sb.append("\n");
			}
			fr.close();
			
			if (outputDir.equals("")){
				outputDir = System.getProperty("user.dir");
			}
			String outputFile = outputDir+"/NGramFreqs_"+dataset+"_M"+mapping+"_"+order+"_N"+n+""+".csv";
			DataWriter.writeStringToFile(outputFile, sb.toString());
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;

        opt = new FlaggedOption(NSConstants.PARAM_DATASET)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(false)
		.setDefault(NSConstants.DEFAULT_DATASET)
		.setShortFlag('d') 
		.setLongFlag(NSConstants.PARAM_DATASET);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_MAPPING_MODE)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault(""+NSConstants.DEFAULT_MAPPING_MODE)
        .setShortFlag('m') 
        .setLongFlag(NSConstants.PARAM_MAPPING_MODE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(NSConstants.DEFAULT_ORDER)
        .setShortFlag('o') 
        .setLongFlag(NSConstants.PARAM_ORDER);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_OUTPUT)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("")
        .setShortFlag('O') 
        .setLongFlag(PARAM_OUTPUT);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_N)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault(""+DEFAULT_N)
        .setShortFlag('N') 
        .setLongFlag(PARAM_N);
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CorpusNGramGenerator.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
}
