package org.cn3.ns.analysis;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.MetaDataManager;
import org.cn3.morpho.domain.DistributionBean;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.morpho.domain.XYBean;
import org.cn3.morpho.utils.ClassificationUtils;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.ProjectProperties;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Node type (A or C) profile along branch order
 * 	Averaged over tree set OR summed over tree set
 *
 */
public class NodeTypeBranchOrderAnalysis {

	private static final String PARAM_PROFILE_MODE = "profile-mode";
	
	// Combines all instances of nodes at a given branch order, then gets %C
	private static final int PROFILE_MODE_SUM = 1;
	// For a given branch order, a percent of total branch order is taken
	// That % branch order and then the %C for the sequence is a data point
	private static final int PROFILE_MODE_PROPORTION = 2;
	
	private static final int MAPPING = 2;
	
	public static void main(String[] args){
		try{
			JSAPResult config = readParameters(args);
			int profileMode = config.getInt(PARAM_PROFILE_MODE);
			List<String> arborizations = NSDataReader.loadSequenceList(MAPPING);
			Map<String,NeuronMetadata> cellData = MetaDataManager.getCellMetaData();

			String arborizationFileBase = DataProperties.getInstance().getFastaDir()+"Arborization/Mapping"+MAPPING+"/Size/";
			String arborization, fastaFilename;
			NeuronMetadata neuron;
			FastaReader fr;
			
			Map<String,List<Object>> classDataMap = new HashMap<String,List<Object>>();
			classDataMap.put("All", new ArrayList<Object>());
			List<Object> globalData, allTypesGlobalData = classDataMap.get("All");
			String arborType;
			Set<String> arborTypes = new TreeSet<String>();
			
			XYBean xy;
			DistributionBean globalDist, instanceDist;
			
			// Run analysis on each arbor sequence
			for (int i = 0; i < arborizations.size(); i++){
				arborization = (String)arborizations.get(i);
				System.out.println(arborization);

				// Get neuron data for arborization's lab - used to find file; + classification
				neuron = (NeuronMetadata)cellData.get(NeuriteSequenceStringUtils.getNeuronNameFromSequence(arborization));
				neuron.setCellType(ClassificationUtils.reclassifyCellTypes(neuron.getCellType()));
				arborType = ClassificationUtils.reclassifyArborizationTypes(
						NeuriteSequenceStringUtils.getArborizationType(arborization));
				arborTypes.add(arborType);
				
				fastaFilename = arborizationFileBase + neuron.getLabName() 
					+ "/" + arborization + ".lts.fasta";
				// Get sequence
				fr = new FastaReader(fastaFilename);
				fr.next();
				
				// Get %C profile
				List percentC = generateProfileFromRoot(fr.getSequence());
				if (!classDataMap.containsKey(arborType)){
					classDataMap.put(arborType, new ArrayList());
				}
				
				// Get dataset for given arbor type
				globalData = classDataMap.get(arborType);
				// run through all branch orders of the arbor
				for (int j = 0; j < percentC.size(); j++){
					// Pull distribution (%C) for the branch order
					instanceDist = (DistributionBean)percentC.get(j);
					switch (profileMode){
					case PROFILE_MODE_SUM:
						// Get global %C distribution for this branch order
						if (j >= globalData.size()){
							globalData.add((Object)new DistributionBean());
						}
						globalDist = (DistributionBean)globalData.get(j);
						// Add all node instances to distribution
						globalDist.addData(instanceDist.getData());

						// Same as above, but for All arbor types set
						if (j >= allTypesGlobalData.size()){
							allTypesGlobalData.add(new DistributionBean());
						}
						globalDist = (DistributionBean)allTypesGlobalData.get(j);
						globalDist.addData(instanceDist.getData());
						break;
					case PROFILE_MODE_PROPORTION:
						// Add normalized branch order, %C XY pair 
						xy = new XYBean(
								(double)j/percentC.size(),
								instanceDist.getMean());
						globalData.add(xy);
						allTypesGlobalData.add(xy);
						break;
					}
				}
			}

			System.gc();
			
			// Write to file
			Set<String> allClasses = new LinkedHashSet<String>();
			allClasses.add("All");
			allClasses.addAll(arborTypes);
			
			String outputData = "";
			String outfile = ProjectProperties.getInstance().getResultsDir()+"NodeTypeBranchOrder-";
			if (profileMode == PROFILE_MODE_SUM){
				outfile = outfile + "sum-";
				outputData = writeCombinedAbsoluteFile(classDataMap, allClasses);
			}
			else if (profileMode == PROFILE_MODE_PROPORTION){
				outfile = outfile + "proportion-";
				outputData = writeProportionFile(classDataMap, allClasses);
			}
			outfile = outfile + "Analysis.csv";
			DataWriter.writeStringToFile(outfile, outputData);
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static String writeCombinedAbsoluteFile(Map classDataMap, Set<String> allClasses){
		StringBuffer sb = new StringBuffer();
		int maxBranchOrder = 0;
		String className;
		List list;
		NumberFormat nf = NumberFormat.getInstance();
		nf.setGroupingUsed(false);
		nf.setMaximumFractionDigits(4);
		for (Iterator it = allClasses.iterator(); it.hasNext();){
			list = (List)classDataMap.get(it.next());
			if (list.size() > maxBranchOrder){
				maxBranchOrder = list.size()-1;
			}
		}
		sb.append("BranchOrder");
		for (int branchOrder = 0; branchOrder <= maxBranchOrder; branchOrder++){
			sb.append(","+branchOrder);
		}
		sb.append("\r\n");
		for (Iterator it = allClasses.iterator(); it.hasNext();){
			className = (String)it.next();
			sb.append(className+" Mean");
			list = (List)classDataMap.get(className);
			for (int i = 0; i <= maxBranchOrder; i++){
				sb.append(",");
				if (i < list.size()){
					//sb.append(nf.format(((DistributionBean)list.get(i)).getCount()));
					sb.append(nf.format(((DistributionBean)list.get(i)).getMean()));
				}
			}
			sb.append("\r\n");
			sb.append(className+" N");
			for (int i = 0; i <= maxBranchOrder; i++){
				sb.append(",");
				if (i < list.size()){
					//sb.append(nf.format(((DistributionBean)list.get(i)).getCount()));
					sb.append(nf.format(((DistributionBean)list.get(i)).getCount()));
				}
			}
			sb.append("\r\n");
			sb.append(className+" SD");
			for (int i = 0; i <= maxBranchOrder; i++){
				sb.append(",");
				if (i < list.size()){
					//sb.append(nf.format(((DistributionBean)list.get(i)).getCount()));
					sb.append(nf.format(((DistributionBean)list.get(i)).getStandardDeviation()));
				}
			}
			sb.append("\r\n");
		}

		return sb.toString();
	}
	
	private static String writeProportionFile(Map classDataMap, Set allClasses){
		StringBuffer sb = new StringBuffer();
		String className;
		int size, maxDataSize = 0;
		NumberFormat nf = NumberFormat.getInstance();
		nf.setGroupingUsed(false);
		nf.setMaximumFractionDigits(4);
		for (Iterator it = allClasses.iterator(); it.hasNext();){
			className = (String)it.next();
			System.out.println(className);
			sb.append(className+",,");
			size = ((List)classDataMap.get(className)).size();
			if (size > maxDataSize){
				maxDataSize = size;
			}
		}
		sb.append("\r\n");
		for (Iterator it = allClasses.iterator(); it.hasNext();){
			it.next();
			sb.append("ProportionalBranchOrder,%C,");
		}
		sb.append("\r\n");
		List dataList;
		XYBean xy;
		//NumberFormat nf = NumberFormat.getInstance();
		for (int i = 0; i < maxDataSize; i++){
			for (Iterator it = allClasses.iterator(); it.hasNext();){
				className = (String)it.next();
				dataList = (List)classDataMap.get(className);
				if (i < dataList.size()){
					xy = (XYBean)dataList.get(i);
					sb.append(nf.format(xy.getX())+","+nf.format(xy.getY())+",");
				}
				else{
					sb.append(",,");
				}
			}
			sb.append("\r\n");
		}
		return sb.toString();
	}
	
	private static List generateProfileFromRoot(String sequence){
		List percentC = new ArrayList();
		DistributionBean dist;
		//LinkedStack helpers = new LinkedStack();
		SequenceTraversalHelper helper = new SequenceTraversalHelper();
		
		for (int i = 0; i < sequence.length(); i++){
			helper.traverse(sequence.charAt(i));
			// Add to data
			if (helper.getLevel()+1 > percentC.size()){
				percentC.add(new DistributionBean());
			}
			dist = (DistributionBean)percentC.get(helper.getLevel());
			if (sequence.charAt(i) == 'A'){
				dist.addData(0);
			}
			else if (sequence.charAt(i) == 'C'){
				dist.addData(1);
			}
		}
		return percentC;
	}
	
	private static List generateProfileFromTermination(String sequence){
		List percentC = new ArrayList();
		return percentC;
	}
		
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;

        opt = new FlaggedOption(PARAM_PROFILE_MODE)
        	.setStringParser(JSAP.INTEGER_PARSER)
        	.setRequired(false)
        	.setDefault(""+PROFILE_MODE_SUM)
        	.setShortFlag('m') 
        	.setLongFlag(PARAM_PROFILE_MODE);
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + NodeTypeBranchOrderAnalysis.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
