package org.cn3.ns.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.ns.domain.FastaBean;

public class FastaReader {

	private String filename;
	private File file;
	private boolean closed = true;

	private FastaBean fastaBean = new FastaBean();
	
	private String waitingString = null;
	private BufferedReader reader = null;
	
	public FastaReader(String filename) throws FileNotFoundException{
		setFile(filename);
	}
	
	public FastaReader(File file) throws FileNotFoundException{
		setFile(file);
	}
	
	public FastaReader(){}
	
	public void setFile(String filename) throws FileNotFoundException{
		setFile(new File(filename));
	}
	
	public void setFile(File file) throws FileNotFoundException{
		this.filename = file.getPath();
		this.file = file;
		reader = new BufferedReader(new FileReader(file));
		closed = false;
	}
	
	public boolean next() throws IOException{
		if (closed){
			return false;
		}
		String dataString;
		int state = 0;
		Pattern pHead = Pattern.compile("^>.+");
		Pattern pSeq = Pattern.compile("^[ABCDEFGHIKLMNPQRSTUVWY]+");
		Matcher m;
		boolean hasNext = false;
		fastaBean.setNeuron(null);
		fastaBean.setSequenceName(null);

		if (waitingString != null){
			m = pHead.matcher(waitingString);
			if (m.matches()){
				state = 1;
				fastaBean.setHeader(waitingString);
			}
			waitingString = null;
		}
		while ( waitingString == null
				&& reader != null && (dataString = reader.readLine()) != null){
			switch (state){
			// Ready to read header
			case 0:
				m = pHead.matcher(dataString);
				if (m.matches()){
					state = 1;
					fastaBean.setHeader(dataString);
				}
				break;
			// Ready to read sequence
			case 1:
				m = pSeq.matcher(dataString);
				if (m.matches()){
					state = 2;
					fastaBean.setSequence(dataString);
					hasNext = true;
				}
				else{
					state = 0;
				}
				break;
			// Continuing to read sequence
			case 2:
				m = pSeq.matcher(dataString);
				if (m.matches()){
					fastaBean.setSequence(fastaBean.getSequence().concat(dataString));
				}
				else{
					// Not a sequence string, may be next header
					waitingString = dataString;
				}
				break;
			}
		}
		if (!hasNext){
			close();
		}
		return hasNext;
	}
	
	public String getHeader(){
		return fastaBean.getHeader();
	}
	
	public String getSequence(){
		return fastaBean.getSequence();
	}

	public File getFile() {
		return file;
	}

	public String getFilename() {
		return filename;
	}

	public String getSequenceName(){
		if (fastaBean.getSequenceName() == null){
			processHeader();
		}
		return fastaBean.getSequenceName();
	}
	
	public int getSequenceLength(){
		return getSequence().length();
	}
	
	public NeuronMetadata getNeuron(){
		if (fastaBean.getNeuron() == null){
			processHeader();
		}
		return fastaBean.getNeuron();
	}
	
	private void processHeader(){
		Pattern p = Pattern.compile(">?(.*?)((-(ApicalDendrite|BasalDendrite|Dendrite|Axon)\\d*)? \\| ([^\\|]*) \\| ([^\\|]*) \\| ([^\\|]*) \\| ([^\\|]*))?");
		Matcher m = p.matcher(fastaBean.getHeader());
		NeuronMetadata neuron;
		if (m.matches()){
			if (m.group(2) != null){
				neuron = new NeuronMetadata();
				fastaBean.setNeuron(neuron);
				fastaBean.setSequenceName(m.group(1).trim()+m.group(3).trim());
				neuron.setName(m.group(1).trim());
				neuron.setCellType(m.group(5).trim());
				neuron.setBrainRegion(m.group(6).trim());
				neuron.setAnimal(m.group(7).trim());
				neuron.setLabName(m.group(8).trim());
				fastaBean.setArborType(m.group(4));
			}
			else{
				fastaBean.setSequenceName(m.group(1).trim());
			}
		}
	}
	
	public String getArborType(){
		if (fastaBean.getNeuron() == null){
			getNeuron();
		}
		return fastaBean.getArborType();
	}

	public FastaBean getFastaBean(){
		if (fastaBean.getNeuron() == null){
			getNeuron();
		}
		return fastaBean;
	}

	/**
	 * Returns the number of sequences in the fasta file.
	 * @return
	 */
	public static int getNumSequences(String filename)throws IOException{
		return getNumSequences(new File(filename));
	}
	public static int getNumSequences(File file) throws IOException{
		FastaReader fr = new FastaReader(file);
		int count = 0;
		while (fr.next()){
			count++;
		}
		return count;
	}
	
	public static List<String> getQueryNames(File fastaFile) throws IOException{
		List<String> queryNames = new ArrayList<String>(FastaReader.getNumSequences(fastaFile));
		FastaReader fr = new FastaReader(fastaFile);
		while (fr.next()){
			queryNames.add(fr.getSequenceName());
		}
		return queryNames;
	}

	public static Map<String,Integer> getQueryLengths(File fastaFile) throws IOException{
		Map<String,Integer> queryNames = new LinkedHashMap<String,Integer>(FastaReader.getNumSequences(fastaFile));
		FastaReader fr = new FastaReader(fastaFile);
		while (fr.next()){
			queryNames.put(fr.getSequenceName(),fr.getSequenceLength());
		}
		return queryNames;
	}

	public boolean isClosed(){
		return closed;
	}
	
	public void close(){
		try{
			if (reader != null){
				reader.close();
			}
		} catch (IOException ioe){
			ioe.printStackTrace();
		}
		closed = true;
	}

	public static String readFastaHeader(String inputFileString)throws Exception{
		return readFastaHeader(new File(inputFileString));
	}

	public static String readFastaHeader(File inputFile)throws Exception{
		FileReader freader = null;
		BufferedReader reader = null;
		String header = null;
	
		try{
			freader = new FileReader(inputFile);
		    reader = new BufferedReader(freader);
		    
			String dataString = "";
			Pattern p = Pattern.compile("^>.+");
			Matcher m;
			
			while ( (dataString = reader.readLine()) != null ) { 
				m = p.matcher(dataString);
				if (m.matches()){
					header = dataString;
					break;
				}
			}
		
		} catch (IOException ioe){
			throw new Exception(ioe);
		} finally{
			if (reader != null){
				reader.close();
			}
			if (freader != null){
				freader.close();
			}
		}
	
		return header;
	}
}
