package org.cn3.ns.data;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.MetaDataManager;
import org.cn3.morpho.data.MorphoDataReader;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.ns.data.persistence.GACTCompressor;
import org.cn3.ns.nodeencoding.EncodingMode;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.resources.ProjectProperties;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;


public class NSDataReader extends MorphoDataReader {
	
	private static Set<String> flaggedNeurons = null;
	public static Set<String> getFlaggedNeurons() throws IOException{
		if (flaggedNeurons == null){
			flaggedNeurons = new TreeSet<String>();
			FileReader freader = null;
		    BufferedReader reader = null;

		    try{
			    freader = new FileReader(DataProperties.getInstance().getFlaggedNeuronsFile());
			    reader = new BufferedReader(freader);

				String dataString = "";
				
				while ( (dataString=reader.readLine()) != null ) {
					dataString = dataString.replaceFirst(" *#.*", "");
					if (!NeuriteSequenceStringUtils.isEmpty(dataString)){
						flaggedNeurons.add(dataString);
					}
				}
				
			} catch (IOException ioe){
				throw ioe;
			} finally{
				if (reader != null){
					reader.close();
				}
				if (freader != null){
					freader.close();
				}
			}
		}
		Set<String> copy = new TreeSet<String>(flaggedNeurons);
		return copy;
	}
	
	private static Map<String,List<String>> flaggedSequences = new HashMap<String,List<String>>(2);
	private static Map<String,List<Integer>> uniqueSequenceLengths = new HashMap<String,List<Integer>>(2);
	private static Map<String,Map<String,Integer>> lengthDatabases = new HashMap<String,Map<String,Integer>>(2);
	private static Map<String,Map<String,Integer>> lengthDatabasesFromFastaFile = new HashMap<String,Map<String,Integer>>(2);
	private static Map<String,List<String>> sequenceLists = new HashMap<String,List<String>>(2);
	private static Map<String,List<String>> sequenceListsFromFastaFile = new HashMap<String,List<String>>(2);
	private static Map<String,Map<String,Integer>> dataMaps = new HashMap<String,Map<String,Integer>>(2);
	private static Map<String,Set<String>> dataSets = new HashMap<String,Set<String>>(2);
	private static Map<Integer,Map<Integer,List<String>>> lengthArborizationsMaps = 
		new HashMap<Integer,Map<Integer,List<String>>>(2);
	private static Map<Integer,Map<String,Integer>> sequenceIdMaps = new HashMap<Integer,Map<String,Integer>>(2);

	// Returns all possible sequence names for flagged neurons, based on the classification type 
	public static List<String> getFlaggedSequences() throws IOException{
		return getFlaggedSequences(NSConstants.CLASSIFICATION_ARBORIZATION);
	}
	public static List<String> getFlaggedSequences(String classificationType) throws IOException{
		if (!flaggedSequences.containsKey(classificationType)){
			List<String> flaggedSequenceList = new ArrayList<String>();
			Set<String> flaggedNeurons = getFlaggedNeurons();
			
			if (classificationType.equals(NSConstants.CLASSIFICATION_CELL)){
				flaggedSequenceList.addAll(flaggedNeurons);
			}
			else if (classificationType.equals(NSConstants.CLASSIFICATION_TREE)){
				for (String neuronName : flaggedNeurons){
					if (!neuronName.matches(".+-(Axon|BasalDendrite|Dendrite|ApicalDendrite)")){
						// Cover all the bases
						flaggedSequenceList.add(neuronName+"-Axon");
						flaggedSequenceList.add(neuronName+"-ApicalDendrite");
						for (int i = 0; i < 12; i++){
							flaggedSequenceList.add(neuronName+"-BasalDendrite"+i);
							flaggedSequenceList.add(neuronName+"-Dendrite"+i);
						}
					}else{
						if (!neuronName.matches(".+-(BasalDendrite|Dendrite)")){
							for (int i = 0; i < 12; i++){
								flaggedSequenceList.add(neuronName+i);
								flaggedSequenceList.add(neuronName+i);
							}
						}else{
							flaggedSequenceList.add(neuronName);
						}
					}
				}
			}
			else{ // Classification type "Arborization"
				for (String neuronName : flaggedNeurons){
					if (!neuronName.matches(".+-(Axon|BasalDendrite|Dendrite|ApicalDendrite)")){
						// Cover all the bases
						flaggedSequenceList.add(neuronName+"-Axon");
						flaggedSequenceList.add(neuronName+"-BasalDendrite");
						flaggedSequenceList.add(neuronName+"-ApicalDendrite");
						flaggedSequenceList.add(neuronName+"-Dendrite");
					}else{
						flaggedSequenceList.add(neuronName);
					}
				}
			}
			flaggedSequences.put(classificationType, flaggedSequenceList);
		}
		return flaggedSequences.get(classificationType);
	}

	public static List<String> loadSequenceListFromFastaFile() throws Exception{
		return loadSequenceListFromFastaFile(NSConstants.DEFAULT_MAPPING_MODE, NSConstants.DEFAULT_DATASET, DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceListFromFastaFile(int mapping) throws Exception{
		return loadSequenceListFromFastaFile(mapping, NSConstants.DEFAULT_DATASET, DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceListFromFastaFile(int mapping, boolean excludeFlaggedNeurons) throws Exception{
		return loadSequenceListFromFastaFile(mapping, NSConstants.DEFAULT_DATASET, excludeFlaggedNeurons);
	}
	public static List<String> loadSequenceListFromFastaFile(int mapping, String dataset) throws Exception{
		return loadSequenceListFromFastaFile(
				DataProperties.getInstance().getFastaFilename(dataset, mapping, NSConstants.DEFAULT_ORDER),
				DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceListFromFastaFile(int mapping, String dataset, String order) throws Exception{
		return loadSequenceListFromFastaFile(
				DataProperties.getInstance().getFastaFilename(dataset, mapping, order),
				DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceListFromFastaFile(int mapping, String dataset, String order, boolean excludeFlaggedNeurons)
	throws Exception{
		return loadSequenceListFromFastaFile(
				DataProperties.getInstance().getFastaFilename(dataset, mapping, order),
				excludeFlaggedNeurons);
	}
	public static List<String> loadSequenceListFromFastaFile(int mapping, String dataset, boolean excludeFlaggedNeurons) throws Exception{
		return loadSequenceListFromFastaFile(
				DataProperties.getInstance().getFastaFilename(dataset, mapping, NSConstants.DEFAULT_ORDER),
				excludeFlaggedNeurons);
	}
	public static List<String> loadSequenceListFromFastaFile(String fastaFile) throws Exception{
		return loadSequenceListFromFastaFile(fastaFile, DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceListFromFastaFile(String lengthDatabase, boolean excludeFlaggedNeurons) throws Exception{
		if (!sequenceListsFromFastaFile.containsKey(lengthDatabase)){
			loadLengthDatabaseFromFastaFile(lengthDatabase);
		}
		List<String> sequenceList = new ArrayList<String>(sequenceListsFromFastaFile.get(lengthDatabase));
		if (excludeFlaggedNeurons){
			sequenceList.removeAll(new ArrayList<String>(getFlaggedSequences()));
		}
		return sequenceList;
	}

	public static String getSequenceById(int id) throws IOException{
		if (!sequenceLists.containsKey(new Integer(NSConstants.DEFAULT_MAPPING_MODE))){
			loadSequenceList();
		}
		List<String> sequenceList = sequenceLists.get(new Integer(NSConstants.DEFAULT_MAPPING_MODE));
		return (String)sequenceList.get(id);
	}
	private static final boolean DEFAULT_EXCLUDE_FLAGGED_NEURONS = false;
	public static List<String> loadSequenceList() throws IOException{
		return loadSequenceList(NSConstants.DEFAULT_MAPPING_MODE);
	}
	public static List<String> loadSequenceList(int mapping) throws IOException{
		return loadSequenceList(mapping, NSConstants.DEFAULT_DATASET, DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceList(int mapping, boolean excludeFlaggedNeurons) throws Exception{
		return loadSequenceList(mapping, NSConstants.DEFAULT_DATASET, excludeFlaggedNeurons);
	}
	public static List<String> loadSequenceList(int mapping, String dataset) throws Exception{
		return loadSequenceList(
				DataProperties.getInstance().getSequenceLengthsFile(mapping, "arborization", dataset),
				DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceList(int mapping, String dataset, boolean excludeFlaggedNeurons) throws IOException{
		return loadSequenceList(
				DataProperties.getInstance().getSequenceLengthsFile(mapping, "arborization", dataset),
				excludeFlaggedNeurons);
	}
	public static List<String> loadSequenceListBySeqType(int seqType, String dataset) throws IOException{
		return loadSequenceList(
				DataProperties.getInstance().getSequenceLengthsFileBySeqType(seqType, "arborization", dataset));
	}
	public static List<String> loadSequenceList(String lengthDatabase) throws IOException{
		return loadSequenceList(lengthDatabase, DEFAULT_EXCLUDE_FLAGGED_NEURONS);
	}
	public static List<String> loadSequenceList(String lengthDatabase, boolean excludeFlaggedNeurons) throws IOException{
		if (!sequenceLists.containsKey(lengthDatabase)){
			loadLengthDatabase(lengthDatabase);
		}
		List<String> sequenceList = new ArrayList<String>(sequenceLists.get(lengthDatabase));
		if (excludeFlaggedNeurons){
			sequenceList.removeAll(new ArrayList<String>(getFlaggedSequences()));
		}
		return sequenceList;
	}

	public static Map<String,Integer> loadSequenceIdMap() throws Exception{
		return loadSequenceIdMap(NSConstants.DEFAULT_MAPPING_MODE);
	}
	public static Map<String,Integer> loadSequenceIdMap(int mapping) throws IOException{
		Integer mapObj = new Integer(mapping);
		if (!sequenceIdMaps.containsKey(mapObj)){
			List<String> sequenceNames = loadSequenceList(mapping);
			Map<String,Integer> sequenceIdMap = new HashMap<String,Integer>();
			for (int j = 0; j < sequenceNames.size(); j++){
				sequenceIdMap.put(sequenceNames.get(j), new Integer(j));
			}
			sequenceIdMaps.put(mapObj, sequenceIdMap);
		}
		return new HashMap<String,Integer>(sequenceIdMaps.get(mapObj));
	}
	public static int getSequenceId(String sequence) throws IOException{
		return getSequenceId(NSConstants.DEFAULT_MAPPING_MODE, sequence);
	}
	public static int getSequenceId(int mapping, String sequence) throws IOException{
		Integer mapObj = new Integer(mapping);
		if (!sequenceIdMaps.containsKey(mapObj)){
			loadSequenceIdMap(mapping);
		}
		Map<String,Integer> sequenceIdMap = sequenceIdMaps.get(mapObj);
		return (sequenceIdMap.containsKey(sequence) ? 
				((Integer)sequenceIdMap.get(sequence)).intValue() : -1);
	}
	
	public static Map<Integer,List<String>> loadTreeSizeArborizationsMap() throws Exception{
		return loadTreeSizeArborizationsMap(NSConstants.DEFAULT_MAPPING_MODE);
	}
	public static Map<Integer,List<String>> loadTreeSizeArborizationsMap(int mapping) throws Exception{
		if (!lengthArborizationsMaps.containsKey(mapping)){
			List<String> orderedSequences = loadSequenceListOrderedByLength(mapping);
			Map<String,Integer> sequenceLengthMap = loadLengthDatabase(mapping);
			Map<Integer,List<String>> lengthArborsMap = new HashMap<Integer,List<String>>();
			int arborSize, size = 0;
			String arborization;
			List<String> currentList = null;
			for (int i = 0; i < orderedSequences.size(); i++){
				arborization = orderedSequences.get(i);
				arborSize = sequenceLengthMap.get(arborization);
				if (arborSize != size){
					size = arborSize;
					currentList = new ArrayList<String>();
					lengthArborsMap.put(size, currentList);
				}
				currentList.add(arborization);
			}
			lengthArborizationsMaps.put(mapping, lengthArborsMap);	
		}
		return lengthArborizationsMaps.get(mapping);
	}
	public static Map<Integer,List<String>> loadTreeSizeArborizationsMap(
			Map<String,Integer> sequenceLengthMap) throws Exception{
		int seqLength;
		List<String> currentList = null;
		Map<Integer,List<String>> lengthSequencesMap = new HashMap<Integer,List<String>>();
		for (String seqName : sequenceLengthMap.keySet()){
			seqLength = sequenceLengthMap.get(seqName);
			if (!lengthSequencesMap.containsKey(seqLength)){
				lengthSequencesMap.put(seqLength, new ArrayList<String>());
			}
			currentList = lengthSequencesMap.get(seqLength);
			currentList.add(seqName);
		}
		return lengthSequencesMap;
	}
	
	public static List<String> loadSequenceListOrderedByLength() throws IOException{
		return loadSequenceListOrderedByLength(NSConstants.DEFAULT_MAPPING_MODE);
	}
	public static List<String> loadSequenceListOrderedByLength(int mapping) throws IOException{
		Map<String,Integer> lengthMap = loadLengthDatabase(NSConstants.MAPPING_MODE_TOPO_ACT);
		List<Integer> sizes = loadUniqueSequenceLengths(NSConstants.MAPPING_MODE_TOPO_ACT);
		return loadSequenceListOrderedByLength(
				DataProperties.getInstance().getSequenceLengthsFile(mapping),
				lengthMap, sizes);
	}
	public static List<String> loadSequenceListOrderedByLength(
			String lengthDatabase, Map<String,Integer> arborLengthMap, List<Integer> lengthList) throws IOException{
		String key = lengthDatabase+"-sorted";
		if (!sequenceLists.containsKey(key)){
			if (!lengthDatabases.containsKey(key)){
				loadLengthDatabase(lengthDatabase);
			}
			List<String> orderedSequenceList = orderArborizations(arborLengthMap, lengthList);
			sequenceLists.put(key, orderedSequenceList);
		}
		return sequenceLists.get(key);
	}
	private static List<String> orderArborizations(Map<String,Integer> lengthMap, List<Integer> lengthList){
		Map<Integer,List<String>> sizeToArborMap = new HashMap<Integer,List<String>>();
		List<String> arborizations = new ArrayList<String>();
		Integer size;
		for (String arbor : lengthMap.keySet()){
			size = lengthMap.get(arbor);
			if (!sizeToArborMap.containsKey(size)){
				sizeToArborMap.put(size, new ArrayList<String>());
			}
			sizeToArborMap.get(size).add(arbor);
		}
		for (int i = 0; i < lengthList.size(); i++){
			arborizations.addAll(sizeToArborMap.get(lengthList.get(i)));
		}
		return arborizations;
	}

	public static List<Integer> loadUniqueSequenceLengths() throws IOException{
		return loadUniqueSequenceLengths(NSConstants.DEFAULT_MAPPING_MODE);
	}
	public static List<Integer> loadUniqueSequenceLengths(int mapping) throws IOException{
		return loadUniqueSequenceLengths(mapping, NSConstants.DEFAULT_DATASET);
	}
	public static List<Integer> loadUniqueSequenceLengths(int mapping, String dataset) throws IOException{
		return loadUniqueSequenceLengths(
				NSDataReader.loadProperties().getProperty(
						NSConstants.LENGTH_DATABASE+"."+dataset+".arborization.type"+NSConstants.mappingModeSequenceType[mapping]));
	}
	public static List<Integer> loadUniqueSequenceLengths(EncodingMode encoding, String dataset) throws IOException{
		return loadUniqueSequenceLengths(
				NSDataReader.loadProperties().getProperty(
						NSConstants.LENGTH_DATABASE+"."+dataset+".arborization.type"+NSConstants.mappingModeSequenceType[encoding.getId()]));
	}
	public static List<Integer> loadUniqueSequenceLengths(String lengthDatabase)
	throws IOException{
		if (!uniqueSequenceLengths.containsKey(lengthDatabase)){
			if (!lengthDatabases.containsKey(lengthDatabase)){
				loadLengthDatabase(lengthDatabase);
			}
			Map<String,Integer> lengthMap = lengthDatabases.get(lengthDatabase);
			uniqueSequenceLengths.put(lengthDatabase, loadUniqueSequenceLengths(lengthMap));
		}
		return uniqueSequenceLengths.get(lengthDatabase);
	}
	public static List<Integer> loadUniqueSequenceLengths(Map<String,Integer> lengthMap)
	throws IOException{
		Set<Integer> lengths = new TreeSet<Integer>();
		Integer length;
		for (Iterator<Integer> it = lengthMap.values().iterator(); it.hasNext();){
			length = it.next();
			if (!lengths.contains(length)){
				lengths.add(length);
			}
		}
		Integer[] list = new Integer[lengths.size()];
		list = (Integer[])lengths.toArray(list);
		Arrays.sort(list);
		return new LinkedList<Integer>(Arrays.asList(list));
	}
	public static Map<String,Integer> loadLengthDatabase() throws IOException{
		return loadLengthDatabase(NSConstants.DEFAULT_MAPPING_MODE, NSConstants.DEFAULT_DATASET);
	}
	public static Map<String,Integer> loadLengthDatabase(int mapping) throws IOException{
		return loadLengthDatabase(mapping, NSConstants.DEFAULT_DATASET);
	}
	public static Map<String,Integer> loadLengthDatabase(int mapping, String dataset) throws IOException{
		return loadLengthDatabase(
				NSDataReader.loadProperties().getProperty(
						NSConstants.LENGTH_DATABASE+"."+dataset+".arborization.type"+NSConstants.mappingModeSequenceType[mapping]));
	}
	public static Map<String,Integer> loadLengthDatabase(String lengthDatabase) 
	throws IOException{
		if (!lengthDatabases.containsKey(lengthDatabase)){
			FileReader freader = null;
		    BufferedReader reader = null;
		    Map<String,Integer> lengthMap = null;
		    List<String> seqList = null;
//		    System.out.println(lengthDatabase);
		    try{
			    freader = new FileReader(lengthDatabase);
			    reader = new BufferedReader(freader);
			    lengthMap = new HashMap<String,Integer>();
			    seqList = new ArrayList<String>();

				String dataString = "";
				String[] data;
				
				
				while ( (dataString=reader.readLine()) != null ) { 
			        data = dataString.split("\t");
			        lengthMap.put(data[0], new Integer(data[1]));
			        seqList.add(data[0]);
				}
				
			} catch (IOException ioe){
				throw ioe;
			} finally{
				if (reader != null){
					reader.close();
				}
				if (freader != null){
					freader.close();
				}
			}
			lengthDatabases.put(lengthDatabase, lengthMap);
			sequenceLists.put(lengthDatabase, seqList);
		}
	
		return lengthDatabases.get(lengthDatabase);
	}
	
	public static Map<String,Integer> loadLengthDatabaseFromFastaFile(int mapping, String dataset) throws Exception{
		return loadLengthDatabaseFromFastaFile(
				DataProperties.getInstance().getFastaFilename(dataset, mapping, NSConstants.DEFAULT_ORDER));
	}

	public static Map<String,Integer> loadLengthDatabaseFromFastaFile(int mapping, String dataset, String order) throws Exception{
		return loadLengthDatabaseFromFastaFile(
				DataProperties.getInstance().getFastaFilename(dataset, mapping, order));
	}

	public static Map<String,Integer> loadLengthDatabaseFromFastaFile(String fastaFilename) 
	throws Exception{
		if (!lengthDatabasesFromFastaFile.containsKey(fastaFilename)){
			FastaReader fr = new FastaReader(fastaFilename);
		    Map<String,Integer> lengthMap = null;
		    List<String> seqList = null;

		    lengthMap = new HashMap<String,Integer>();
		    seqList = new ArrayList<String>();

			while (fr.next()) { 
		        lengthMap.put(fr.getSequenceName(), new Integer(fr.getSequenceLength()));
		        seqList.add(fr.getSequenceName());
			}
				
			lengthDatabasesFromFastaFile.put(fastaFilename, lengthMap);
			sequenceListsFromFastaFile.put(fastaFilename, seqList);
		}
	
		return lengthDatabasesFromFastaFile.get(fastaFilename);
	}
	
	public static Map<String,Integer> generateLengthMapFromFasta(String fastaFilename) throws IOException{
		Map<String,Integer> lengthMap = new HashMap<String,Integer>();
		FastaReader fr = new FastaReader(fastaFilename);
		while (fr.next()){
			lengthMap.put(fr.getSequenceName(), new Integer(fr.getSequenceLength()));
		}
		return lengthMap;
	}
	
	public static Map<String,Integer> loadOutliers() throws Exception{
		return loadOutliers(NSConstants.DEFAULT_MAPPING_MODE, NSConstants.DEFAULT_ORDER, 
				NSConstants.DEFAULT_BASELINE);
	}
	public static Map<String,Integer> loadOutliers(int mapping, String order) throws Exception{
		return loadOutliers(mapping, order, NSConstants.DEFAULT_BASELINE);
	}
	public static Map<String,Integer> loadOutliers(int mapping, String order, int method) throws Exception{
		return loadOutliers(mapping, order, method, 0, "");
	}
	public static Map<String,Integer> loadOutliers(int mapping, String order, int method, float threshold) throws Exception{
		return loadOutliers(mapping, order, method, threshold, "");
	}
	public static Map<String,Integer> loadOutliers(int mapping, String order, int method, float threshold, String thresholdModifier) throws Exception{
		DataProperties props = DataProperties.getInstance();
		return loadOutliers(props.getOutlierFile(mapping, order, method, threshold, thresholdModifier));
	}
	public static Map<String,Integer> loadOutliers(int mapping, String order, int method, double threshold, String thresholdModifier, long processorMinuteCutoff) throws Exception{
		DataProperties props = DataProperties.getInstance();
		return loadOutliers(props.getOutlierFile(mapping, order, method, threshold, thresholdModifier), processorMinuteCutoff);
	}
	public static Map<String,Integer> loadOutliers(String outlierFilename)throws Exception{
		return loadOutliers(outlierFilename, 0);
	}
	public static Map<String,Integer> loadOutliers(String outlierFilename, long processorMinuteCutoff)throws Exception{
		if (!dataMaps.containsKey(outlierFilename)){
			File file = new File(outlierFilename);
			FileReader freader = null;
		    BufferedReader reader = null;
		    String dataString = "";
			
		    // If no specific outlier file, create with initial contents of master file Outliers.txt
		    if (!file.exists()){
				file.createNewFile();
				File masterFile = new File(
						DataProperties.getInstance().getOutlierDirectory()+"Outliers.txt");
				StringBuffer sb = new StringBuffer();
				try{
					freader = new FileReader(masterFile);
				    reader = new BufferedReader(freader);
				    while ( (dataString=reader.readLine()) != null ) {
						if (!NeuriteSequenceStringUtils.isEmpty(dataString)){
					        sb.append(dataString+"\n");
						}
					}
				    DataWriter.writeStringToFile(file, sb.toString());
				} catch (IOException ioe){
					throw new Exception(ioe);
				} finally{
					if (reader != null){
						reader.close();
					}
					if (freader != null){
						freader.close();
					}
				}
			}

			Map<String,Integer> outlierMap = new HashMap<String,Integer>();
			String[] parts;
			int minutes, processorMinutes;
		    try{
			    freader = new FileReader(outlierFilename);
			    reader = new BufferedReader(freader);
				
				while ( (dataString=reader.readLine()) != null ) {
					if (!NeuriteSequenceStringUtils.isEmpty(dataString)){
						parts = dataString.split(",");
						// If time data doesn't exist, this one isn't worth looking at at all
						minutes = (int)Double.parseDouble(parts[3]);
						processorMinutes = (parts.length > 6 ? Integer.parseInt(parts[6])*minutes : minutes);
						if (parts.length < 4 || processorMinutes >= processorMinuteCutoff)
							outlierMap.put(parts[0],new Integer(parts[1]));
					}
				}
			} catch (IOException ioe){
				throw new Exception(ioe);
			} finally{
				if (reader != null){
					reader.close();
				}
				if (freader != null){
					freader.close();
				}
			}
			dataMaps.put(outlierFilename, outlierMap);
		}
		return dataMaps.get(outlierFilename);
	}

	public static Set<String> loadOutliersSet(int mapping, int k) throws Exception{
		return loadOutliersSet(mapping, NSConstants.DEFAULT_ORDER, k);
	}
	public static Set<String> loadOutliersSet(int mapping, String order, int k) throws Exception{
		Map<String,Integer> map = loadOutliers(mapping, order, NSConstants.DEFAULT_BASELINE);
		Set<String> set = new TreeSet<String>();
		String key;
		int j;
		for (Iterator<String> it = map.keySet().iterator(); it.hasNext();){
			key = it.next();
			j = map.get(key).intValue();
			if (j <= k){
				set.add(key);
			}
		}
		return set;
	}

	public static LinkedHashSet<String> loadOrderedSet(String filename)
	throws Exception{
		LinkedHashSet<String> set = new LinkedHashSet<String>();
		FileReader freader = null;
	    BufferedReader reader = null;
	    try{
		    freader = new FileReader(filename);
		    reader = new BufferedReader(freader);
			String dataString = "";
			
			while ( (dataString=reader.readLine()) != null ) { 
		        set.add(dataString.trim());
			}
		} catch (IOException ioe){
			throw new Exception(ioe);
		} finally{
			if (reader != null){
				reader.close();
			}
			if (freader != null){
				freader.close();
			}
		}
		return set;
	}
	public static Set<String> loadSet(String filename)
	throws Exception{
		Set<String> set = new TreeSet<String>();
		FileReader freader = null;
	    BufferedReader reader = null;
	    try{
		    freader = new FileReader(filename);
		    reader = new BufferedReader(freader);
			String dataString = "";
			
			while ( (dataString=reader.readLine()) != null ) { 
		        set.add(dataString.trim());
			}
		} catch (IOException ioe){
			throw new Exception(ioe);
		} finally{
			if (reader != null){
				reader.close();
			}
			if (freader != null){
				freader.close();
			}
		}
		return set;
	}
	public static Set<String> loadDimerTooUnlikely(int mapping, int method) throws Exception{
		return loadSet(
				DataProperties.getInstance().getDimerTooUnlikelyFile(mapping, method));
	}
	public static Set<String> loadTrimerTooUnlikely(int mapping, int method) throws Exception{
		return loadSet(
				DataProperties.getInstance().getTrimerTooUnlikelyFile(mapping, method));
	}
	public static Set<String> loadOutliersFromAllTrees(int mapping) throws Exception{
		return loadOutliersFromAllTrees(
				NSDataReader.loadProperties().getProperty(
						NSConstants.OUTLIERS_ALL_TREE+".map"+mapping));
	}
	public static Set<String> loadOutliersFromAllTrees(String outlierFilename)
	throws Exception{
		if (!dataSets.containsKey(outlierFilename)){
			dataSets.put(outlierFilename, loadSet(outlierFilename));
		}
		return dataSets.get(outlierFilename);
	}

	public static Set<String> loadCA1Set()
	throws Exception{
		String filename = 
			ProjectProperties.getInstance().getResourceDirectory()+"/ca1.txt";
		if (!dataSets.containsKey(filename)){
			dataSets.put(filename, loadSet(filename));
		}
		return dataSets.get(filename);
	}
	public static Set<String> loadCA3Set()
	throws Exception{
		String filename = 
			ProjectProperties.getInstance().getResourceDirectory()+"/ca3.txt";
		if (!dataSets.containsKey(filename)){
			dataSets.put(filename, loadSet(filename));
		}
		return dataSets.get(filename);
	}

	
	public static StringBuffer fileToStringBuffer(String file, StringBuffer sb)
	throws Exception{
		return fileToStringBuffer(new File(file), sb);
	}
	public static StringBuffer fileToStringBuffer(File file, StringBuffer sb)
	throws Exception{
		if (sb == null){
			sb = new StringBuffer();
		}
		FileReader freader = null;
	    BufferedReader reader = null;

	    try{
		    freader = new FileReader(file);
		    reader = new BufferedReader(freader);

			String dataString = "";
			
			while ( (dataString=reader.readLine()) != null ) { 
		        sb.append(dataString+"\r\n");
			}
			
		} catch (IOException ioe){
			throw new Exception(ioe);
		} finally{
			if (reader != null){
				reader.close();
			}
			if (freader != null){
				freader.close();
			}
		}
		return sb;
	}
	
	private static Map<String,Properties> propMap = new HashMap<String,Properties>();
	
	public static Properties loadProperties() {
		return loadProperties(NSConstants.DATA_PROPERTIES_FILE);
	}

	public static Properties loadProperties(String propertiesFile) {
		return loadProperties(propertiesFile, false);
	}
	public static Properties loadProperties(String propertiesFile, boolean refresh) {
		try{
			if (refresh || !propMap.containsKey(propertiesFile)){
				Pattern p = Pattern.compile(".+\\.properties");
				Matcher m = p.matcher(propertiesFile);
				if (!m.matches()){
					propertiesFile = propertiesFile + ".properties";
				}
				Properties properties = new Properties();
				properties.load(new FileInputStream("resources/"+propertiesFile));
				propMap.put(propertiesFile, properties);
			}
			return (Properties)propMap.get(propertiesFile);
		} catch (IOException ioe){
			ioe.printStackTrace();
		}
		return null;
	}
	
	public static String readNeuromorphoSequence(int mapping, String order) throws Exception{
		FileReader freader = null;
		BufferedReader reader = null;
		StringBuffer sb = new StringBuffer();

		try{
			freader = new FileReader(DataProperties.getInstance()
					.getNeuromorphoSequenceFile(mapping, order));
		    reader = new BufferedReader(freader);
		    
			String dataString = "";
			while ( (dataString = reader.readLine()) != null ) { 
				sb.append(dataString);
			}
		
		} catch (IOException ioe){
			throw new Exception(ioe);
		} finally{
			if (reader != null){
				reader.close();
			}
			if (freader != null){
				freader.close();
			}
		}

		return sb.toString();
	}
	
	public static List<String> readCompressedSequences(String filename) throws Exception{
		FileReader freader = null;
		BufferedReader reader = null;
		List<String> sequences = new ArrayList<String>();
		String dataString;
		try{
			freader = new FileReader(filename);
		    reader = new BufferedReader(freader);

			while ( (dataString=reader.readLine()) != null) {
				sequences.add(GACTCompressor.decompressSequence(dataString.toCharArray()));
		    }
		} catch (IOException ioe){
			throw new Exception(ioe);
		} finally{
			if (reader != null){
				reader.close();
			}
			if (freader != null){
				freader.close();
			}
		}
		return sequences;
	}
	
	public static String getSequenceByName(String sequenceName, Map<String,?> metaData) throws Exception{
		return getSequenceByName(sequenceName, NSConstants.DEFAULT_MAPPING_MODE, metaData);
	}
	public static String getSequenceByName(String sequenceName) throws Exception{
		return getSequenceByName(sequenceName, NSConstants.DEFAULT_MAPPING_MODE, MetaDataManager.getCellMetaData());
	}
	public static String getSequenceByName(String sequenceName, int mapping) throws Exception{
		return getSequenceByName(sequenceName, mapping, MetaDataManager.getCellMetaData());
	}
	public static String getSequenceByName(String sequenceName, int mapping, Map<String,?> metaData) throws Exception{
		FastaReader fr = new FastaReader(DataProperties.getInstance().getFastaArborDir()+"Mapping"+mapping+"/Size/"+
				((NeuronMetadata)metaData.get(NeuriteSequenceStringUtils.getNeuronNameFromSequence(sequenceName))).getLabName().toLowerCase()+
				"/"+sequenceName+".lts.fasta");
		if (fr.next()){
			return fr.getSequence();
		}
		else return null;
	}
	
	public static String readLine(InputStream bis) throws Exception{
		return readLine(bis, true);
	}
	public static String readLine(InputStream bis, boolean doubleLineBreak) throws IOException{
		String dataString = "";
		char nextChar = ' ';
		
		do {
			dataString = dataString + nextChar;
			nextChar = (char)bis.read();
		} while ((int)nextChar != 65535 && nextChar != '\n' && nextChar != '\r');
		if (doubleLineBreak && nextChar != -1){
			nextChar = (char)bis.read(); // Clear second line break
			//System.out.println("Should be line break: "+(byte)nextChar);
		}
		dataString = dataString.substring(1);

		if (nextChar == 65535 && dataString.isEmpty()){
			return null;
		}
		return dataString;
	}


}
