package org.cn3.ns.data.persistence;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.resources.DataProperties;

/**
 * 
 * @author gillette
 *
 * Enables compression and extraction of ACT (and G if needed) sequences
 *
 */
public class GACTCompressor {

	private static boolean debug = false;
	public static boolean isDebug(){
		return debug;
	}
	public static void setDebug(boolean debug){
		GACTCompressor.debug = debug;
	}
	
	public static String compressSequenceIntoString(String sequence){
		char[] compressed = compressSequence(sequence);
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < compressed.length; i++){
			sb.append(compressed[i]);
		}
		return sb.toString();
	}
	public static char[] compressSequence(String sequence){
		int numBytes = (int)Math.ceil((double)sequence.length()/4);
		// 4 is for the numberOfBytes which will come first
		char[] chars = new char[numBytes+4];
		char[] seqLen = intToCharArray(sequence.length());
		
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < 4; i++){
			chars[i] = seqLen[i];
			for (int j = 7; debug && j >= 0; j--){
				sb.append((seqLen[i] >> j) & 1);
			}
		}
		if (debug) System.out.println("SeqLen: "+sequence.length()+" "+sb.toString());
		
		int bt = 0;
		for (int i = 0; i < sequence.length(); i++){
			switch (sequence.charAt(i)){
			case 'G':
				bt = 0x00;
				break;
			case 'A':
				bt = 0x40;
				break;
			case 'C':
				bt = 0x80;
				break;
			case 'T':
				bt = 0xC0;
				break;
			default:
				break;
			}
			chars[4 + i/4] = (char)(chars[4 + i/4] | (bt >> 2*(i%4)));
			//System.out.println(sequence.charAt(i)+" " +i + " "+Integer.toBinaryString(chars[4+i/4]));
		}
		return chars;
	}

	public static byte[] getByteSequence(DataInputStream dis) throws IOException{
		byte[] seqLenArr = new byte[4];
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < 4; i++){
			seqLenArr[i] = dis.readByte();
			for (int j = 7; debug && j >= 0; j--){
				sb.append((seqLenArr[i] >> j) & 1);
			}
		}
		int seqLen = byteArrayToInt(seqLenArr);
		if (debug) System.out.println("SeqLen: "+seqLen+" "+sb.toString());
		
		int numChars = (int)Math.ceil((double)seqLen/4);
		byte[] byteArr = new byte[numChars+4];
		
		for (int i = 0; i < 4; i++){
			byteArr[i] = seqLenArr[i];
		}
		for (int i = 0; i < numChars; i++){
			byteArr[i+4] = dis.readByte();
		}
		return byteArr;
	}
	
	public static String decompressSequence(DataInputStream dis) throws IOException{
		byte[] seqLenArr = new byte[4];
		for (int i = 0; i < 4; i++){
			seqLenArr[i] = dis.readByte();
		}
		int seqLen = byteArrayToInt(seqLenArr);
		//System.out.println("SeqLen: "+seqLen);
		
		int numChars = (int)Math.ceil((double)seqLen/4);
		byte[] byteArr = new byte[numChars];
		
		for (int i = 0; i < numChars; i++){
			byteArr[i] = dis.readByte();
		}
		return decodeSequence(seqLen, byteArr);
	}

	public static String decompressSequence(byte[] bytes) throws IOException{
		byte[] seqLenArr = new byte[4];
		for (int i = 0; i < 4; i++){
			seqLenArr[i] = bytes[i];
		}
		int seqLen = byteArrayToInt(seqLenArr);
		
		int numChars = (int)Math.ceil((double)seqLen/4);
		byte[] byteArr = new byte[numChars];
		
		for (int i = 0; i < numChars; i++){
			byteArr[i] = bytes[i+4];
		}
		return decodeSequence(seqLen, byteArr);
	}

	public static String decompressSequence(Reader reader) throws Exception{
		char[] seqLenB = new char[4];
		reader.read(seqLenB);
		int seqLen = charArrayToInt(seqLenB);
		
		int numChars = (int)Math.ceil((double)seqLen/4);
		char[] charArr = new char[numChars];
		
		reader.read(charArr);
		return decodeSequence(seqLen, charArr);
	}
	
	public static String decompressSequence(char[] chars){
		int seqLen = getSequenceLength(chars);

		int numChars = chars.length-4;
		//System.out.println("numChars 1: "+numChars);
		
		char[] newChars = new char[numChars];
		for (int i = 0; i < numChars; i++){
			newChars[i] = chars[i+4];
		}
		return decodeSequence(seqLen, newChars);
	}

	public static int getSequenceLength(char[] chars){
		// 4 is for the numberOfBytes which will come first
		char[] seqLenB = new char[4];
		for (int i = 0; i < 4; i++){
			seqLenB[i] = chars[i];
		}
		int seqLen = charArrayToInt(seqLenB);

		return seqLen;
	}

	public static int getSequenceLength(byte[] bytes){
		// 4 is for the numberOfBytes which will come first
		byte[] seqLenB = new byte[4];
		for (int i = 0; i < 4; i++){
			seqLenB[i] = bytes[i];
		}
		int seqLen = byteArrayToInt(seqLenB);
//		System.out.println("seqLen: "+seqLen);

		return seqLen;
	}

	public static String decodeSequence(int seqLen, char[] chars){
		StringBuffer sequence = new StringBuffer();
		int numBytes = (int)Math.ceil((double)seqLen/4);
		//System.out.println("SeqLen: "+seqLen+"; numBytes: "+numBytes);
		for (int i = 0; i < numBytes; i++){
			//System.out.println(Integer.toBinaryString(chars[i]));
			for (int j = 0; j < 4; j++){
				if (i*4 + j < seqLen){
					switch ((((chars[i] >> (3-j)*2)) & 0x03)){
					case 0:
						sequence.append("G");
						break;
					case 1:
						sequence.append("A");
						break;
					case 2:
						sequence.append("C");
						break;
					case 3:
						sequence.append("T");
						break;
					default:
						break;
					}
				}
			}
		}
		return sequence.toString();
	}
	public static String decodeSequence(int seqLen, byte[] bytes){
		StringBuffer sequence = new StringBuffer();
		int numBytes = (int)Math.ceil((double)seqLen/4);
		
		//System.out.println("NumBytes: "+numBytes);
		for (int i = 0; i < numBytes; i++){
			//System.out.println(Integer.toString((char)bytes[i+4],2));
			for (int j = 0; j < 4; j++){
				if (i*4 + j < seqLen){
					//System.out.println(bytes[i+4] + " " + (((bytes[i] >> (3-j)*2)) & 0x03));
					switch (((((char)bytes[i] >> (3-j)*2)) & 0x03)){
					case 0:
						sequence.append("G");
						break;
					case 1:
						sequence.append("A");
						break;
					case 2:
						sequence.append("C");
						break;
					case 3:
						sequence.append("T");
						break;
					default:
						break;
					}
				}
			}
		}
		return sequence.toString();
	}
	
	private static char[] intToCharArray(int integer){
		char[] ret = new char[4];
		ret[0] = (char)(integer & 0x00FF);
		ret[1] = (char)((integer >> 8) & 0x000000FF);
		ret[2] = (char)((integer >> 16) & 0x000000FF);
		ret[3] = (char)((integer >> 24) & 0x000000FF);
		return ret;
	}
	
	public static int byteArrayToInt(byte[] bytes){
		int ret = 0, tmp;
		if (debug) System.out.println((int)bytes[0]+" BA2I");
		for (int i = 0; i < 4; i++){
			if ((int)bytes[i] < 0){
				tmp = (256 + (int)bytes[i]) << (i*8);
				ret += tmp;
				//System.out.println((int)bytes[i]+" "+tmp);
			}
			else{
				ret += (int)bytes[i] << (i*8);
				//System.out.println((int)bytes[i]+" "+((int)bytes[i] << (i*8)));
			}
		}
		return ret;
	}
	private static int charArrayToInt(char[] chars){
		int ret = 0;
		for (int i = 0; i < 4; i++){
			ret += chars[i] << (8*i);
		}
		return ret;
	}

	public static void main(String args[]){
		try{
			String sequence = "CAACACACCCACACCACAAAACCTCTTCTTACCACCACTTTTACCTTACACCAACCACCCTACTTTCCCCTACCTCTTTAACTCTCTCCCCACACTTCTACAAAACCCACTTTCCTCCCCTACCCTCTCCACACTTACTT";
			String filename = DataProperties.getInstance().getResultsDir()+"/testCompress.txt";
			System.out.println(sequence);
			StringBuffer byteStrBuff = new StringBuffer();

			byteStrBuff.append(compressSequenceIntoString(sequence));
			DataWriter.writeStringToFile(filename, byteStrBuff.toString()+"\n");

			FileReader freader = null;
			BufferedReader reader = null;
			
			FileInputStream fis = null;
			BufferedInputStream bis = null;
			DataInputStream dis = null;
			
			String byteStr = "", virtualSequence;
			try{
				/*freader = new FileReader(filename);
			    reader = new BufferedReader(freader);*/
			    
				fis = new FileInputStream(filename);
				bis = new BufferedInputStream(fis);
				dis = new DataInputStream(bis);
				
				//virtualSequence = GACTCompressor.decompressSequence(dis);
				//virtualSequence = GACTCompressor.decompressSequence(dis);
				
	        	byte[] compressedSequence = GACTCompressor.getByteSequence(dis);
	        	virtualSequence = GACTCompressor.decompressSequence(compressedSequence);

				
			    //virtualSequence = GACTCompressor.decompressSequence(reader);
			    //byteStr = compressSequenceIntoString(sequence);
			    //virtualSequence = decompressSequence(byteStr.toCharArray());
			    //byteStr = reader.readLine();
			} catch (IOException ioe){
				throw new Exception(ioe);
			} finally{
				if (reader != null){
					reader.close();
				}
				if (freader != null){
					freader.close();
				}
			}
			System.out.println(byteStr);
			//String decompressed = decompressSequence(byteStr.getBytes());
			//System.out.println("Decompressed: "+decompressed);
			System.out.println("Decompressed: "+virtualSequence);
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
}
