package org.cn3.ns.generation;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.ns.data.DataConverter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.motif.analysis.MotifFinder;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappingUtils;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Generates a FASTA file containing a number of random tree-sequences based on the distribution of lengths
 * and optionally C-Count in an original "real" FASTA file (of neurite-derived sequences).
 * Random sequences can be generated with several possible tree generation methods. Current options (using -b/--baseline) include:
 * 1 constrained by real sequence lengths, generated using string growth
 * 2 constrained by real sequence lengths, generated using terminal branch growth 
 * 3 constrained by real sequence lengths, generated using Harding method (uniform distribution of tree shapes)
 * 5 constrained by real sequence lengths and monomer frequencies
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.generation.CreateRandomDBWithDistribution -i C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS -b 3 -s 1
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.generation.CreateRandomDBWithDistribution -i C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS -b 5 -s 1
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.generation.CreateRandomDBWithDistribution -i /homes/tgillett/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS -b 5 -s 1
 */
public class CreateRandomDBWithDistribution {

	public static final String PARAM_INPUT = "input";
	public static final String PARAM_DATABASE_NAME = "database-name";
	public static final String PARAM_SET_NUM = "set-num";

	private static String traversal;
	private static String order;
	private static int baseline;
	private static File originalFasta;
	private static int setNum;
	private static File outfile;
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			
			boolean traverse = false;
			if (config.getString(NSConstants.PARAM_TRAVERSAL) != null){
				traverse = true;
			}
			traversal = config.getString(NSConstants.PARAM_TRAVERSAL);
			order = config.getString(NSConstants.PARAM_ORDER);
			baseline = config.getInt(NSConstants.PARAM_BASELINE);
			originalFasta = new File(config.getString(PARAM_INPUT));
			setNum = config.getInt(PARAM_SET_NUM);

			Pattern p = Pattern.compile("(.*)\\.fasta");
			Matcher m = p.matcher(originalFasta.getName());
			if (m.matches()){
				outfile = new File(originalFasta.getParent()+"/"+m.group(1)
						+"-randomsB"+baseline+"_"+setNum+".fasta");
			}
			else{
				throw new Exception("First parameter not a Fasta file");
			}

			String randomDatabase = null;
			Map<Integer,Integer> lengthDistribution;
			Map<Integer,Map<String,Integer>> cCountDistribution;
			switch (baseline){
			case NSConstants.DISTRIBUTION_SEQUENCE_GROWTH:
				lengthDistribution = getLengthDistributionFromFasta(originalFasta);
				randomDatabase = generateRandomDatabaseByStringGrowth(
						lengthDistribution, traverse, traversal, order);
				break;
			case NSConstants.DISTRIBUTION_TERMINAL_GROWTH:
			case NSConstants.DISTRIBUTION_UNIFORM:
				lengthDistribution = getLengthDistributionFromFasta(originalFasta);
				randomDatabase = generateRandomDatabaseOrdered(
						lengthDistribution, baseline, traversal, order);
				break;
			case NSConstants.DISTRIBUTION_RESIDUE_FREQ_CONSTRAINED:
				cCountDistribution = getLengthAndMonomerCountsFromFasta(originalFasta);
				randomDatabase = generateMonomerConstrainedRandomDatabaseOrdered(
						cCountDistribution, baseline, traversal, order);
				break;
			default:
				throw new Exception("Unsupported random generation method: "+baseline);
			}

			DataWriter.writeStringToFile(outfile, randomDatabase);
			System.out.println("Wrote file "+outfile);
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static Map<Integer,Integer> getLengthDistributionFromFasta(File originalFastaFile)
	throws Exception{

		int count;
		Integer length;
		Map<Integer,Integer> distribution = new HashMap<Integer,Integer>();
		
		FastaReader fr = new FastaReader(originalFastaFile);

		while (fr.next()) {
			count = 1;
			length = new Integer(fr.getSequenceLength());
			if (distribution.containsKey(length)){
				count = distribution.get(length);
				count++;
			}
			distribution.put(length, count);
		}
		return distribution;
	}

	private static Map<Integer,Map<String,Integer>> getLengthAndMonomerCountsFromFasta(File originalFastaFile)
	throws Exception{

		Integer cCount;
		Integer length;
		Map<Integer,Map<String,Integer>> mapByLength = new HashMap<Integer,Map<String,Integer>>();
		Map<String,Integer> mapBySequence;
		
		FastaReader fr = new FastaReader(originalFastaFile);

		while (fr.next()) {
			cCount = (Integer)MotifFinder.findMotifs(fr.getSequence(), 1).get("C");
			
			length = new Integer(fr.getSequenceLength());
			if (!mapByLength.containsKey(length)){
				mapByLength.put(length, new HashMap<String,Integer>());
			}
			mapBySequence = mapByLength.get(length);
			mapBySequence.put(fr.getSequenceName(), cCount);
		}
		return mapByLength;
	}

	private static String generateRandomDatabaseByStringGrowth(Map<Integer,Integer> distribution, 
			boolean traverse, String traversal, String order)
	throws Exception{
		StringBuffer sb = new StringBuffer();
		String sequence;
		int length;
		int count, sequenceCount = 0, gcCount = 0;
		int largest = getMax(distribution);
		int smallest = getMin(distribution);
		BinaryTreeNode headNode;
		
		while (!distribution.isEmpty()){
			sequence = NodeSequenceBinaryTreeGenerator
				.generateRandomBinaryTreeStringByLengths(smallest,largest);
			length = sequence.length();
			if (distribution.containsKey(length)){
				count = distribution.get(length)-1;
				if (count == 0){
					distribution.remove(length);
					if (length == largest){
						largest = getMax(distribution);
					}
					else if (length == smallest){
						smallest = getMin(distribution);
					} 
					if (distribution.keySet().size() % 10 == 0
							|| distribution.keySet().size() <= 10){
						System.out.println("length "+length+" complete; "
							+distribution.keySet().size()+" remain");
					}
				}
				else{
					distribution.put(length, new Integer(count));
				}
				if (traverse){
					headNode = DataConverter.fastaStringToTree(sequence);
					sequence = MappingUtils.traverseTreeToString(headNode,NSConstants.DEFAULT_MAPPING_MODE,order);
				}
				sb.append(">R"+(++sequenceCount)+"L"+length+"\r\n");
				sb.append(NeuriteSequenceStringUtils.insertBreaksAtLength(sequence,80));
				sb.append("\r\n");
			}
			else{
				gcCount++;
			}
			sequence = null;
			
			if (gcCount == 100){
				System.gc();
				gcCount = 0;
			}
		}
		
		return sb.toString();
	}

	private static String generateRandomDatabaseOrdered(Map<Integer,Integer> distribution, 
			int baseline, String traversal, String order)
	throws Exception{
		StringBuffer sb = new StringBuffer();
		String sequence;
		int sequenceCount = 0;
		int count, degree;
		BinaryTreeNode headNode;
		
		for (int length : distribution.keySet()){
			degree = length+1;
			count = distribution.get(length).intValue();
			for (int i = 0; i < count; i++){
				switch(baseline){
				case NSConstants.DISTRIBUTION_TERMINAL_GROWTH:
					headNode = NodeSequenceBinaryTreeGenerator.generateTerminalGrowthBinaryTreeByDegree(degree);
					break;
				case NSConstants.DISTRIBUTION_UNIFORM:
					headNode = NodeSequenceBinaryTreeGenerator.generateRandomHardingTreeByDegree(degree);
					break;
				default:
					return null;
				}
				sequence = MappingUtils.traverseTreeToString(
						headNode, NSConstants.MAPPING_MODE_TOPO_ACT, order);
				sb.append(">R"+(++sequenceCount)+"L"+length+"\r\n");
				sb.append(NeuriteSequenceStringUtils.insertBreaksAtLength(sequence,80));
				sb.append("\r\n");
			}
		}
		
		return sb.toString();
	}

	private static String generateMonomerConstrainedRandomDatabaseOrdered(Map<Integer,Map<String,Integer>> distribution, 
			int baseline, String traversal, String order)
	throws Exception{
		StringBuffer sb = new StringBuffer();
		String sequence;
		int sequenceCount = 0;
		int cCount, degree;
		BinaryTreeNode headNode;
		Map<String,Integer> cCountMap;
		
		for (int length : distribution.keySet()){
			degree = length+1;
			cCountMap = distribution.get(length);
			for (String neuriteName : cCountMap.keySet()){
				cCount = cCountMap.get(neuriteName);
				headNode = NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedNonuniformTree(degree, cCount);
				sequence = MappingUtils.traverseTreeToString(
						headNode, NSConstants.MAPPING_MODE_TOPO_ACT, order);
				sb.append(">R"+(++sequenceCount)+"L"+length+"\r\n");
				sb.append(NeuriteSequenceStringUtils.insertBreaksAtLength(sequence,80));
				sb.append("\r\n");
			}
		}
		
		return sb.toString();
	}

	private static int getMax(Map<Integer,Integer> distribution){
		int max = 0;
		for (int length : distribution.keySet()){
			if (length > max){
				max = length;
			}
		}
		return max;
	}

	private static int getMin(Map<Integer,Integer> distribution){
		int min = 0;
		for (int length : distribution.keySet()){
			if (min == 0 || length < min){
				min = length;
			}
		}
		return min;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_INPUT)
        	.setStringParser(JSAP.STRING_PARSER)
        	.setRequired(true) 
        	.setShortFlag('i') 
        	.setLongFlag(PARAM_INPUT);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_BASELINE)
			.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(true)
			.setDefault(""+NSConstants.DEFAULT_BASELINE)
			.setShortFlag('b') 
			.setLongFlag(NSConstants.PARAM_BASELINE);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_TRAVERSAL)
        	.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(false) 
    		.setDefault(NSConstants.DEFAULT_TRAVERSAL)
    		.setShortFlag('t') 
    		.setLongFlag(NSConstants.PARAM_TRAVERSAL);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
    		.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(false)
    		.setDefault(NSConstants.DEFAULT_ORDER)
    		.setShortFlag('o') 
    		.setLongFlag(NSConstants.PARAM_ORDER);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_DATABASE_NAME)
        	.setStringParser(JSAP.STRING_PARSER)
			.setRequired(false)
			.setDefault("0")
			.setShortFlag('n') 
			.setLongFlag(PARAM_DATABASE_NAME);
        jsap.registerParameter(opt);
    
        opt = new FlaggedOption(PARAM_SET_NUM)
	    	.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(false)
			.setDefault("0")
			.setShortFlag('s') 
			.setLongFlag(PARAM_SET_NUM);
	    jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CreateRandomDBWithDistribution.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
