package org.cn3.ns.generation;

import java.io.File;
import java.util.Random;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.motif.analysis.MotifFinder;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappingUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Generates a FASTA files containing a given number of pairs of sequences generated randomly
 * but constrained by real sequence monomer distributions. Each FASTA file has sequences of 
 * a particular length.
 * 
 * java org.cn3.ns.motif.generation.CreateRandomDBsByLengths -i C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS
 * 
 * On Krasnow Cluster
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.CreateRandomPairsDBByLengths -i /homes/tgillett/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS --output-dir /homes/tgillett/RandomSequenceFiles/
 *
 * For a random, non-tree-based sequence (but still with the given node-type frequencies)
 * java org.cn3.ns.generation.CreateRandomPairsDBByLengths -i C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2.fasta -o LTS --tree false
 */
public class CreateRandomPairsDBByLengths {

	public static final String PARAM_INPUT = "input";
	public static final String PARAM_OUTPUT_DIR = "output-dir";
	public static final String PARAM_NUM_PAIRS = "numPairs";
	public static final String PARAM_TREE_SEQUENCE = "tree";
	public static final String PARAM_SEQUENCE_LENGTH = "length";
	public static final String PARAM_START_LENGTH = "startLength";
	public static final String PARAM_STOP_LENGTH = "stopLength";
	public static final String PARAM_LENGTH_ADD = "lengthAdd";
	public static final String PARAM_LENGTH_MULT = "lengthMult";
	
	private static final int DEFAULT_NUM_PAIRS = 10000;
	private static final int DEFAULT_START_LENGTH = 150;
	private static final int DEFAULT_STOP_LENGTH = 180;
	private static final int DEFAULT_LENGTH_ADD = 1;
	private static final double DEFAULT_LENGTH_MULT = 1;
	private static final long RANDOM_SEED = 1200;

	private static String order;
	private static File originalFasta;
	private static Random rand = new Random(RANDOM_SEED);

	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			
			order = config.getString(NSConstants.PARAM_ORDER);
			originalFasta = new File(config.getString(PARAM_INPUT));
			int seqLen = config.getInt(PARAM_SEQUENCE_LENGTH);
			int numPairs = config.getInt(PARAM_NUM_PAIRS);
			int startLength = config.getInt(PARAM_START_LENGTH);
			int stopLength = config.getInt(PARAM_STOP_LENGTH);
			int lengthAdd = config.getInt(PARAM_LENGTH_ADD);
			double lengthMult = config.getDouble(PARAM_LENGTH_MULT);
			boolean treeSequence = config.getBoolean(PARAM_TREE_SEQUENCE);
			File outputDir = new File(config.getString(PARAM_OUTPUT_DIR));
			if (!outputDir.isDirectory()){
				throw new Exception("Output directory required. "+outputDir+" is not a directory.");
			}
			
			double[] cFrequencies = calculateCFrequencies(originalFasta);
			
			StringBuffer sb;
			
			int cCount;
			if (seqLen > 0){
				startLength = stopLength = seqLen;
				lengthAdd = 1;
				lengthMult = 1;
			}
			for (int sequenceLength = startLength; sequenceLength <= stopLength; sequenceLength=(int)Math.round(sequenceLength*lengthMult)+lengthAdd){
				sb = new StringBuffer();
				for (int pair = 1; pair <= numPairs; pair++){
					sb.append(">Pair"+pair+"_A\n");
					cCount = selectCCount(cFrequencies, sequenceLength);
					if (treeSequence){
						sb.append(MappingUtils.traverseTreeToString(
							NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedNonuniformTree(sequenceLength+1, cCount),
							NSConstants.MAPPING_MODE_TOPO_ACT,order)+"\n");
					}
					else{
						sb.append(generateRandomSequence(sequenceLength, cCount)+"\n");
					}
					sb.append(">Pair"+pair+"_B\n");
					cCount = selectCCount(cFrequencies, sequenceLength);
					if (treeSequence){
						sb.append(MappingUtils.traverseTreeToString(
							NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedNonuniformTree(sequenceLength+1, cCount),
							NSConstants.MAPPING_MODE_TOPO_ACT,order)+"\n");
					}
					else{
						sb.append(generateRandomSequence(sequenceLength, cCount)+"\n");
					}
				}
				// Write sequences to file
				String filename = outputDir.getAbsolutePath()+"/BackgroundConstrained";
				if (treeSequence){
					filename += order;
				}
				else{
					filename += "NonTree";
				}
				filename += "-Length"+sequenceLength+"_Pairs"+numPairs+".fasta";
				
				DataWriter.writeStringToFile(
						filename,
						sb.toString());
			}
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static String generateRandomSequence(int sequenceLength, int cCount){
		int aCount = (sequenceLength - cCount)/2;
		int tCount = (sequenceLength - cCount)/2 + 1;
		
		String seq = "";
		int r;
		while (cCount+aCount+tCount > 0){
			r = (int)(Math.random()*(cCount+aCount+tCount));
			if (r < cCount){
				seq += "C";
				cCount--;
			}
			else if (r < cCount+aCount){
				seq += "A";
				aCount--;
			}
			else{
				seq += "T";
				tCount--;
			}
		}
		return seq;
	}
		
	private static int selectCCount(double[] cFrequencies, int sequenceLength){
		// A and T node count must be odd, therefore length-C must be odd
		int cCount;
		do{
			cCount = (int)Math.round(
					cFrequencies[rand.nextInt(cFrequencies.length)]*sequenceLength);
		} while((sequenceLength-cCount)%2 == 1);
		return cCount;
	}
	
	private static double[] calculateCFrequencies(File originalFastaFile) throws Exception{
		int numSequences = FastaReader.getNumSequences(originalFastaFile);
		FastaReader fr = new FastaReader(originalFastaFile);
		double[] frequencies = new double[numSequences];
		int ind = 0;
		while (fr.next()){
			frequencies[ind] = ((Double)
					MotifFinder.findMotifFrequencies(fr.getSequence(),1).get("C"))
					.doubleValue();
			ind++;
		}
		return frequencies;
	}
	

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_INPUT)
        	.setStringParser(JSAP.STRING_PARSER)
        	.setRequired(true) 
        	.setShortFlag('i') 
        	.setLongFlag(PARAM_INPUT);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
    		.setStringParser(JSAP.STRING_PARSER)
    		.setRequired(false)
    		.setDefault(NSConstants.DEFAULT_ORDER)
    		.setShortFlag('o') 
    		.setLongFlag(NSConstants.PARAM_ORDER);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SEQUENCE_LENGTH)
		.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault("0")
		.setShortFlag('l') 
		.setLongFlag(PARAM_SEQUENCE_LENGTH);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_OUTPUT_DIR)
        	.setStringParser(JSAP.STRING_PARSER)
			.setRequired(false)
			.setDefault("G:/RandomSequenceFiles/")
			.setShortFlag('O') 
			.setLongFlag(PARAM_OUTPUT_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_NUM_PAIRS)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_NUM_PAIRS)
		.setShortFlag('n') 
		.setLongFlag(PARAM_NUM_PAIRS);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_TREE_SEQUENCE)
    	.setStringParser(JSAP.BOOLEAN_PARSER)
		.setRequired(false)
		.setDefault("true")
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_TREE_SEQUENCE);
        jsap.registerParameter(opt);
        
        opt = new FlaggedOption(PARAM_START_LENGTH)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_START_LENGTH)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_START_LENGTH);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_STOP_LENGTH)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_STOP_LENGTH)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_STOP_LENGTH);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_LENGTH_ADD)
    	.setStringParser(JSAP.INTEGER_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_LENGTH_ADD)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_LENGTH_ADD);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_LENGTH_MULT)
    	.setStringParser(JSAP.DOUBLE_PARSER)
		.setRequired(false)
		.setDefault(""+DEFAULT_LENGTH_MULT)
		.setShortFlag(JSAP.NO_SHORTFLAG) 
		.setLongFlag(PARAM_LENGTH_MULT);
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CreateRandomPairsDBByLengths.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
