package org.cn3.ns.generation;

import java.io.File;
import java.text.NumberFormat;
import java.util.Map;
import java.util.Stack;
import java.util.TreeMap;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.data.SWCReader;
import org.cn3.morpho.data.SWCWriter;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.Soma;
import org.cn3.morpho.domain.Soma.ArborType;
import org.cn3.morpho.domain.SwcTreeNode;
import org.cn3.morpho.utils.StringUtils;
import org.cn3.ns.nodeencoding.NodeEncoder;
import org.cn3.ns.nodeencoding.Topological1stOrderEncoder;
import org.cn3.ns.resources.NSConstants;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Removes segments extending from any segment that falls outside of some z-distance from the soma.
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.generation.SwcZDistanceCutter -s "/home/tgillett/swcFiles/NM/ascoli+buzsaki/CNG version/ca3b-N2.CNG.swc" -a axon -o /home/tgillett/zCuts/ -c 3
 *
 */
public class SwcZDistanceCutter {

	private static final String PARAM_SWC_FILE = "swc-file";
	private static final String PARAM_ARBOR = "arbor";
	private static final String PARAM_OUTPUT_DIRECTORY = "output-dir";
	private static final String PARAM_Z_INTERVAL = "z-interval";
	private static final String PARAM_NUM_CUTS = "num-cuts";
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			String swcFilename = config.getString(PARAM_SWC_FILE);
			if (!swcFilename.endsWith(".swc") && !swcFilename.endsWith(".SWC")){
				throw new Exception("SWC file must have .swc extension");
			}
			File swcFile = new File(swcFilename);
			if (!swcFile.exists()){
				throw new Exception("SWC file does not exist ("+swcFilename+")");
			}
			if (swcFile.isDirectory()){
				throw new Exception("SWC file is a directory ("+swcFilename+")");
			}
			ArborType arborType = ArborType.find(config.getString(PARAM_ARBOR));
			
			String outputDir = config.getString(PARAM_OUTPUT_DIRECTORY);
			if (StringUtils.isEmpty(outputDir)){
				outputDir = swcFile.getParent();
			}
			float zInterval = config.getFloat(PARAM_Z_INTERVAL);
			int numCuts = config.getInt(PARAM_NUM_CUTS);

			// Read in SWC
			Soma<SwcTreeNode> soma = SWCReader.convertSwcToSoma(swcFile);
			SwcTreeNode tree;
			switch(arborType){
			case Axon:
				tree = soma.getAxonalTree().getTreeHead();
				break;
			case ApicalDendrite:
				tree = soma.getApicalDendriteTree().getTreeHead();
				break;
			case Dendrite:
				tree = soma.getBasalDendriteTree().getTreeHead();
				break;
			default:
				throw new Exception("Unknown arbor type: "+config.getString(PARAM_ARBOR));
			}
			
			// If no zInterval is given, calculate it based on z range of tree 
			if (zInterval == 0){
				zInterval = calculateZInterval(tree, numCuts);
			}
			
			String neuronName = swcFile.getName().replaceAll("(\\.CNG)?\\.swc","");
			String neuriteName = neuronName + "-"+arborType;

			Map<Float,SwcTreeNode> cutTrees = cutTreeMultiple(tree,zInterval,numCuts);
			StringBuffer fastaBuffer = new StringBuffer();
			NumberFormat nf = NumberFormat.getNumberInstance();
			nf.setMaximumFractionDigits(0);
			nf.setGroupingUsed(false);
			String cutName, outputFilename;
			SwcTreeNode cutTree;
			BinaryTreeNode binaryCutTree;
			NodeEncoder encoder = new Topological1stOrderEncoder();
			for (float zCutoff : cutTrees.keySet()){
				cutTree = cutTrees.get(zCutoff);
				cutName = neuriteName + "_ZCut-"+nf.format(zCutoff);
				
				// Write to SWC file
				outputFilename = outputDir + "/" + cutName + ".swc";
				SWCWriter.writeSwcTreeToFile(new File(outputFilename), cutTree);
				
				// Convert to fasta and append to file
				fastaBuffer.append("> "+cutName+"\n");
				binaryCutTree = SWCReader.convertToBinaryTreeNode(cutTree);
				fastaBuffer.append(
						encoder.encodeTree(binaryCutTree, "STL")+
						"\n");
			}
			BinaryTreeNode originalBinaryTree =
				SWCReader.convertToBinaryTreeNode(tree);
			fastaBuffer.append("> " + neuriteName + "\n"+
					encoder.encodeTree(originalBinaryTree, "STL")+
					"\n");
			DataWriter.writeStringToFile(outputDir + "/ZCuts-"+numCuts+"_" + neuriteName + ".fasta", fastaBuffer.toString());
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	/**
	 * Determine zInterval based on number of cuts and size of tree in z-dimension
	 * @param tree
	 * @param numCuts
	 * @return
	 */
	private static float calculateZInterval(SwcTreeNode tree, int numCuts){
		// Find min and max z values
		double somaZ = tree.getZ(), minZ = somaZ, maxZ = somaZ;
		Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
		stack.push(tree);
		SwcTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (node.getZ() < minZ){
				minZ = node.getZ();
			}
			if (node.getZ() > maxZ){
				maxZ = node.getZ();
			}
			if (node.hasChildren()){
				stack.addAll(node.getChildren());
			}
		}
		
		// Calculate interval with numCuts between soma and farthest of min and max
		double belowSoma = somaZ - minZ, aboveSoma = maxZ - somaZ;
		double greatestDist = Math.max(belowSoma,aboveSoma);
		return((float)greatestDist/(numCuts+1));
	}
	
	public static Map<Float,SwcTreeNode> cutTreeMultiple(SwcTreeNode tree, float zInterval, int numCuts){
		Map<Float,SwcTreeNode> cutTrees = new TreeMap<Float,SwcTreeNode>();
		for (int i = 1; i <= numCuts; i++){
			float zCutoff = zInterval*i;
			cutTrees.put(zCutoff,cutTree(tree, zCutoff));
		}
		return cutTrees;
	}

	public static SwcTreeNode cutTree(SwcTreeNode tree, float zCutoff){
		SwcTreeNode newTree = tree.copy();
		Stack<SwcTreeNode> stack = new Stack<SwcTreeNode>();
		stack.push(newTree);
		double top = newTree.getZ() + zCutoff, bottom = newTree.getZ() - zCutoff;
		SwcTreeNode node;
		while (!stack.isEmpty()){
			node = stack.pop();
			if (node.getZ() > top || node.getZ() < bottom){
				node.detach();
			}else if (node.hasChildren()){
				stack.addAll(node.getChildren());
			}
		}
		return newTree;
	}
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_SWC_FILE)
			.setStringParser(JSAP.STRING_PARSER)
			.setDefault(NSConstants.DEFAULT_DATASET)
			.setRequired(false) 
			.setShortFlag('s') 
			.setLongFlag(PARAM_SWC_FILE);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_ARBOR)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true) 
		.setShortFlag('a') 
		.setLongFlag(PARAM_ARBOR);
        jsap.registerParameter(opt);


        opt = new FlaggedOption(PARAM_OUTPUT_DIRECTORY)
		.setStringParser(JSAP.STRING_PARSER)
		.setDefault("")
		.setRequired(false) 
		.setShortFlag('o') 
		.setLongFlag(PARAM_OUTPUT_DIRECTORY);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_Z_INTERVAL)
			.setStringParser(JSAP.FLOAT_PARSER)
			.setRequired(false)
			.setDefault("0")
			.setShortFlag('z') 
			.setLongFlag(PARAM_Z_INTERVAL);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_NUM_CUTS)
			.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(false)
			.setDefault("3")
			.setShortFlag('c') 
			.setLongFlag(PARAM_NUM_CUTS);
        jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + SwcZDistanceCutter.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}

}
