package org.cn3.ns.motif.analysis;

import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.domain.DistributionBean;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.motif.generation.ConstraintsDataReader;
import org.cn3.ns.motif.generation.MotifConstants;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappedBinaryTreeUtils;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Produces 2 files:
 * 	1) Standard deviations for each k-mer of a given neurite's surrogate set (k-mer based on k-1-mer constrained surrogates)
 *  2) Number of distinct k-mer counts in k-mer count distribution within each neurite's surrogate set
 *
 * 12/23/13, set 1 as baseline, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.BaselineKmerVarianceAnalyzer -o STL --startK 0 -C 4 -s 40 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/
 *
 * 12/27/13, set 1 as baseline, LTS given source traversal STL for k=3-5 (constraints 2-4)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.BaselineKmerVarianceAnalyzer -o LTS --startK 2 -C 4 --sourceTraversal STL -s 40 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/
 *
 * 
 */
public class BaselineKmerVarianceAnalyzer {
	
	private static final String PARAM_MAX_CONSTRAINT = "max-constraint";
	private static final String PARAM_START_SET = "startSet";
	private static final String PARAM_START_K = "startK";

	private static final String PARAM_OUTPUT_DIR = "outputDir";
	private static final String PARAM_SURROGATE_SET = "surrogateSet";
	
	private static final int DEFAULT_MAX_CONSTRAINT = 3;

	private static double threshold = 0;
	private static int kmerThresholdType = MotifConstants.KMER_THRESHOLD_NONE;
	private static int seqThresholdType = MotifConstants.SEQ_THRESHOLD_WHOLE;
	private static String thresholdTypeString = null;
	private static String dataset = "NM";
	private static int method;
	private static String order;
	private static int numInSet;
	private static int mapping = NSConstants.MAPPING_MODE_TOPO_ACT;
	private static int maxConstraint = DEFAULT_MAX_CONSTRAINT;
	private static int maxCheck;
	private static int startK = 0;
	private static int kPlus = 1;
	private static String sourceTraversal;
	private static boolean retraverse = false;
	
	private static List<String> sequenceNames;
	private static List<String>[] removedSequences;
	private static List<String>[] motifsByLength;

	@SuppressWarnings("unchecked")
	public static void main(String args[]) throws Exception{
		JSAPResult config = readParameters(args);
		dataset = config.getString(NSConstants.PARAM_DATASET);
		thresholdTypeString = MotifConstants.getThresholdTypeString(kmerThresholdType, seqThresholdType);
		// Surrogate generate method
		method = config.getInt(NSConstants.PARAM_RANDOM_METHOD);
		// Traversal order
		order = config.getString(NSConstants.PARAM_ORDER);
		// Set number for surrogates
		int surrogateSet = config.getInt(PARAM_SURROGATE_SET);
		// Number of surrogate sequences for each neurite and k
		numInSet = config.getInt(MotifConstants.PARAM_SET_NUM);
		// Constraint from which to start analysis
		startK = config.getInt(PARAM_START_K);
		//startSet = config.getInt(PARAM_START_SET);
		
		// Constraint on which to stop analysis
		maxConstraint = config.getInt(PARAM_MAX_CONSTRAINT);
		maxCheck = maxConstraint+kPlus;
		String outputDir = config.getString(PARAM_OUTPUT_DIR);

		sourceTraversal = order;
		if (config.userSpecified(NSConstants.PARAM_SOURCE_TRAVERSAL)){
			sourceTraversal = config.getString(NSConstants.PARAM_SOURCE_TRAVERSAL);
			if (!sourceTraversal.equals(order)){
				retraverse = true;
			}
		}

		System.out.println("Initializing...");
		
		removedSequences = new List[maxCheck];
		for (int i = 0; i < maxCheck; i++){
			removedSequences[i] = new ArrayList<String>();
		}

		// Get sequence names and map them to unique ids
		sequenceNames = NSDataReader.loadSequenceList(mapping);
		Map<String,Integer> sequenceIdMap = new HashMap<String,Integer>();
		for (int j = 0; j < sequenceNames.size(); j++){
			sequenceIdMap.put(sequenceNames.get(j), new Integer(j));
		}
		
		// Initializing motifs
		motifsByLength = new List[maxCheck+1];
		for (int i = 1; i <= maxCheck; i++){
			if (i > 1){
				motifsByLength[i] = NeuriteSequenceStringUtils.generatePossibleACTKmers(i,order);
			}
			else{
				motifsByLength[1] = new ArrayList<String>();
				motifsByLength[1].add("C");
			}
		}

		// Initializing string buffers for output
		StringBuffer[] sdSb = new StringBuffer[maxCheck];
		StringBuffer[] numValsSb = new StringBuffer[maxCheck];
		StringBuffer[] zeroBoundSb = new StringBuffer[maxCheck];
		for (int k = 0; k <= maxConstraint; k++){
			sdSb[k] = new StringBuffer();
			numValsSb[k] = new StringBuffer();
			zeroBoundSb[k] = new StringBuffer();
			sdSb[k].append("Sequence");
			numValsSb[k].append("Sequence");
			zeroBoundSb[k].append("Sequence");
			for (int j = 0; j < motifsByLength[k+kPlus].size(); j++){
				sdSb[k].append(","+motifsByLength[k+kPlus].get(j));
				numValsSb[k].append(","+motifsByLength[k+kPlus].get(j));
				zeroBoundSb[k].append(","+motifsByLength[k+kPlus].get(j));
			}
			sdSb[k].append("\n");
			numValsSb[k].append("\n");
			zeroBoundSb[k].append("\n");
		}

		// Load up virtual sequences (key by sequenceId) from the N virtual sequence sets (standard N=20)
		System.out.println("Loading virtual sequences");
		Map<Integer,Map<String,List<String>>> kBaselineSequenceMap;
		// Only using newer storage method now
		kBaselineSequenceMap = ConstraintsDataReader.loadSurrogates(
				surrogateSet, sourceTraversal, method, numInSet, false);

		// Determine whether nmers are constrained
		Map<String,Number> baselineKmerCountMap;
		Map<String,List<String>> baselineSequenceMap;
		List<String> baselineSequences;
		List<Map<String,Number>> virtualSequenceKMerCountMaps;
		String kmer;
		DistributionBean baselineKmerCountDist;
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(3);
		
		String sequence;
		// Loop through constraints (by k)
		for (int constraint = startK; constraint <= maxConstraint; constraint++){
			// Load list of virtual sequences associated with real sequence and k-constraint
			baselineSequenceMap = kBaselineSequenceMap.get(constraint);
			
			// Loop through sequences and k to calculate whether a given real sequence was constrained
			for (String sequenceName : baselineSequenceMap.keySet()){
				baselineSequences = baselineSequenceMap.get(sequenceName);
				//System.out.println(sequenceName+" ");
				/*System.out.println(kVirtualSequenceMap.get(startK).get(sequenceId));
				System.out.println(kVirtualSequenceMap.get(startK).get(sequenceId).get(0));
				System.out.println(kVirtualSequenceMap.get(startK).get(sequenceId).get(0).length());*/
	
				//System.out.println(constraint+" " +sequenceName);
				//System.out.println("  "+kVirtualSequenceMap.get(constraint).get(sequenceName).size());
				if (!kBaselineSequenceMap.get(constraint).containsKey(sequenceName)){
					continue;
				}
				
				if (baselineSequences == null){
					System.out.println("No sequences for "+sequenceName+" / "+constraint);
				}
				else if (baselineSequences.size() == 1){
					System.out.println("Only one sequence for "+sequenceName+" / "+constraint);
				}
				else{
					//System.out.println(k+" "+sequenceId+" "+sequenceNames.get(sequenceId.intValue()));
					
					// Calculate and assemble n-mer counts of k+1 (or kPlus, for which default is 1) across the virtual sequence sets
					virtualSequenceKMerCountMaps = new ArrayList<Map<String,Number>>();
					for (int i = 0; i < baselineSequences.size(); i++){
						sequence = baselineSequences.get(i);
						if (retraverse) sequence = MappedBinaryTreeUtils.reorderSequence(sequence, order);
						virtualSequenceKMerCountMaps.add(MotifFinder.findMotifs(sequence,constraint+kPlus));
					}

					// Start new line in output file
					sdSb[constraint].append(sequenceName);
					numValsSb[constraint].append(sequenceName);
					zeroBoundSb[constraint].append(sequenceName);
					
					/** Loop through motifs at k+1 and determine the real sequence's
					 *  kmer count percentile rank within the virtual sequence distribution
					 */
					for (int i = 0; i < motifsByLength[constraint+kPlus].size(); i++){
						// Get distribution of k+1-mer counts across virtual sequences
						kmer = motifsByLength[constraint+kPlus].get(i);
						baselineKmerCountDist = new DistributionBean();
						for (int j = 0; j < baselineSequences.size(); j++){
							baselineKmerCountMap = virtualSequenceKMerCountMaps.get(j);
							if (baselineKmerCountMap.containsKey(kmer)){
								baselineKmerCountDist.addData(baselineKmerCountMap.get(kmer));
							}
							else{
								baselineKmerCountDist.addData(0);
							}
						}
						
						sdSb[constraint].append(","+nf.format(baselineKmerCountDist.getStandardDeviation()));
						numValsSb[constraint].append(","+nf.format(baselineKmerCountDist.getUniqueValues().size()));
						zeroBoundSb[constraint].append(","+(baselineKmerCountDist.getPercentileRankForValue(0)>0.025 ? 1 : 0));
					}
					sdSb[constraint].append("\n");
					numValsSb[constraint].append("\n");
					zeroBoundSb[constraint].append("\n");
					
					/** 
					 * Determine whether this sequence is constrained for k based on its distribution 
					 * of percentile ranks (given its virtual sequence set) across k-mers. Percentile
					 * ranks should fall along a uniform distribution, so a Wilcox test comparing the
					 * distribution of k-mer percentile ranks with randomly generated values should
					 */
					// Off-loading this process to R. Easier to do analysis on subsets.
				}
			}
		}
		
 		// Write results to a csv file
		String suffix = "-"+dataset+"-"+numInSet+"Sets";
		if (surrogateSet > -1){
			suffix += "-"+surrogateSet;
		}
		suffix += "-Size"+order+"-B"+method+sourceTraversal+"-T"+threshold+thresholdTypeString;

		writeToFile(outputDir+"/SD",suffix,sdSb);
		writeToFile(outputDir+"/NumVals",suffix,numValsSb);
		writeToFile(outputDir+"/ZeroBound",suffix,zeroBoundSb);
	}
	
	private static void writeToFile(String baseFilename, String suffix, StringBuffer[] sb) throws IOException{
		String filename;
		for (int k = startK; k <= maxConstraint; k++){
			filename = baseFilename + suffix + "_K"+k+"-"+(k+kPlus)+".csv";
			DataWriter.writeStringToFile(filename, sb[k].toString());
		}
	}
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = MotifConstants.getParameters();
        
        FlaggedOption opt;

        opt = new FlaggedOption(NSConstants.PARAM_DATASET)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(NSConstants.DEFAULT_DATASET)
        .setShortFlag('D') 
        .setLongFlag(NSConstants.PARAM_DATASET);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_MAX_CONSTRAINT)
        .setStringParser(JSAP.INTEGER_PARSER)
	       .setRequired(false)
	       .setDefault("2")
	       .setShortFlag('C') 
	       .setLongFlag(PARAM_MAX_CONSTRAINT);
	    jsap.registerParameter(opt);
	    
	    opt = new FlaggedOption(PARAM_START_SET)
	    	.setStringParser(JSAP.INTEGER_PARSER)
	        .setRequired(false)
	        .setDefault("1")
	        .setShortFlag(JSAP.NO_SHORTFLAG) 
	        .setLongFlag(PARAM_START_SET);
	    jsap.registerParameter(opt);
	    
	    opt = new FlaggedOption(PARAM_START_K)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_START_K);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_DEBUG)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(NSConstants.PARAM_DEBUG);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_OUTPUT_DIR)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("C:/Documents and Settings/gillette/My Documents/research/MotifAnalysis/PercentileRanks/")
        .setShortFlag('O') 
        .setLongFlag(PARAM_OUTPUT_DIR);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_SURROGATE_SET)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("-1")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_SURROGATE_SET);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_SOURCE_TRAVERSAL)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(""+NSConstants.DEFAULT_ORDER)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(NSConstants.PARAM_SOURCE_TRAVERSAL);
	    jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + BaselineKmerVarianceAnalyzer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

}
