package org.cn3.ns.motif.analysis;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.cn3.morpho.data.DataWriter;
import org.cn3.morpho.domain.DistributionBean;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.motif.generation.ConstraintsDataReader;
import org.cn3.ns.motif.generation.MotifConstants;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappedBinaryTreeUtils;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Gets percentile rank of each real sequence/kmer within constrained surrogates for k=1 to MaxK (constraints 0 to MaxK-1).
 * Replaces ConstraintAnalyzer with a simpler method that also can run on other (non-NMO) datasets.
 * If running on surrogates, reads in new set of sequences for each k in order to see effect
 * of constraint on next level.
 *
 * Parameters
 * -o/--order			Traversal order (STL|LTS)
 * --startK 1			First/lowest constrained run through analysis
 * -O/--outputDir		Directory where output will be written		
 * -C/--max-constraint	Top constrained run through analysis
 * -s/--set-num 		Baseline/surrogate set number
 * -m/--method			Method used to generate surrogates (numeric, see DISTRIBUTION_* constants in NSConstants)
 * --minLength			If given, determines the smallest length sequence to analyze
 * --measurementType	Percentile rank (1 - default) or frequency difference (2)
 * --surrogateSet		The number (if it exists) of the surrogate set to use as baseline
 * --baselineDistNA		Sets the output value to -1 (to be read as NA) if the baseline set has no variance and the percentile rank would be 0.5 
 * --useRandomSet		If given, this determines which surrogate set to use in place of real neurites
 * --useRandomMethod	If using surrogate set in place of real neurites, this determines which generation method produced the additional set 
 * --useRandomK			If using surrogate set in place of real neurites, this will determine which k from which to take (what constraint level)		
 * --useRandomN			If using surrogate set in place of real neurites, this determines the number of surrogates per original sequence and constraint (also partially defines the surrogate filename)
 *
 * 4/7/15, NM_SpurLen2D; set 1 as baseline, set 2 as additional surrogate(s), STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o STL --startK 0 -C 4 -s 100 -m 11 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 * Baseline: set1N100, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o STL --startK 0 -C 4 -s 100 -m 11 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 * 
 * NM_SpurLen2D; set 1 as baseline, set 2 as additional surrogates, LTS given source traversal STL for k=3-5 (constraints 2-4)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o LTS --startK 2 -C 4 --sourceTraversal STL -s 100 -m 11 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 * Baseline: set1N100, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o LTS --startK 2 -C 4 --sourceTraversal STL -s 100 -m 11 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 * 
 * 4/2/15, NM_SpurLen2D; set 1 as baseline, set 2 as additional surrogate(s), STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o STL --startK 0 -C 4 -s 100 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 * Baseline: set1N100, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o STL --startK 0 -C 4 -s 100 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 * 
 * NM_SpurLen2D; set 1 as baseline, set 2 as additional surrogates, LTS given source traversal STL for k=3-5 (constraints 2-4)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o LTS --startK 2 -C 4 --sourceTraversal STL -s 100 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 * Baseline: set1N100, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o LTS --startK 2 -C 4 --sourceTraversal STL -s 100 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 20
 *
 * 2/28/14, NM_SpurLen2D; set 1 as baseline, set 2 as additional surrogate(s), STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o STL --startK 0 -C 4 -s 100 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 * Baseline: set1N40, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o STL --startK 0 -C 4 -s 100 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 * 
 * NM_SpurLen2D; set 1 as baseline, set 2 as additional surrogates, LTS given source traversal STL for k=3-5 (constraints 2-4)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o LTS --startK 2 -C 4 --sourceTraversal STL -s 100 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 * Baseline: set1N40, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer --dataset NM_SpurLen2D -o LTS --startK 2 -C 4 --sourceTraversal STL -s 100 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 * 
 * 12/23/13, set 1 as baseline, set 2 as additional surrogate(s), STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL --startK 0 -C 4 -s 40 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 * Baseline: set1N40, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL --startK 0 -C 4 -s 40 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 *
 * 12/27/13, set 1 as baseline, set 2 as additional surrogates, LTS given source traversal STL for k=3-5 (constraints 2-4)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 2 -C 4 --sourceTraversal STL -s 40 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 * Baseline: set1N40, use surrogates instead of real: set 1 as additional surrogates, STL
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 2 -C 4 --sourceTraversal STL -s 40 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomN 1 --outputDir ~/results/PercentileRanks/ --minLength 18
 *
 * 10/5/12, running on set 5 with corrected dimer and trimer-constrained generation methods (to get closer to uniform distribution)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 0 -C 3 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 18
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL --startK 3 -C 3 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 18
 * 
 * Running additional random set in place of real set
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 0 -C 0 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 25 --useRandomSet 6 --useRandomK 0 --useRandomBaseline 8 --useRandomN 1 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 1 -C 1 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 25 --useRandomSet 6 --useRandomK 1 --useRandomBaseline 8 --useRandomN 1 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 2 -C 2 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 25 --useRandomSet 6 --useRandomK 2 --useRandomBaseline 8 --useRandomN 1 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 3 -C 3 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 25 --useRandomSet 6 --useRandomK 3 --useRandomBaseline 8 --useRandomN 1 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL --sourceTraversal LTS --startK 2 -C 2 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 25 --useRandomSet 6 --useRandomK 2 --useRandomBaseline 8 --useRandomN 1
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL --startK 3 -C 3 -s 40 -m 8 --surrogateSet 6 --outputDir ~/results/PercentileRanks/ --minLength 25 --useRandomSet 6 --useRandomK 3 --useRandomBaseline 8 --useRandomN 1
 * 
 * 6/17/12, running on set 2, LTS and STL, with monomer constrained sequences using the new method that ensure a uniform dist. of tree shapes.
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 3 -s 40 -m 8 --surrogateSet 2 --startK 0 --outputDir ~/results/PercentileRanks/ --baselineDistNA T
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 3 -s 40 -m 8 --surrogateSet 2 --startK 0 --outputDir ~/results/PercentileRanks/ --baselineDistNA T
 *
 * Also using set 2 (having uniform dist monomer constrained sequences), but producing relative kmer frequency rather than percentile rank
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 3 -s 40 -m 8 --surrogateSet 2 --startK 0 --outputDir ~/results/PercentileRanks/ --measurementType 2
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 3 -s 40 -m 8 --surrogateSet 2 --startK 0 --outputDir ~/results/PercentileRanks/ --measurementType 2 --sourceTraversal LTS
 * 
 * 
 * Method 7 (dimer assembly)
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 4 -s 20 -T 0.05 -m 7 --startK 1
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 4 -s 20 -T 0.05 -m 7 --startK 1
 * 
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 4 -s 20 -T 0.05 -m 7 --startK 2 --useRandomSet 21 --useRandomK 2
 * 
 * Method 8 (nmer assembly up through 4-mer)
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 4 -s 20 -m 8 --startK 0 --outputDir ~/results/PercentileRanks/
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 4 -s 20 -m 8 --startK 0
 *
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -S 1 -K 1 -o LTS -C 0 -s 20 -m 3 --startK 0 --useRandomSet 22 --useRandomK 0
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 1 -s 20 -m 8 --startK 1 --useRandomSet 22 --useRandomK 1
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 1 -s 20 -m 8 --startK 1 --useRandomSet 22 --useRandomK 1
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 2 -s 20 -m 8 --startK 2 --useRandomSet 22 --useRandomK 2
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 2 -s 20 -m 8 --startK 2 --useRandomSet 22 --useRandomK 2
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 3 -s 20 -m 8 --startK 3 --useRandomSet 22 --useRandomK 3
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 3 -s 20 -m 8 --startK 3 --useRandomSet 22 --useRandomK 3
 * -C is the maxConstraint, default 3. Use 4 for now.
 * 
 * New surrogate set storage
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS -C 4 -s 40 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o STL -C 4 -s 40 -m 8 --surrogateSet 1 --outputDir ~/results/PercentileRanks/
 * 
 * New storage, using old set for random run
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 0 -C 0 -s 40 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomK 0 --outputDir ~/results/PercentileRanks/
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 1 -C 1 -s 40 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomK 1 --outputDir ~/results/PercentileRanks/
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 2 -C 2 -s 40 -m 8 --surrogateSet 1 --useRandomSet 5 --useRandomK 2 --outputDir ~/results/PercentileRanks/
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 3 -C 3 -s 40 -m 8 --surrogateSet 1 --useRandomSet 1 --useRandomK 3 --outputDir ~/results/PercentileRanks/
 *
 * Run random of one baseline on another baseline
 * java org.cn3.ns.motif.analysis.CompleteConstraintAnalyzer -o LTS --startK 2 -C 2 -s 40 -m 3 --surrogateSet 1 --useRandomSet 5 --useRandomK 2 --useRandomBaseline 8 --outputDir ~/results/PercentileRanks/
 * 
 */
public class CompleteConstraintAnalyzer {
	
	private static final String PARAM_MAX_CONSTRAINT = "max-constraint";
	private static final String PARAM_START_K = "startK";
	//private static final String PARAM_USE_TREES_SET = "useTREESSet";
	//private static final String PARAM_BALANCING_FACTOR = "balancingFactor";
	// If given, this determines which surrogate set to use in place of real neurites
	private static final String PARAM_USE_RANDOM_SET = "useRandomSet";
	// If using surrogate set in place of real neurites, this will determine which k from which to take (what constraint level)
	private static final String PARAM_USE_RANDOM_K = "useRandomK";
	
	private static final String PARAM_USE_RANDOM_METHOD = "useRandomMethod";
	 // Number of random sequences per real sequence; required for name of surrogate filename 
	private static final String PARAM_USE_RANDOM_N = "useRandomN";
	private static final String PARAM_OUTPUT_DIR = "outputDir";
	private static final String PARAM_SURROGATE_SET = "surrogateSet";
	private static final String PARAM_MEASUREMENT_TYPE = "measurementType";
	private static final String PARAM_BASELINE_DIST_NA = "baselineDistNA";
	private static final String PARAM_MIN_LENGTH = "minLength";
	
	private static final int MEASURE_PERCENTILE_RANK = 1;
	private static final int MEASURE_FREQUENCY_DIFFERENCE = 2;
	
	private static final int DEFAULT_MAX_CONSTRAINT = 3;
	private static final int DEFAULT_MEASUREMENT_TYPE = MEASURE_PERCENTILE_RANK;

	private static double threshold = 0;
	private static int kmerThresholdType = MotifConstants.KMER_THRESHOLD_NONE;
	private static int seqThresholdType = MotifConstants.SEQ_THRESHOLD_WHOLE;
	private static String thresholdTypeString = null;
	private static String dataset = "NM";
	private static int method;
	private static String order;
	private static int numInSet;
	private static int mapping = NSConstants.MAPPING_MODE_TOPO_ACT;
	private static int maxConstraint = DEFAULT_MAX_CONSTRAINT;
	private static int maxCheck;
	private static int startK = 0;
	private static int useRandomSet = 0;
	private static int useRandomK = -1;
	private static int useRandomMethod = 0;
	private static int useRandomN = 0;
	private static int kPlus = 1;
	private static int measurementType = 1;
	private static String sourceTraversal;
	private static boolean retraverse = false;
	private static int minLength = 0;
	
	private static List<String> sequenceNames;
	private static List<String>[] motifsByLength;

	@SuppressWarnings("unchecked")
	public static void main(String args[]) throws Exception{
		JSAPResult config = readParameters(args);
		dataset = config.getString(NSConstants.PARAM_DATASET);
		thresholdTypeString = MotifConstants.getThresholdTypeString(kmerThresholdType, seqThresholdType);
		// Surrogate generate method
		method = config.getInt(NSConstants.PARAM_RANDOM_METHOD);
		// Traversal order
		order = config.getString(NSConstants.PARAM_ORDER);
		// Set number for surrogates
		int surrogateSet = config.getInt(PARAM_SURROGATE_SET);
		// Number of surrogate sequences for each neurite and k
		numInSet = config.getInt(MotifConstants.PARAM_SET_NUM);
		// Constraint from which to start analysis
		startK = config.getInt(PARAM_START_K);
		//startSet = config.getInt(PARAM_START_SET);
		
		// If provided, an additional surrogate set will take the place of real neurites
		useRandomSet = config.getInt(PARAM_USE_RANDOM_SET);
		// If using surrogates in place of real neurites, this determines what constraint to use for surrogates
		useRandomK = config.getInt(PARAM_USE_RANDOM_K);
		// Provides the opportunity to use a different surrogate generation method than used by the baseline
		useRandomMethod = config.getInt(PARAM_USE_RANDOM_METHOD);
		if (useRandomSet > 0 && useRandomMethod == 0){
			useRandomMethod = method;
		}
		// For determining which surrogate file to load for use in place of real neurite
		useRandomN = config.getInt(PARAM_USE_RANDOM_N);
		if (useRandomSet > 0 && useRandomN <= 0){
			useRandomN = numInSet;
			//useRandomMethod = setNum;
		}
		// Constraint on which to stop analysis
		maxConstraint = config.getInt(PARAM_MAX_CONSTRAINT);
		maxCheck = maxConstraint+kPlus;
		minLength = config.getInt(PARAM_MIN_LENGTH);
		measurementType = config.getInt(PARAM_MEASUREMENT_TYPE);
		String outputDir = config.getString(PARAM_OUTPUT_DIR);
		boolean baselineDistNA = config.getBoolean(PARAM_BASELINE_DIST_NA);

		sourceTraversal = order;
		if (config.userSpecified(NSConstants.PARAM_SOURCE_TRAVERSAL)){
			sourceTraversal = config.getString(NSConstants.PARAM_SOURCE_TRAVERSAL);
			if (!sourceTraversal.equals(order)){
				retraverse = true;
			}
		}

		System.out.println("Initializing...");

		// Initializing motifs
		motifsByLength = new List[maxCheck+1];
		for (int i = 1; i <= maxCheck; i++){
			if (i > 1){
				motifsByLength[i] = NeuriteSequenceStringUtils.generatePossibleACTKmers(i,order);
			}
			else{
				motifsByLength[1] = new ArrayList<String>();
				motifsByLength[1].add("C");
			}
		}

		// Get sequence names and map them to unique ids
		sequenceNames = NSDataReader.loadSequenceList(mapping);
		Map<String,Integer> sequenceIdMap = new HashMap<String,Integer>();
		for (int j = 0; j < sequenceNames.size(); j++){
			sequenceIdMap.put(sequenceNames.get(j), new Integer(j));
		}
		String fastaFile = DataProperties.getInstance().getFastaArborDir(dataset)+"Mapping"+mapping+"/Size/"+dataset+"ArborSize"+sourceTraversal+"Map"+mapping+".fasta";
		Map<String,Integer> sequenceLengthMap = NSDataReader.loadLengthDatabaseFromFastaFile(fastaFile);
		
		// Initializing string buffers for output
		StringBuffer[] sb = new StringBuffer[maxCheck];
		for (int k = 0; k <= maxConstraint; k++){
			sb[k] = new StringBuffer();
			sb[k].append("Sequence,N");
			for (int j = 0; j < motifsByLength[k+kPlus].size(); j++){
				sb[k].append(","+motifsByLength[k+kPlus].get(j));
			}
			sb[k].append("\n");
		}

		// Load real tree n-mer counts into sequence id keyed map
		System.out.println("Loading real tree motif counts");
		Map<String,Map<String,Number>> targetTreeMotifCounts = null;
		if (useRandomSet > 0){
			targetTreeMotifCounts = loadRandomSampleTreeMotifCounts(motifsByLength);
		}
		else{
			targetTreeMotifCounts = loadRealTreeMotifCounts(motifsByLength);
		}

		// Load up virtual sequences (key by sequenceId) from the N virtual sequence sets (standard N=20)
		System.out.println("Loading virtual sequences");
		Map<Integer,Map<String,List<String>>> kBaselineSequenceMap;
		// Only using newer storage method now
		kBaselineSequenceMap = ConstraintsDataReader.loadSurrogates(
				surrogateSet, dataset, sourceTraversal, method, numInSet, false);
		System.out.println(kBaselineSequenceMap.keySet());
		System.out.println(kBaselineSequenceMap.get(1).keySet());

		// Determine whether nmers are constrained
		Map<String,Number> targetKmerCountMap, baselineKmerCountMap;
		Map<String,List<String>> baselineSequenceMap;
		List<String> baselineSequences;
		List<Map<String,Number>> virtualSequenceKMerCountMaps;
		String kmer;
		DistributionBean baselineKmerCountDist;
		double measurement;
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(3);
		
		String sequence;
		// Loop through constraints (by k)
		for (int constraint = startK; constraint <= maxConstraint; constraint++){
			// Load list of virtual sequences associated with real sequence and k-constraint
			baselineSequenceMap = kBaselineSequenceMap.get(constraint);
			
			// Loop through sequences and k to calculate whether a given real sequence was constrained
			for (String sequenceName : targetTreeMotifCounts.keySet()){
				baselineSequences = baselineSequenceMap.get(sequenceName);
				System.out.println(sequenceName+" ");
				/*System.out.println(kVirtualSequenceMap.get(startK).get(sequenceId));
				System.out.println(kVirtualSequenceMap.get(startK).get(sequenceId).get(0));
				System.out.println(kVirtualSequenceMap.get(startK).get(sequenceId).get(0).length());*/
	
				//System.out.println(constraint+" " +sequenceName);
				//System.out.println("  "+kVirtualSequenceMap.get(constraint).get(sequenceName).size());
				if (!kBaselineSequenceMap.get(constraint).containsKey(sequenceName) || 
						kBaselineSequenceMap.get(constraint).get(sequenceName).get(0).length() < minLength){
					System.out.println(sequenceName+" not in baseline set for constraint k="+constraint);
					continue;
				}
				
				// Get kmer counts
				targetKmerCountMap = targetTreeMotifCounts.get(sequenceName);
				//System.out.println("Is "+sequenceName+" null? "+(targetKmerCountMap==null));
				
				if (baselineSequences == null){
					System.out.println("No sequences for "+sequenceName+" / "+constraint);
				}
				else if (baselineSequences.size() == 1){
					System.out.println("Only one sequence for "+sequenceName+" / "+constraint);
				}
				else{
					//System.out.println(k+" "+sequenceId+" "+sequenceNames.get(sequenceId.intValue()));
					
					// Calculate and assemble n-mer counts of k+1 (or kPlus, for which default is 1) across the virtual sequence sets
					virtualSequenceKMerCountMaps = new ArrayList<Map<String,Number>>();
					for (int i = 0; i < baselineSequences.size(); i++){
						sequence = baselineSequences.get(i);
						if (retraverse) sequence = MappedBinaryTreeUtils.reorderSequence(sequence, order);
						virtualSequenceKMerCountMaps.add(MotifFinder.findMotifs(sequence,constraint+kPlus));
					}

					// Start new line in output file
					sb[constraint].append(sequenceName+","+baselineSequences.size());
					
					/** Loop through motifs at k+1 and determine the real sequence's
					 *  kmer count percentile rank within the virtual sequence distribution
					 */
					for (int i = 0; i < motifsByLength[constraint+kPlus].size(); i++){
						// Get distribution of k+1-mer counts across virtual sequences
						kmer = motifsByLength[constraint+kPlus].get(i);
						baselineKmerCountDist = new DistributionBean();
						for (int j = 0; j < baselineSequences.size(); j++){
							baselineKmerCountMap = virtualSequenceKMerCountMaps.get(j);
							if (baselineKmerCountMap.containsKey(kmer)){
								baselineKmerCountDist.addData(baselineKmerCountMap.get(kmer));
							}
							else{
								baselineKmerCountDist.addData(0);
							}
						}
						switch (measurementType){
						case MEASURE_PERCENTILE_RANK:
							// Determine the percentile of the real sequence k-mer count within virtual distribution
							//System.out.println("kmer: "+kmer+"; "+targetKmerCountMap.keySet());
							measurement = baselineKmerCountDist.getPercentileRankForValue(
									targetKmerCountMap.get(kmer).doubleValue());
							if (baselineDistNA && baselineKmerCountDist.getStandardDeviation() == 0 && measurement == 0.5)
								measurement = -1;
							break;
						case MEASURE_FREQUENCY_DIFFERENCE:
							measurement = (targetKmerCountMap.get(kmer).doubleValue() - baselineKmerCountDist.getMean())/
							(sequenceLengthMap.get(sequenceName)-constraint+1);
							if (baselineDistNA && baselineKmerCountDist.getStandardDeviation() == 0 && measurement == 0.5)
								measurement = -1;
							break;
						default:
							throw new Exception("Unknown measurement type "+measurementType);
						}
						sb[constraint].append(","+nf.format(measurement));
					}
					sb[constraint].append("\n");
					
					/** 
					 * Determine whether this sequence is constrained for k based on its distribution 
					 * of percentile ranks (given its virtual sequence set) across k-mers. Percentile
					 * ranks should fall along a uniform distribution, so a Wilcox test comparing the
					 * distribution of k-mer percentile ranks with randomly generated values should
					 */
					// Off-loading this process to R. Easier to do analysis on subsets.
				}
			}
		}
		
 		// Write results to a csv file
		String suffix = "-"+dataset+"-"+numInSet+"Sets";
		if (surrogateSet > -1){
			suffix += "-"+surrogateSet;
		}
		if (useRandomSet > 0){
			if (useRandomK == -1){
				suffix += "-RandSet"+useRandomSet+"_Plus1";
			}else{
				suffix += "-RandSet"+useRandomSet+"_"+useRandomK;
			}
		}
		suffix += "-Size"+order+"-B"+method+sourceTraversal+"-T"+threshold+thresholdTypeString;

		String baseFilename = outputDir;
		if (measurementType == MEASURE_PERCENTILE_RANK){
			baseFilename += "/PR";
		}
		else if (measurementType == MEASURE_FREQUENCY_DIFFERENCE){
			baseFilename += "/FreqDiff";
		}
		
		String filename;
		for (int k = startK; k <= maxConstraint; k++){
			filename = baseFilename + suffix + "_K"+k+"-"+(k+kPlus)+".csv";
			DataWriter.writeStringToFile(filename, sb[k].toString());
			System.out.println("Writing to "+filename);
		}
		
	}
	
	private static Map<String,Map<String,Number>> loadRealTreeMotifCounts(
			List<String>[] motifsByLength) throws Exception{
		Map<String,Map<String,Number>> realTreeMotifCounts = new HashMap<String,Map<String,Number>>();
		FastaReader fr = new FastaReader();
		Map<String,Number> motifCounts;
		String sequence;
		
		fr.setFile(DataProperties.getInstance().getFastaArborDir(dataset)+"Mapping"+mapping+"/Size/"+dataset+"ArborSize"+sourceTraversal+"Map"+mapping+".fasta");
		List<String> sequenceList = NSDataReader.loadSequenceListFromFastaFile(mapping, dataset);
		//System.out.println(sequenceList);
		while (fr.next()){
			System.out.println(fr.getSequenceName());
			// Avoids case of using updated fasta file without updated lengths file
			if (sequenceList.contains(fr.getSequenceName())){
				sequence = fr.getSequence();
				if (retraverse) sequence = MappedBinaryTreeUtils.reorderSequence(sequence,order);
				
				motifCounts = MotifFinder.findMotifs(sequence,1,maxCheck);
				for (int k = startK+1; k <= maxCheck; k++){
					for (int j = 1; j < motifsByLength[k].size(); j++){
						if (!motifCounts.containsKey(motifsByLength[k].get(j))){
							motifCounts.put(motifsByLength[k].get(j), 0);
						}
					}
				}
				realTreeMotifCounts.put(fr.getSequenceName(), motifCounts);
			}
		}
		return realTreeMotifCounts;
	}

	private static Map<String,Map<String,Number>> loadRandomSampleTreeMotifCounts(
			List<String>[] motifsByLength) throws Exception{
		Map<String,Number> motifCounts;

		/* Old method of loading extra surrogate set
		File file = new File(MotifConstants.getConstrainedTreesFilename(sourceTraversal, method, useRandomSet, threshold, kmerThresholdType, seqThresholdType));
		// Load virtual sequences constrained by K
		Map<Integer,Map<Object,String>> virtualSequencesPerK = 
			ConstraintsDataReader.loadSequencesByRealSequence(file, true, doubleLineBreak);
		Map<Object,String> virtualSequences = virtualSequencesPerK.get(useRandomK);
		*/
		
		// New method of loading extra surrogate set
		Map<Integer,Map<String,List<String>>> virtualSequencesPerK = 
			ConstraintsDataReader.loadSurrogates(
				useRandomSet, dataset, sourceTraversal, useRandomMethod, useRandomN, false);

		// Go through all relevant constraints to pull k-mer counts
		String sequence;
		Map<String,Map<String,Number>> surrogateTreeMotifCounts = new HashMap<String,Map<String,Number>>();
		Map<String,List<String>> surrogateSequences;
		for (int k = startK+kPlus; k <= maxCheck; k++){
			System.out.println("K: "+k);
			// Get sequences lists keyed by neuriteName
			if (useRandomK == -1){
				surrogateSequences = virtualSequencesPerK.get(k-kPlus);
			}else{
				surrogateSequences = virtualSequencesPerK.get(useRandomK);
			}
			// Go through each neurite name and generate motif maps for the surrogates
			for (String seqName : surrogateSequences.keySet()){
				//System.out.println("Surrogate set kmer counting: "+seqName);
				// Just take the first sequence from the list
				sequence = surrogateSequences.get(seqName).get(0);
				
				if (retraverse) sequence = MappedBinaryTreeUtils.reorderSequence(sequence,order);

				// Add all kmers of k to the kmer count list for the current neurite name
				motifCounts = MotifFinder.findMotifs(sequence,k);
				if (k == startK+kPlus){
					surrogateTreeMotifCounts.put(seqName,motifCounts);
				}else{
					surrogateTreeMotifCounts.get(seqName).putAll(motifCounts);
				}
				for (int j = 0; j < motifsByLength[k].size(); j++){
					if (!motifCounts.containsKey(motifsByLength[k].get(j))){
						surrogateTreeMotifCounts.get(seqName).put(motifsByLength[k].get(j), 0);
					}
				}
			}
		}
		
		return surrogateTreeMotifCounts;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = MotifConstants.getParameters();
        
        FlaggedOption opt;

        opt = new FlaggedOption(NSConstants.PARAM_DATASET)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(NSConstants.DEFAULT_DATASET)
        .setShortFlag('D') 
        .setLongFlag(NSConstants.PARAM_DATASET);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_MAX_CONSTRAINT)
        .setStringParser(JSAP.INTEGER_PARSER)
	       .setRequired(false)
	       .setDefault("2")
	       .setShortFlag('C') 
	       .setLongFlag(PARAM_MAX_CONSTRAINT);
	    jsap.registerParameter(opt);
	    
    
	    opt = new FlaggedOption(PARAM_START_K)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_START_K);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_USE_RANDOM_SET)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_USE_RANDOM_SET);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_USE_RANDOM_K)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("-1")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_USE_RANDOM_K);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_USE_RANDOM_METHOD)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_USE_RANDOM_METHOD);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_USE_RANDOM_N)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_USE_RANDOM_N);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_DEBUG)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(NSConstants.PARAM_DEBUG);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_OUTPUT_DIR)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("C:/Documents and Settings/gillette/My Documents/research/MotifAnalysis/PercentileRanks/")
        .setShortFlag('O') 
        .setLongFlag(PARAM_OUTPUT_DIR);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_SURROGATE_SET)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("-1")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_SURROGATE_SET);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_BASELINE_DIST_NA)
    	.setStringParser(JSAP.BOOLEAN_PARSER)
        .setRequired(false)
        .setDefault("false")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_BASELINE_DIST_NA);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_MEASUREMENT_TYPE)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault(""+DEFAULT_MEASUREMENT_TYPE)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_MEASUREMENT_TYPE);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_SOURCE_TRAVERSAL)
    	.setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(""+NSConstants.DEFAULT_ORDER)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(NSConstants.PARAM_SOURCE_TRAVERSAL);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_MIN_LENGTH)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_MIN_LENGTH);
	    jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CompleteConstraintAnalyzer.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

}
