package org.cn3.ns.motif.analysis;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.cn3.morpho.data.MetaDataManager;
import org.cn3.morpho.domain.BinaryTreeNode;
import org.cn3.morpho.domain.DistributionBean;
import org.cn3.morpho.domain.NeuronMetadata;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Find matching segments (motifs) of a given length between two sequences. Keeps count
 * of how many are found of each motif and stores the frequencies in a HashMap
 *
 */
public class MotifFinder {

	private final static int DEFAULT_MOTIF_LENGTH = 6;
	
/*	private Map motifMap = new HashMap();
	private String query;
	private String subject;
	private int motifLength = DEFAULT_MOTIF_LENGTH;
	
	public MotifFinder(String query, String subject){
		this.query = query;
		this.subject = subject;
	}
	*/
	public static Map<String,Number> findMotifs(String query){
		return findMotifs(query, DEFAULT_MOTIF_LENGTH);
	}
	public static Map<String,Number> findMotifs(String query, int motifLength){
		return findMotifs(query, motifLength, motifLength);
	}
	public static Map<String,Number> findMotifs(String query, int motifLength, String order){
		return findMotifs(query, motifLength, motifLength, order);
	}
	public static Map<String,Number> findMotifs(String query, int minMotifLength, int maxMotifLength, String order){
		Set<String> motifs = new HashSet<String>();
		for (int i = minMotifLength; i <= maxMotifLength; i++){
			motifs.addAll(NeuriteSequenceStringUtils.generatePossibleACTKmers(i,order));
		}
		return findMotifs(query, minMotifLength, maxMotifLength, motifs);
	}
	public static Map<String,Number> findMotifs(String query, int minMotifLength, int maxMotifLength){
		Set<String> motifs = new HashSet<String>();
		for (int i = minMotifLength; i <= maxMotifLength; i++){
			motifs.addAll(NeuriteSequenceStringUtils.generatePossibleACTMotifs(i));
		}
		return findMotifs(query, minMotifLength, maxMotifLength, motifs);
	}
	public static Map<String,Number> findMotifs
	(String query, int minMotifLength, int maxMotifLength, Set<String> motifs){
		Map<String,Number> motifMap = new HashMap<String,Number>();
		String motif;
		int motifLength;
		for (Iterator<String> it = motifs.iterator(); it.hasNext();){
			motifMap.put(it.next(), new Integer(0));
		}
		for (int i = 0; i <= query.length() - minMotifLength; i++){
			for (motifLength = minMotifLength; motifLength <= maxMotifLength; motifLength++){
				if (motifLength+i <= query.length()){
					motif = query.substring(i,i+motifLength);
					if (motifs.contains(motif)){
						motifMap.put(motif, 
								new Integer(((Integer)motifMap.get(motif)).intValue()+1));
					}
				}
			}
		}
		return motifMap;
	}

	public static Map<String,Number> findMotifFrequencies(String query){
		return findMotifFrequencies(query, DEFAULT_MOTIF_LENGTH);
	}
	public static Map<String,Number> findMotifFrequencies(String query, int motifLength){
		return findMotifFrequencies(query, motifLength, motifLength);
	}
	public static Map<String,Number> findMotifFrequencies(String query, int minMotifLength, int maxMotifLength){
		Set<String> motifs = new HashSet<String>();
		for (int i = minMotifLength; i <= maxMotifLength; i++){
			motifs.addAll(NeuriteSequenceStringUtils.generatePossibleACTMotifs(i));
		}
		return findMotifFrequencies(query, minMotifLength, maxMotifLength, motifs);
	}
	public static Map<String,Number> findMotifFrequencies
	(String query, int minMotifLength, int maxMotifLength, Set<String> motifs){
		Map<String,Number> motifMap = new HashMap<String,Number>();
		String motif;
		int motifLength, freqInd;
		double[] freqPerCount = new double[maxMotifLength - minMotifLength + 1];
		//System.out.println("minMLen: "+minMotifLength+"; maxMLen: "+maxMotifLength+"; queryLen: "+query.length());

		// Initialize
		for (Iterator<String> it = motifs.iterator(); it.hasNext();){
			motifMap.put(it.next(), new Double(0));
		}
		
		for (motifLength = minMotifLength; motifLength <= maxMotifLength; motifLength++){
			freqInd = motifLength - minMotifLength;
			freqPerCount[freqInd] = (double)1/(query.length() - motifLength + 1);
		}
		for (int i = 0; i <= query.length() - minMotifLength; i++){
			for (motifLength = minMotifLength; motifLength <= maxMotifLength; motifLength++){
				freqInd = motifLength - minMotifLength;
				if (motifLength+i <= query.length()){
					motif = query.substring(i,i+motifLength);
					if (!motifMap.containsKey(motif)){
						motifMap.put(motif, new Double(freqPerCount[freqInd]));
					}
					else{
						motifMap.put(motif, 
							new Double(((Double)motifMap.get(motif)).doubleValue()
									+freqPerCount[freqInd]));
					}
				}
			}
		}
		return motifMap;
	}
	
	public static Map<String,Double> findDependencies(String query, char target){
		Map<String,Double> dependencies = new HashMap<String,Double>();
		String key;
		Map<String,DistributionBean> distMaps = new HashMap<String,DistributionBean>();
		DistributionBean dist;
		for (int i = 0; i < query.length(); i++){
			key = ""+query.charAt(i);
			if (!distMaps.containsKey(key)){
				distMaps.put(key, new DistributionBean());
			}
			dist = (DistributionBean)distMaps.get(key);
			
			if (query.charAt(i+1) == target){
				dist.addData(1);
			}
			else{
				dist.addData(0);
			}
		}
		for (Iterator<String> it = distMaps.keySet().iterator(); it.hasNext();){
			key = (String)it.next();
			dependencies.put(key, new Double(((DistributionBean)distMaps.get(key)).getMean()));
		}
		return dependencies;
	}
	
	public static Map<String,Number> findMotifs(String query, String subject){
		return findMotifs(query, subject, DEFAULT_MOTIF_LENGTH);
	}
	
	public static Map<String,Number> findMotifs(String query, String subject, int motifLength){
		return findMotifs(query, subject, motifLength, motifLength);
	}

	public static Map<String,Number> findMotifs(String query, String subject, 
			int minMotifLength, int maxMotifLength){
		int queryStart = 0, motifLength;
		int subjectStart = subject.length() - minMotifLength - 1;
		int queryInd, subjectInd;
		String motif;
		Map<String,Number> motifMap = new HashMap<String,Number>();
		while (query.length()-queryStart >= minMotifLength){
			for (motifLength = minMotifLength; motifLength <= maxMotifLength; motifLength++){
				queryInd = queryStart;
				subjectInd = subjectStart;
				while (queryInd + motifLength < query.length() 
						&& subjectInd + motifLength < subject.length()){
					if (query.substring(queryInd, queryInd+motifLength).equals(
						subject.substring(subjectInd+motifLength))){
						motif = query.substring(queryInd, queryInd+motifLength);
						if (!motifMap.containsKey(motif)){
							motifMap.put(motif, new Integer(1));
						}
						else{
							motifMap.put(motif, 
								new Integer(((Integer)motifMap.get(motif)).intValue()+1));
						}
					}
					queryInd++;
					subjectInd++;
				}
			}
				
			if (subjectStart > 0){
				subjectStart--;
			}
			else{
				queryStart++;
			}
		}
		return motifMap;
	}

	public static Map<String,Number> findMotifFrequencies(BinaryTreeNode tree){
		return findMotifFrequencies(tree, DEFAULT_MOTIF_LENGTH);
	}
	public static Map<String,Number> findMotifFrequencies(BinaryTreeNode tree, int motifLength){
		return findMotifs(tree, motifLength, motifLength, true);
	}
	public static Map<String,Number> findMotifFrequencies(BinaryTreeNode tree, int minMotifLength, int maxMotifLength){
		return findMotifs(tree, minMotifLength, maxMotifLength, true);
	}
	public static Map<String,Number> findMotifs(BinaryTreeNode tree){
		return findMotifs(tree, DEFAULT_MOTIF_LENGTH);
	}
	public static Map<String,Number> findMotifs(BinaryTreeNode tree, int motifLength){
		return findMotifs(tree, motifLength, motifLength, false);
	}
	public static Map<String,Number> findMotifs(BinaryTreeNode tree, int minMotifLength, int maxMotifLength){
		return findMotifs(tree, minMotifLength, maxMotifLength, false);
	}
	public static Map<String,Number> findMotifs(
			BinaryTreeNode tree, int motifLength, boolean frequencies){
		return findMotifs(tree, motifLength, motifLength, frequencies);
	}
	public static Map<String,Number> findMotifs(
			BinaryTreeNode tree, int minMotifLength, int maxMotifLength, boolean frequencies){
		Map<String,Number> motifMap = new HashMap<String,Number>();
		Stack<BinaryTreeNode> initStack = new Stack<BinaryTreeNode>(),
			nodeStack = new Stack<BinaryTreeNode>();
		
		initStack.push(tree);
		nodeStack.push(tree);
		String motif;
		int motifLength, count;
		int[] totalMotifs = new int[maxMotifLength-minMotifLength+1];
		BinaryTreeNode baseNode, currNode;
		while (!initStack.isEmpty()){
			baseNode = (BinaryTreeNode)initStack.pop();
			if (baseNode.hasChildren()){
				nodeStack.push(baseNode);
				initStack.push(baseNode.getLeft());
				initStack.push(baseNode.getRight());
			}
		}
		
		while (!nodeStack.isEmpty()){
			baseNode = (BinaryTreeNode)nodeStack.pop();
			currNode = baseNode;
			for (motifLength = minMotifLength; motifLength <= maxMotifLength; motifLength++){
				count = 0;
				motif = "";
				while (count < motifLength && currNode != null){
					count++;
					motif = motif.concat(""+currNode.getData());
					currNode = currNode.getParent();
				}
				if (count == motifLength){
					totalMotifs[motifLength-minMotifLength]++;
					if (!motifMap.containsKey(motif)){
						motifMap.put(motif, new Double(1));
					}
					else{
						motifMap.put(motif, 
							new Double(((Double)motifMap.get(motif)).doubleValue()+1));
					}
				}
			}
		}

		if (frequencies){
			Double freq;
			for (Iterator<String> it = motifMap.keySet().iterator(); it.hasNext();){
				motif = it.next();
				freq = (Double)motifMap.get(motif);
				motifMap.put(motif, new Double(
						freq.doubleValue()/totalMotifs[motif.length()-minMotifLength]));
			}
		}

		return motifMap;
	}

	private final static String PARAM_ARBOR_NAME = "name";
	private final static String PARAM_MOTIF_SIZE = "motif-size";
	private final static String PARAM_PRENORMALIZE = "prenormalize";
	private final static int DEFAULT_MOTIF_SIZE = 6;

	public static void main(String[] args){
		try{
			JSAPResult config = readParameters(args);
			String name = config.getString(PARAM_ARBOR_NAME);
			int motifLength = config.getInt(PARAM_MOTIF_SIZE);
			boolean prenorm = config.getBoolean(PARAM_PRENORMALIZE);
			
			Map<String,?> metadata = MetaDataManager.getCellMetaData();
			NeuronMetadata neuron = (NeuronMetadata)metadata.get(NeuriteSequenceStringUtils.getNeuronNameFromSequence(name));
			String filename = DataProperties.getInstance().getFastaArborDir()+"Mapping2/Size/"+neuron.getLabName()+"/"+name+".lts.fasta";
			System.out.println(filename);
			
			String sequence, motif;
			FastaReader fr = new FastaReader(filename);
			if (fr.next()){
				sequence = fr.getSequence();
				Map<String,Number> results;
				if (prenorm){
					results = findMotifFrequencies(sequence, motifLength);
				}
				else{
					results = findMotifs(sequence, motifLength);
				}
				for (Iterator<String> it = results.keySet().iterator(); it.hasNext();){
					motif = it.next();
					System.out.println(motif+": "+results.get(motif));
				}
			}
			
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_ARBOR_NAME)
			.setStringParser(JSAP.STRING_PARSER)
			.setRequired(true)
			.setShortFlag('n') 
			.setLongFlag(PARAM_ARBOR_NAME);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_MAPPING_MODE)
    		.setStringParser(JSAP.INTEGER_PARSER)
    		.setRequired(false)
    		.setDefault(""+NSConstants.DEFAULT_MAPPING_MODE)
    		.setShortFlag('M') 
    		.setLongFlag(NSConstants.PARAM_MAPPING_MODE);
        jsap.registerParameter(opt);
    
        opt = new FlaggedOption(PARAM_MOTIF_SIZE)
			.setStringParser(JSAP.INTEGER_PARSER)
			.setRequired(false)
			.setDefault(""+DEFAULT_MOTIF_SIZE)
			.setShortFlag('S') 
			.setLongFlag(PARAM_MOTIF_SIZE);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_PRENORMALIZE)
			.setStringParser(JSAP.BOOLEAN_PARSER)
			.setRequired(false)
			.setDefault("TRUE")
			.setShortFlag('N') 
			.setLongFlag(PARAM_PRENORMALIZE);
        jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + MotifFinder.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
