package org.cn3.ns.motif.generation;

import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.data.persistence.GACTCompressor;
import org.cn3.ns.generation.NodeSequenceBinaryTreeGenerator;
import org.cn3.ns.motif.analysis.MotifFinder;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.MappedBinaryTreeUtils;
import org.cn3.ns.utils.MappingUtils;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 *	Generate numSurrogates surrogates for each real sequence and k, fully constrained
 *	-n --numSurrogates | required
 *	-o --order (LTS)
 *	-b --baseline (8)	Uniform (3) not recommended for large datasets since it will take too long to run, unless this is to generate a set to compare against the "constructed" methods
 *	-s --setNum (-1)
 *	-M --maxConstraint (4)
 *	--maxSequenceLength (0)
 *	-p --numProcessors (1)
 *	-u --uniformApprox (false)
 *  --startConstraint (0)
 *	--sampledSurrogates (100000)
 *	--compress (false)
 *	--debug (false)
 *
 * 4/2/15 - Same as before, but with baseline 11 (composite k-mers removed from constraints)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o STL --startConstraint 0 -M 4 -s 1 -b 11 --maxSequenceLength 1200 -p 3 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o STL --startConstraint 0 -M 4 -s 2 -b 11 --maxSequenceLength 1200 -p 3 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o LTS --startConstraint 3 -M 4 -s 1 -b 11 --maxSequenceLength 1200 -p 3 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o LTS --startConstraint 3 -M 4 -s 2 -b 11 --maxSequenceLength 1200 -p 3 &
 *
 * Companion set (additional surrogate for outlier test of randoms)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 1 -o STL --startConstraint 0 -M 4 -s 1 -b 11 --maxSequenceLength 1200 -p 3 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 1 -o LTS --startConstraint 3 -M 4 -s 1 -b 11 --maxSequenceLength 1200 -p 3 &
 *
 * New 3/19/15 - For NM0 6.0 - 100 surrogates/sequence of current database SpurLen2D
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o STL --startConstraint 0 -M 4 -s 1 -b 8 --maxSequenceLength 1200 -p 6 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o STL --startConstraint 0 -M 4 -s 2 -b 8 --maxSequenceLength 1200 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o LTS --startConstraint 3 -M 4 -s 1 -b 8 --maxSequenceLength 1200 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o LTS --startConstraint 3 -M 4 -s 2 -b 8 --maxSequenceLength 1200 &
 *
 * Companion set (additional surrogate for outlier test of randoms)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 1 -o STL --startConstraint 0 -M 4 -s 1 -b 8 --maxSequenceLength 1200 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 1 -o LTS --startConstraint 3 -M 4 -s 1 -b 8 --maxSequenceLength 1200 &
 *
 * New 2/18/14 - 100 surrogates/sequence of current database SpurLen2D
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o STL --startConstraint 0 -M 4 -s 1 -b 8 --maxSequenceLength 1200 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o STL --startConstraint 0 -M 4 -s 2 -b 8 --maxSequenceLength 1200 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o LTS --startConstraint 3 -M 4 -s 1 -b 8 --maxSequenceLength 1200 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 100 -o LTS --startConstraint 3 -M 4 -s 2 -b 8 --maxSequenceLength 1200 &
 *
 * Companion set (additional surrogate for outlier test of randoms)
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 1 -o STL --startConstraint 0 -M 4 -s 1 -b 8 --maxSequenceLength 1200 &
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -D NM_SpurLen2D -n 1 -o LTS --startConstraint 3 -M 4 -s 1 -b 8 --maxSequenceLength 1200 &
 *
 * New 8/20/13 - 20 surrogates/sequence of current database - k=3 to be pulled into specific class files
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 20 -o STL --startConstraint 3 -M 3 -s 1 -b 8 --maxSequenceLength 2000
 * New 11/2/13 - 20 surrogates/sequence of current database - k=0 to be pulled into specific class files
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 20 -o STL --startConstraint 0 -M 0 -s 1 -b 8 --maxSequenceLength 2000
 * New 12/20/13 - 40 STL surrogates/sequence of current database - k=0-2, for a full rerun for Motif results 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o STL --startConstraint 0 -M 2 -s 1 -b 8 --maxSequenceLength 2000
 *
 * Original runs
 *	java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o STL -M 4 -s 2 --maxSequenceLength 1200
 *	java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o LTS -M 4 -s 2 --maxSequenceLength 1200
 *
 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o LTS -M 4 -s 3 --maxSequenceLength 1200
 *  
 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o LTS --startConstraint 0 -M 2 -s 5 --maxSequenceLength 1200
 *  
 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o LTS --startConstraint 0 -M 2 -s 7 -b 8 --maxSequenceLength 2000
 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o STL --startConstraint 0 -M 2 -s 7 -b 8 --maxSequenceLength 2000

 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o LTS --startConstraint 0 -M 3 -s 6 -b 8 --maxSequenceLength 2000
 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 40 -o STL --startConstraint 3 -M 3 -s 6 -b 8 --maxSequenceLength 2000
 * Companion set (additional surrogate for outlier test of randoms)
 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 1 -o LTS --startConstraint 0 -M 3 -s 6 -b 8 --maxSequenceLength 2000
 *  java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.motif.generation.ConstrainedSurrogateGenerator -n 1 -o STL --startConstraint 3 -M 3 -s 6 -b 8 --maxSequenceLength 2000
 *
 *
 *
 */
public class ConstrainedSurrogateGenerator {
	
	public static final String PARAM_NUM_SURROGATES = "numSurrogates";
	public static final String PARAM_UNIFORM_APPROXIMATION = "uniformApprox";
	public static final String PARAM_SAMPLED_SURROGATES = "sampledSurrogates";
	public static final String PARAM_COMPRESS = "compress";
	public static final String PARAM_START_CONSTRAINT = "startConstraint";
	
	private static final String PARAM_MAX_SEQUENCE_LENGTH = "maxSequenceLength";
	private static final String PARAM_NUM_PROCESSORS = "numProcessors";

	public static final int DEFAULT_SAMPLED_SURROGATES = 100000;
	
	private static final int DEFAULT_MAX_SEQUENCE_LENGTH = 0;
	private static final int DEBUG_AT = 1000000;
	
	private static String dataset;
	private static String order;
	private static int numSurrogates;
	private static int setNum;
	private static int method;
	private static int mapping = NSConstants.MAPPING_MODE_TOPO_ACT;
	private static int startConstraint = 0;
	private static int maxConstraint = 4;
	private static boolean uniformApproximation = false;
	private static int sampledSurrogates = DEFAULT_SAMPLED_SURROGATES;
	private static boolean compress = false;
	private static int debug = 0;
	
	// Shared between threads
	private static List<String> sequenceNames;
	private static List<String> removedSequences;
	private static Map<String,Map<String,Number>> realTreeMotifCounts;
	private static Map<Integer,List<String>> motifsByLength;
	private static int[] firsts;
	private static Map<String,Integer> lengthMap;
	private static int numProcessors;
	private static int numGeneratedSequences;
	private static CountDownLatch latch;
	private static Set<String> motifs;
	private static String outputFilename;
	private static Map<String,List<String>> completedSequencesForK;
	
	public static void main(String args[]) throws Exception{
		JSAPResult config = readParameters(args);
		dataset = config.getString(NSConstants.PARAM_DATASET);
		order = config.getString(NSConstants.PARAM_ORDER);
		method = config.getInt(NSConstants.PARAM_BASELINE);
		numSurrogates = config.getInt(PARAM_NUM_SURROGATES);
		setNum = config.getInt(MotifConstants.PARAM_SET_NUM);
		int maxSequenceLength = config.getInt(PARAM_MAX_SEQUENCE_LENGTH);
		numProcessors = config.getInt(PARAM_NUM_PROCESSORS);
		startConstraint = config.getInt(PARAM_START_CONSTRAINT);
		maxConstraint = config.getInt(MotifConstants.PARAM_MAX_CONSTRAINT);
		startConstraint = config.getInt(PARAM_START_CONSTRAINT);
		uniformApproximation = config.getBoolean(PARAM_UNIFORM_APPROXIMATION);
		compress = config.getBoolean(PARAM_COMPRESS);
		if (uniformApproximation){
			sampledSurrogates = config.getInt(PARAM_SAMPLED_SURROGATES);
			if (sampledSurrogates <= 0){
				throw new Exception("Number of sampled surrogates must be greater than 0");
			}
			numGeneratedSequences = sampledSurrogates;
		}
		else{
			sampledSurrogates = 0;
			numGeneratedSequences = numSurrogates;

		}
		debug = config.getInt(NSConstants.PARAM_DEBUG);
		NodeSequenceBinaryTreeGenerator.debugLevel = debug;
		
		System.out.println("Initializing...");
		
		int length, lengthIndex;

		outputFilename = MotifConstants.getSurrogatesFilename(
				setNum, dataset, order, method, numSurrogates, sampledSurrogates);
		System.out.println("Output file: "+outputFilename);
		// Read output file to determine what has already been run
		Map<Integer,Map<String,List<String>>> completedSequences = 
			ConstraintsDataReader.loadSurrogates(outputFilename, numSurrogates, compress);
		
		// Get sequence names and map them to unique IDs
		sequenceNames = NSDataReader.loadSequenceListFromFastaFile(mapping,dataset,order);
		
		System.out.println("Loading tree size data");

		// Loading sequence:length Map
		lengthMap = NSDataReader.loadLengthDatabaseFromFastaFile(mapping,dataset,order);

		// Loading tree sizes represented in database
		List<Integer> sequenceLengths = NSDataReader.loadUniqueSequenceLengths(lengthMap);

		// Loading map keyed by seq length, values are lists of sequence names of seqs of that length
		Map<Integer,List<String>> lengthArborsMap = NSDataReader.loadTreeSizeArborizationsMap(lengthMap);
		//loadTreeSizeArborizationsMap(sequenceIdMap);
		
		// Initializing
		motifsByLength = new TreeMap<Integer,List<String>>();
		for (int i = 1; i <= maxConstraint; i++){
			if (i > 1){
				motifsByLength.put(i,NeuriteSequenceStringUtils.generatePossibleACTMotifs(i));
			}
			else{
				motifsByLength.put(1,new ArrayList<String>());
				motifsByLength.get(1).add("C");
			}
		}

		// Load real tree n-mer counts into sequence id keyed map
		System.out.println("Loading real tree motif counts");
		realTreeMotifCounts = loadRealTreeMotifCounts(motifsByLength);
		
		// Remove lengths that are too big
		removedSequences = new ArrayList<String>();
		if (maxSequenceLength > 0){
			System.out.println("Removing very large sequences (larger than "+maxSequenceLength+")");
			for (lengthIndex = sequenceLengths.size()-1; sequenceLengths.get(lengthIndex).intValue() > maxSequenceLength; lengthIndex--){
				length = sequenceLengths.get(lengthIndex);
				removedSequences.addAll(lengthArborsMap.remove(length));
				sequenceLengths.remove(lengthIndex);
			}
		}
		
		// Add flagged neurons to removedSequences lists
		List<String> flaggedSequences = NSDataReader.getFlaggedSequences();
		removedSequences.addAll(flaggedSequences);
		
		// Remove all removedSequences from the list of sequenceNames
		sequenceNames.removeAll(removedSequences);

		ExecutorService executor = Executors.newFixedThreadPool(numProcessors);
		SurrogateRunnable[] runnables = new SurrogateRunnable[numProcessors];
		// Breaking up the sequences to run on each processor
		firsts = new int[numProcessors];
		int interval = sequenceNames.size()/numProcessors;
		for (int i = 0; i < numProcessors; i++){
			firsts[i] = i*interval;
		}
		
		// Loop through constraint lengths
		/*StringBuffer everything = new StringBuffer();
		everything.append("NumSurrogates:"+numSurrogates+"\n");*/
		new File(outputFilename).getParentFile().mkdirs();
		if (!new File(outputFilename).exists()){
			DataWriter.writeStringToFile(outputFilename,"NumSurrogates:"+numSurrogates+"\n"); 
		}
		for (int k = startConstraint; k <= maxConstraint; k++){
			System.out.println("K: "+k);
			//everything.append("K:"+k+"\n");
			DataWriter.appendStringToFile(outputFilename, "K:"+k+"\n");
			if (completedSequences!= null && completedSequences.containsKey(k)){
				completedSequencesForK = completedSequences.get(k);
			}
			else{
				completedSequencesForK = null;
			}
			
			// Set up motifs set for this k
			motifs = new HashSet<String>();
			for (int i = k; i <= maxConstraint; i++){
				if (i > 0) motifs.addAll(motifsByLength.get(i));
			}
			
			latch = new CountDownLatch(numProcessors);
			//runnables
			for (int i = 0; i < numProcessors; i++){
				runnables[i] = new SurrogateRunnable(i, k);
				runnables[i].localDebug = debug;
				executor.execute(runnables[i]);
			}
			
			// Wait until all threads are finish
			if (debug > 1){
				int loops = 0;
				while (latch.getCount() > 0){
					Thread.sleep(60000);
					loops++;
					System.out.println((loops)+" minutes");
					for (int i = 0; i < numProcessors; i++){
						System.out.println(i+": "+runnables[i].getCurrentSequence()+" "+runnables[i].getGenerationCount()+"; debugPos: "+runnables[i].getDebugPosition());
					}
				}
			} else{
				latch.await();
			}
			System.out.println("Finished all threads");

			// Grab summary data from all runs
			/*
			for (int i = 0; i < numProcessors; i++){
				everything.append(runnables[i].getOutput());
				runnables[i] = null;
			}*/

			System.out.println("Generated "+k);

			for (int i = 0; i < numProcessors; i++){
				runnables[i] = null;
			}
		}
		executor.shutdown();


		/*DataWriter.writeStringToFile(
				outputFilename,
				everything.toString());

		System.out.println("Wrote results to "+outputFilename);
		*/
	}
	
	private static synchronized void appendToFile(String string) throws IOException{
		DataWriter.appendStringToFile(outputFilename, string);
	}
	
	private static Map<String,Map<String,Number>> loadRealTreeMotifCounts(
			Map<Integer,List<String>> motifsByLength) throws Exception{
		Map<String,Map<String,Number>> realTreeMotifCounts = new HashMap<String,Map<String,Number>>();
		FastaReader fr = new FastaReader();
		Map<String,Number> motifCounts;
		
		fr.setFile(DataProperties.getInstance().getFastaArborDir(dataset)+"Mapping"+mapping+"/Size/"+dataset+"ArborSize"+order+"Map"+mapping+".fasta");
		List<String> sequenceList = NSDataReader.loadSequenceListFromFastaFile(mapping, dataset);
		while (fr.next()){
			// Avoids case of using updated fasta file without updated lengths file
			if (sequenceList.contains(fr.getSequenceName())){
				motifCounts = MotifFinder.findMotifs(fr.getSequence(),1,maxConstraint);
				for (int i = 1; i <= maxConstraint; i++){
					for (int j = 1; j < motifsByLength.get(i).size(); j++){
						if (!motifCounts.containsKey(motifsByLength.get(i).get(j))){
							motifCounts.put(motifsByLength.get(i).get(j), 0);
						}
					}
				}
				realTreeMotifCounts.put(fr.getSequenceName(), motifCounts);
			}
		}
		return realTreeMotifCounts;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(NSConstants.PARAM_DATASET)
        .setStringParser(JSAP.STRING_PARSER)
        .setDefault(NSConstants.DEFAULT_DATASET)
        .setRequired(false)
        .setShortFlag('D') 
        .setLongFlag(NSConstants.PARAM_DATASET);
	    jsap.registerParameter(opt);

        opt = new FlaggedOption(NSConstants.PARAM_ORDER)
        .setStringParser(JSAP.STRING_PARSER)
        .setDefault(NSConstants.DEFAULT_ORDER)
        .setRequired(false)
        .setShortFlag('o') 
        .setLongFlag(NSConstants.PARAM_ORDER);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_BASELINE)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault(""+NSConstants.DISTRIBUTION_NODE_TYPE_ASSEMBLED3)
        .setShortFlag('b') 
        .setLongFlag(NSConstants.PARAM_BASELINE);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_NUM_SURROGATES)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(true)
        .setShortFlag('n') 
        .setLongFlag(PARAM_NUM_SURROGATES);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(MotifConstants.PARAM_SET_NUM)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("-1")
        .setShortFlag('s') 
        .setLongFlag(MotifConstants.PARAM_SET_NUM);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_MAX_SEQUENCE_LENGTH)
	   	.setStringParser(JSAP.INTEGER_PARSER)
	    .setRequired(false)
	    .setDefault(""+DEFAULT_MAX_SEQUENCE_LENGTH)
	    .setShortFlag(JSAP.NO_SHORTFLAG) 
	    .setLongFlag(PARAM_MAX_SEQUENCE_LENGTH);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_NUM_PROCESSORS)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("1")
        .setShortFlag('p') 
        .setLongFlag(PARAM_NUM_PROCESSORS);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_UNIFORM_APPROXIMATION)
    	.setStringParser(JSAP.BOOLEAN_PARSER)
        .setRequired(false)
        .setDefault("false")
        .setShortFlag('u') 
        .setLongFlag(PARAM_UNIFORM_APPROXIMATION);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_SAMPLED_SURROGATES)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault(""+DEFAULT_SAMPLED_SURROGATES)
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_SAMPLED_SURROGATES);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_COMPRESS)
    	.setStringParser(JSAP.BOOLEAN_PARSER)
        .setRequired(false)
        .setDefault("false")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_COMPRESS);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(MotifConstants.PARAM_MAX_CONSTRAINT)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("4")
        .setShortFlag('M') 
        .setLongFlag(MotifConstants.PARAM_MAX_CONSTRAINT);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_START_CONSTRAINT)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_START_CONSTRAINT);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(NSConstants.PARAM_DEBUG)
    	.setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(NSConstants.PARAM_DEBUG);
	    jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + ConstrainedSurrogateGenerator.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

	private static class SurrogateRunnable implements Runnable {
		private int threadNum;
		private int k;
		private int localDebug;
		private StringBuffer virtualSequencesSB;
		private String currentSequence;
		private int generationCount;
		private int debugPosition;
		
		SurrogateRunnable(int threadNum, int k){
			this.threadNum = threadNum;
			this.k = k;
		}
		
		public String getOutput(){
			return virtualSequencesSB.toString();
		}
		public String getCurrentSequence(){
			return currentSequence;
		}
		public int getGenerationCount(){
			return generationCount;
		}
		public int getDebugPosition(){
			return debugPosition;
		}
		
		private void print(String debugStr){
			System.out.println("Thread "+threadNum+": "+debugStr);
		}
		private void trace(String debugStr){
			debug(debugStr, 2);
		}
		private void debug(String debugStr){
			debug(debugStr, 1);
		}
		private void debug(String debugStr, int level){
			if (localDebug >= level){
				System.out.println("Thread "+threadNum+": "+debugStr);
			}
		}
		
		public void run(){
			String virtualSequence = null;
			int length, first, last, degree;
			Map<String,Number> motifCounts;
			char firstChar = 'A'; // Initialize
			Set<String> uniqueSequences = new TreeSet<String>();
			List<String> sequences = new ArrayList<String>(numGeneratedSequences);
			
			BigInteger shapeNum = null, possibleShapes = null;
			Map<String,Number> virtualMotifCounts;
			long generations = 0;

			virtualSequencesSB = new StringBuffer();
			first = firsts[threadNum];
			if (threadNum == numProcessors-1){
				last = sequenceNames.size();
			}
			else{
				last = firsts[threadNum+1];
			}
			System.out.println("First: "+first+"; Last: "+last);

			try{

				// Loop through sequences to set the threshold at which the difference count causes a virtual sequence not to match
				for (int seqInd = first; seqInd < last; seqInd++){
					currentSequence = sequenceNames.get(seqInd);
					//System.out.println("K"+k+"; "+seqInd + " " +realTreeMotifCounts.containsKey(currentSequence));
					// Skip this one if the real sequence isn't in the fasta file or if it's already in the output file
					if (realTreeMotifCounts.containsKey(currentSequence) &&
							(completedSequencesForK == null || !completedSequencesForK.containsKey(currentSequence))){
						//System.out.println("Generating "+numGeneratedSequences+" sequences");
						currentSequence = sequenceNames.get(seqInd);
						length = lengthMap.get(currentSequence);
						degree = length+1;
						debugPosition = 1;
						if (method == NSConstants.DISTRIBUTION_UNIFORM){
							possibleShapes = MappedBinaryTreeUtils.numShapes(degree);
						}
						debugPosition = 2;
						//motifCounts = realTreeMotifCounts.get(seqInd);
						motifCounts = realTreeMotifCounts.get(currentSequence);
					
						//virtualSequencesSB.append("S:"+seqInd+"\n");
						virtualSequencesSB.append("S:"+currentSequence+"\n");
						debugPosition = 3;
						firstChar = getFirstChar(motifCounts);

						sequences.clear();
						debugPosition = 4;
						if (k == 1) debug("K"+k+"; Sequence id: "+seqInd+"; name: "+currentSequence + " degree "+degree+" cnodes "+motifCounts.get("C").intValue(),1);
						else if (k > 1) debug("K"+k+"; Sequence id: "+seqInd+"; name: "+currentSequence + " degree "+degree);
						for (generationCount = 0; generationCount < numGeneratedSequences; generationCount++){
							if (k == 1 && degree > 500) debug("Surrogate "+generationCount,2);
							generations++;
							if (generations%DEBUG_AT==0){
								debug(seqInd+" sequences completed; Currently on sequence "+seqInd+"; generation "+generationCount+" of "+numGeneratedSequences);
							}
							debugPosition = 5;
							switch (method){
							case NSConstants.DISTRIBUTION_NODE_TYPE_ASSEMBLED3:
								switch (k){
								case 0:
									virtualSequence = NodeSequenceBinaryTreeGenerator.generateRandomHardingTreeMap2Sequence(degree,order);
									//virtualSequence = MappingUtils.traverseTreeToString(
									//		BinaryTreeGenerator.generateRandom???(length+1,order),
									//		Constants.MAPPING_MODE_TOPO_ACT);
									break;
								case 1:
									/*virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedNonuniformTree(
														degree, ((Integer)motifCounts.get("C")).intValue()),
												NSConstants.MAPPING_MODE_TOPO_ACT);*/
									//System.out.println("Run "+k);
									virtualSequence = MappingUtils.traverseTreeToString(
											NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedHardingTreeByDegree(
													degree, motifCounts.get("C").intValue()),
											NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								case 2:
									//NodeSequenceBinaryTreeGenerator.debugLevel = 3;
									virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomDimerConstrainedNonuniformTree(
														motifCounts, order),
												NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								case 3:
									virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomTrimerConstrainedNonuniformTree(
														firstChar, motifCounts, order),
												NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								case 4:
									virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomTetramerConstrainedNonuniformTree(
														firstChar, motifCounts, order),
												NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								default:
									throw new Exception("K is "+k+" but should only be 0-4");
								}
								break;
							case NSConstants.DISTRIBUTION_NODE_TYPE_ASSEMBLED4:
								switch (k){
								case 0:
									virtualSequence = NodeSequenceBinaryTreeGenerator.generateRandomHardingTreeMap2Sequence(degree,order);
									break;
								case 1:
									virtualSequence = MappingUtils.traverseTreeToString(
											NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedHardingTreeByDegree(
													degree, motifCounts.get("C").intValue()),
											NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								case 2:
									virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomDimerConstrainedNonuniformTree(
														motifCounts, order),
												NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								case 3:
									virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomTrimerConstrainedNonuniformTree(
														firstChar, motifCounts, order, true),
												NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								case 4:
									virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomTetramerConstrainedNonuniformTree(
														firstChar, motifCounts, order, true),
												NSConstants.MAPPING_MODE_TOPO_ACT, order);
									break;
								default:
									throw new Exception("K is "+k+" but should only be 0-4");
								}
								break;
							case NSConstants.DISTRIBUTION_UNIFORM:
								virtualSequence = null;
								do{
									switch (k){
									case 1:
										/*
										shapeNum = NodeSequenceBinaryTreeGenerator.randomBigInteger(possibleShapes);
										virtualMotifCounts = new HashMap<String,Number>();
										virtualMotifCounts.put("C", LengthMonomerDistGenerator.getCCount(degree, shapeNum));
										*/
										// Length-monomer constrained version has been shown to conform to sampling uniformly
										virtualSequence = MappingUtils.traverseTreeToString(
												NodeSequenceBinaryTreeGenerator.generateRandomMonomerConstrainedHardingTreeByDegree(
														degree, motifCounts.get("C").intValue()),
												NSConstants.MAPPING_MODE_TOPO_ACT, order);
										virtualMotifCounts = MotifFinder.findMotifs(virtualSequence, k, k);
										break;
									case 2:
										shapeNum = NodeSequenceBinaryTreeGenerator.randomBigInteger(possibleShapes);
										virtualMotifCounts = LengthDimerDistGenerator.getMotifCounts(degree, shapeNum, order);
										break;
									case 3:
										shapeNum = NodeSequenceBinaryTreeGenerator.randomBigInteger(possibleShapes);
										virtualMotifCounts = LengthTrimerDistGenerator.getMotifCounts(degree, shapeNum, order);
										break;
									default:
										// Generate random tree number and determine its motif counts
										virtualSequence = NodeSequenceBinaryTreeGenerator.generateRandomHardingTreeMap2Sequence(degree, order);
										virtualMotifCounts = MotifFinder.findMotifs(virtualSequence, k, maxConstraint, motifs);
									}
								} while(!isMatch(realTreeMotifCounts.get(k), virtualMotifCounts, motifsByLength.get(k)));
								if (virtualSequence == null){
									virtualSequence = NodeSequenceBinaryTreeGenerator.generateHardingTreeMap2Sequence(degree, shapeNum);
								}
								break;
							}
							debugPosition = 6;
							/*
							if (sequenceName.equals("ca1a-Axon")){
								System.out.println("ca1a-Axon "+virtualSequence);
								GACTCompressor.setDebug(true);
							}
							compressedSequence = GACTCompressor.compressSequenceIntoString(virtualSequence);
							virtualSequences.append(compressedSequence);
							if (sequenceName.equals("ca1a-Axon")){
								System.out.println(compressedSequence.getBytes()[0]+" "+compressedSequence.getBytes()[1]+" "+compressedSequence.getBytes()[2]+" "+compressedSequence.getBytes()[3]);
								System.out.println(compressedSequence.getBytes()[0]+" "+compressedSequence.getBytes()[1]+" "+compressedSequence.getBytes()[2]+" "+compressedSequence.getBytes()[3]);
								System.out.println(GACTCompressor.byteArrayToInt(compressedSequence.getBytes()));
							}
							GACTCompressor.setDebug(false);*/
							sequences.add(virtualSequence);
						}
						//virtualSequences.append("\n");
						if (uniformApproximation){
							uniqueSequences.clear();
							// Remove duplicates
							uniqueSequences.addAll(sequences);
							sequences.clear();
							// Put unique sequences back into the list
							sequences.addAll(uniqueSequences);
							for (int i = 0; i < numSurrogates; i++){
								virtualSequence = sequences.get((int)(Math.random()*sequences.size()));
								if (compress){
									virtualSequence = GACTCompressor.compressSequenceIntoString(virtualSequence);
									virtualSequencesSB.append(virtualSequence);
								}
								else{
									virtualSequencesSB.append(virtualSequence+"\n");
								}
							}
						}
						else{
							for (int i = 0; i < sequences.size(); i++){
								virtualSequence = sequences.get(i);
								if (compress){
									virtualSequence = GACTCompressor.compressSequenceIntoString(virtualSequence);
									virtualSequencesSB.append(virtualSequence);
								}
								else{
									virtualSequencesSB.append(virtualSequence+"\n");
								}
							}
						}
						if (compress){
							virtualSequencesSB.append("\n");
						}
						// Append virtual sequences for real sequence to the file
						appendToFile(virtualSequencesSB.toString());
						virtualSequencesSB = new StringBuffer();
					}
				}
				
			} catch (Exception e){
				e.printStackTrace();
			} finally {
				//System.out.println("Exiting ConstraintRunnable. Countdown now at "+(latch.getCount()-1));
				latch.countDown();
				debug("Leaving");
			}
		}
	}

	private static char getFirstChar(Map<String,Number> trimerCounts){
		String trimer;
		int aCount1 = 0, aCount2 = 0;
		for (Iterator<String> it = trimerCounts.keySet().iterator(); it.hasNext();){
			trimer = it.next();
			if (trimer.length() == 3){
				if (trimer.charAt(0) == 'A'){
					aCount1 += ((Integer)trimerCounts.get(trimer)).intValue();
				}
				if (trimer.charAt(1) == 'A'){
					aCount2 += ((Integer)trimerCounts.get(trimer)).intValue();
				}
			}
		}
		if (aCount1 > aCount2){
			return 'A';
		}
		else {
			return 'C';
		}
	}

	// Determines whether the real and virtual motif counts match
	private static boolean isMatch(Map<String,Number> realCounts, Map<String,Number> virtualCounts, List<String> motifs){
		int realKmerCount, virtualKmerCount;
		String motif;

		for (int i = 0; i < motifs.size(); i++){
			motif = (String)motifs.get(i);
			// Pull real kmer count 
			realKmerCount = realCounts.get(motif).intValue();
			virtualKmerCount = virtualCounts.get(motif).intValue();
			
			if (realKmerCount != virtualKmerCount){
				return false;
			}
		}
		
		return true;
	}

}
