package org.cn3.ns.motif.generation;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.motif.analysis.MotifFinder;
import org.cn3.ns.resources.DataProperties;
import org.cn3.ns.resources.NSConstants;
import org.cn3.ns.utils.NeuriteSequenceStringUtils;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 *
 * Produces a report on the progress of surrogate generation. This is particularly
 * useful when producing surrogates constrained by dimers or higher and sampling
 * strictly uniformly (rather than an approximation) from the set of tree shapes.
 * That is equivalent to generation method/distribution type 3.
 *
 */
public class DetermineSurrogateCompletion {
	
	private static final String PARAM_MAX_CONSTRAINT = "max-constraint";
	private static final String PARAM_OUTPUT_DIR = "output-dir";
	private static final String PARAM_START = "start";
	
	private static final int MAX_SEQUENCE_LENGTH = 2000;
	
	private static final int DEFAULT_MAX_CONSTRAINT = 3;

	private static double threshold;
	private static int kmerThresholdType = MotifConstants.KMER_THRESHOLD_NONE;
	private static int seqThresholdType = MotifConstants.SEQ_THRESHOLD_WHOLE;
	private static int method;
	private static String order;
	private static int mapping = NSConstants.MAPPING_MODE_TOPO_ACT;
	private static int maxConstraint = DEFAULT_MAX_CONSTRAINT;
	private static int startConstraint = 0;
	private static int maxCheck;
	private static boolean doubleLineBreak = false;
	
	public static void main(String args[]) throws Exception{
		JSAPResult config = readParameters(args);
		threshold = config.getDouble(MotifConstants.PARAM_THRESHOLD);
		seqThresholdType = config.getInt(MotifConstants.PARAM_SEQ_THRESHOLD_TYPE);
		if (seqThresholdType != MotifConstants.SEQ_THRESHOLD_WHOLE){
			kmerThresholdType = config.getInt(MotifConstants.PARAM_KMER_THRESHOLD_TYPE);
		}
		method = config.getInt(NSConstants.PARAM_RANDOM_METHOD);
		order = config.getString(NSConstants.PARAM_ORDER);
		int numSets = config.getInt(MotifConstants.PARAM_SET_NUM);
		startConstraint = config.getInt(PARAM_START);
		maxConstraint = config.getInt(PARAM_MAX_CONSTRAINT);
		maxCheck = maxConstraint+1;
		doubleLineBreak = config.getBoolean(MotifConstants.PARAM_DOUBLE_LINEBREAK);
		String outputDirectory = config.getString(PARAM_OUTPUT_DIR);
		
		System.out.println("Initializing...");
		
		StringBuffer sb;
		
		Integer sequenceId;
		
		int length, lengthIndex;

		// Get sequence names and map them to unique ids
		List sequenceNames = NSDataReader.loadSequenceList(mapping);
		Set remainingSequenceNames = new HashSet(NSDataReader.loadSequenceList(mapping));
		Map sequenceIdMap = new HashMap();
		for (int j = 0; j < sequenceNames.size(); j++){
			sequenceIdMap.put(sequenceNames.get(j), new Integer(j));
		}
		int databaseSize = sequenceNames.size();
		
		System.out.println("Loading tree size data");
		// Loading tree sizes represented in database
		List sequenceLengths = NSDataReader.loadUniqueSequenceLengths(mapping);

		// Loading map keyed by seq length, values are lists of sequence ids of seqs of that length
		Map lengthArborsMap = loadTreeSizeArborizationsMap(sequenceIdMap);
		Map seqLengthMap = NSDataReader.loadLengthDatabase(mapping);
		
		// Array for each sequence at each k (n-mer length) of number of virtual sequence sets for which a sequence was generated
		int[][] completed = new int[databaseSize][maxCheck];
		// Load results from prior runs
		loadCompletions(numSets, completed);

		// Initializing
		List[] motifsByLength = new List[maxCheck+1];
		for (int i = 1; i <= maxCheck; i++){
			if (i > 1){
				motifsByLength[i] = NeuriteSequenceStringUtils.generatePossibleACTKmers(i,order);
			}
			else{
				motifsByLength[1] = new ArrayList();
				motifsByLength[1].add("C");
			}
		}
		
		// Load real tree n-mer counts into sequence id keyed map
		System.out.println("Loading real tree motif counts");
		Map realTreeMotifCounts = loadRealTreeMotifCounts(sequenceIdMap, motifsByLength);
		
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(2);

		// Remove lengths that are too big
		List ids;
		System.out.println("Removing very large sequences (larger than "+MAX_SEQUENCE_LENGTH+")");
		for (lengthIndex = sequenceLengths.size()-1; ((Integer)sequenceLengths.get(lengthIndex)).intValue() > MAX_SEQUENCE_LENGTH; lengthIndex--){
			length = ((Integer)sequenceLengths.get(lengthIndex)).intValue();
			ids = (List)lengthArborsMap.remove(new Integer(length));
			for (int i = 0; i < ids.size(); i++){
				remainingSequenceNames.remove(sequenceNames.get(i));
			}
			sequenceLengths.remove(lengthIndex);
		}
		
		// Remove sequences that are no longer in the database
		sb = new StringBuffer();
		for (int i = 0; i < sequenceNames.size(); i++){
			String sequenceName = (String)sequenceNames.get(i);
			sequenceId = (Integer)sequenceIdMap.get(sequenceName);
			if (!realTreeMotifCounts.containsKey(sequenceId)){
				for (int j = 0; j < maxCheck; j++){
					remainingSequenceNames.remove(sequenceName);
				}
				sb.append(sequenceName+"  ");
			}
		}
		System.out.println("Sequences no longer in db: "+sb.toString());
		
		// DO NOT Remove sequences that are "outliers" (took too long to process)
		/*Map outliers = DataReader.loadOutliers(mapping, order, method, threshold, 
				thresholdTypeString, 60*100);*/

		// Add flagged neurons to removedSequences lists
		List flaggedSequences = NSDataReader.getFlaggedSequences();
		for (int i = 0; i < flaggedSequences.size(); i++){
			remainingSequenceNames.remove(flaggedSequences.get(i));
		}
		
		sb = new StringBuffer();
		sb.append("Sequence,Length");
		for (int k = startConstraint; k <= maxConstraint; k++){
			sb.append(","+k);
		}
		sb.append("\n");
		
		// Loop through constraint lengths
		String sequenceName;
		for (int i = 0; i < sequenceNames.size(); i++){
			sequenceName = (String)sequenceNames.get(i);
			if (remainingSequenceNames.contains(sequenceName)){
				sb.append(sequenceName+","+seqLengthMap.get(sequenceName));
				for (int k = startConstraint; k <= maxConstraint; k++){
					sb.append(","+completed[i][k]);
				}
				sb.append("\n");
			}
		}
		String filename = outputDirectory+
		"/constrainedTreesCompleted"+order+"-T"+threshold
			+ MotifConstants.getThresholdTypeString(kmerThresholdType, seqThresholdType)
			+ ".csv";

		DataWriter.writeStringToFile(filename, sb.toString());
	}
	
	// Maps sequence lengths to lists of sequence IDs associated with sequences of that length
	private static Map loadTreeSizeArborizationsMap(Map sequenceIdMap) throws Exception{
		Map treeSizeArborMapTmp = NSDataReader.loadTreeSizeArborizationsMap(mapping);
		Map treeSizeArborMap = new HashMap();
		Object key;
		List list1, list2;
		for (Iterator it = treeSizeArborMapTmp.keySet().iterator(); it.hasNext();){
			key = it.next();
			list1 = (List)treeSizeArborMapTmp.get(key);
			list2 = new ArrayList();
			for (int i = 0; i < list1.size(); i++){
				list2.add(sequenceIdMap.get(list1.get(i)));
				//System.out.println("ltsam "+list1.get(i)+" " +sequenceIdMap.get(list1.get(i)));
			}
			treeSizeArborMap.put(key, list2);
		}
		return treeSizeArborMap;
	}
	
	private static void loadCompletions(int setNum, int[][] completed) throws Exception{
		FileInputStream fis = null;
		DataInputStream dis = null;
		
		Set[] completions = new Set[maxCheck];
		for (int i = 0; i < maxCheck; i++){
			completions[i] = new TreeSet();
		}

		File file = new File(
				MotifConstants.getConstrainedTreesFilename(
						order, method, setNum, threshold, kmerThresholdType, seqThresholdType));
		if (file.exists()){
			try{
				fis = new FileInputStream(file);
				dis = new DataInputStream(fis);
	
				String dataString = "";
				Pattern p = Pattern.compile("(\\d+)\\|(\\d)\\-(\\d)");
				Matcher m;
				int id, kmin, kmax;

				dataString = NSDataReader.readLine(dis, doubleLineBreak);
				
				while (dataString != null) {
				    m = p.matcher(dataString);
			        if (m.matches()){
			        	id = Integer.parseInt(m.group(1));
			        	kmin = Integer.parseInt(m.group(2));
			        	kmax = Integer.parseInt(m.group(3));
			        	//System.out.println(id+" "+kmin+" "+kmax);
			        	for (int k = kmin; k < kmax && k < maxCheck; k++){
			        		// Constraint is accomplished up to kmax-1 on kmax
			        		completed[id][k]++;
			        	}

			        	// Clear the line
			        	if (doubleLineBreak){
			        		dis.read(); dis.read();
			        	}
			        }
			        dataString = NSDataReader.readLine(dis, doubleLineBreak);
			    }
				
			} catch (IOException ioe){
				throw new Exception(ioe);
			} finally{
				if (dis != null){
					dis.close();
				}
				if (fis != null){
					fis.close();
				}
			}
		}
	}
	
	private static Map loadRealTreeMotifCounts(Map sequenceIdMap, List[] motifsByLength) throws Exception{
		Map realTreeMotifCounts = new HashMap();
		FastaReader fr = new FastaReader();
		Map motifCounts;
		
		fr.setFile(DataProperties.getInstance().getFastaArborDir()+"Mapping"+mapping+"/Size/NMArborSize"+order+"Map"+mapping+".fasta");
		List sequenceList = NSDataReader.loadSequenceList(mapping);
		while (fr.next()){
			// Avoids case of using updated fasta file without updated lengths file
			if (sequenceList.contains(fr.getSequenceName())){
				motifCounts = MotifFinder.findMotifs(fr.getSequence(),1,maxCheck);
				for (int i = 1; i <= maxCheck; i++){
					for (int j = 1; j < motifsByLength[i].size(); j++){
						if (!motifCounts.containsKey(motifsByLength[i].get(j))){
							motifCounts.put(motifsByLength[i].get(j), new Integer(0));
						}
					}
				}
				realTreeMotifCounts.put(sequenceIdMap.get(fr.getSequenceName()),
					motifCounts);
			}
		}
		return realTreeMotifCounts;
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = MotifConstants.getParameters();
        
        FlaggedOption opt;
        
        opt = new FlaggedOption(PARAM_MAX_CONSTRAINT)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault(""+DEFAULT_MAX_CONSTRAINT)
        .setShortFlag('C') 
        .setLongFlag(PARAM_MAX_CONSTRAINT);
	    jsap.registerParameter(opt);
	    
	    opt = new FlaggedOption(PARAM_OUTPUT_DIR)
	    .setStringParser(JSAP.STRING_PARSER)
	    .setRequired(true)
	    .setShortFlag('O') 
	    .setLongFlag(PARAM_OUTPUT_DIR);
	    jsap.registerParameter(opt);

	    opt = new FlaggedOption(PARAM_START)
	    .setStringParser(JSAP.INTEGER_PARSER)
	    .setRequired(false)
	    .setDefault("0")
	    .setShortFlag(JSAP.NO_SHORTFLAG) 
	    .setLongFlag(PARAM_START);
	    jsap.registerParameter(opt);

    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + DetermineSurrogateCompletion.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

}
