package org.cn3.ns.scripts;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.data.FastaReader;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.motif.generation.ConstraintsDataReader;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 *
 * 	Takes list of neurite names from a file and then takes those sequences from
 *  given fasta file and puts into a separate fasta file.
 *  
 *  If value for -f parameter is a file with .dat extension, file is assumed to be
 *  a surrogate file and associated parameters are used (constraint, numSurrogates, surrogateNum). 
 *  
 *  
 * java org.cn3.ns.scripts.CreateGroupFastaFile -f "C:/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeLTSMap2_Min20.fasta" -g "C:/Documents and Settings/gillette/My Documents/research/NS-BLAST/MultipleAlignment/SimpleClassLists" -o "C:/Documents and Settings/gillette/My Documents/research/NS-BLAST/MultipleAlignment/ClassLTSFastaFiles"
 *
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.CreateGroupFastaFile -f "/home/tgillett/fastaFromSwc/NM/Arborization/Mapping2/Size/NMArborSizeSTLMap2.fasta" -g "/home/tgillett/Domains/" -o "/home/tgillett/Domains/ClusterLists/"
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.CreateGroupFastaFile -f "/home/tgillett/fastaFromSwc/NM/Random/Method8/Mapping2/surrogatesSTL-20sets_B8_Set1.dat" -g "/home/tgillett/Domains/SeqList-Axon-G3E1Identity-Dims7-Clust3-DistLim1.txt" -o "/home/tgillett/Domains/ClusterLists/" --surrogateNum 1 --constraint 3 --numSurrogates 100
 * 
 * java -classpath .:lib/JSAP-2.1.jar org.cn3.ns.scripts.CreateGroupFastaFile -f "/home/tgillett/fastaFromSwc/NM/Random/Method8/Mapping2/surrogatesSTL-100sets_B8_Set1.dat" -g "/home/tgillett/Domains/SeqList-Axon-G3E1Identity-Dims2of7-VII4-Clust3-DistLim1.txt" -o "/home/tgillett/Domains/ClusterLists/" --surrogateNum 1 --constraint 3 --numSurrogates 100
 */

public class CreateGroupFastaFile {

	private static Map<String,String> createSequenceMap(String fastaFilename, int surrogateNum, int constraint, int numSurrogates)
	throws Exception{
		Map<String,String> sequenceMap = new HashMap<String,String>();
		Random rand = new Random();

		if (fastaFilename.endsWith(".fasta")){
			FastaReader fr = new FastaReader(fastaFilename);
			while (fr.next()){
				sequenceMap.put(fr.getSequenceName(), fr.getSequence());
			}
		}else if (fastaFilename.endsWith(".dat")){
			Map<String,List<String>> surrogateMap = ConstraintsDataReader.loadSurrogates(fastaFilename, numSurrogates, false).get(constraint);
			for (String sequenceName : surrogateMap.keySet()){
				if (surrogateNum == -1){
					sequenceMap.put(sequenceName, surrogateMap.get(sequenceName).get(rand.nextInt(numSurrogates)));
				}else{
					sequenceMap.put(sequenceName, surrogateMap.get(sequenceName).get(surrogateNum-1));
				}
			}
		}
		return sequenceMap;
	}
	
	private static void createGroupFastaFile(String groupFilename, Map<String,String> sequenceMap, String outputDir, String surrogateString)
	throws Exception{
		File groupFile = new File(groupFilename);
		String groupName = groupFile.getName().replaceAll("\\.(txt|csv)", "");
		File outputFile = new File(outputDir+"/"+groupName+surrogateString+".fasta");
		
		Set<String> neuriteNames = NSDataReader.loadOrderedSet(groupFilename);
		//System.out.println("Start "+groupFilename+" : "+neuriteNames.size());
		StringBuffer groupFasta = new StringBuffer();
		
		// Write sequences to file in order of names in the group file
		for (String name : neuriteNames){
			if (name.startsWith("\"") && name.endsWith("\"")){
				name = name.substring(1, name.length()-1);
			}
			if (sequenceMap.containsKey(name)){
				groupFasta.append(">"+name+"\n"+sequenceMap.get(name)+"\n");
			}else{
				System.out.println("Sequence "+name+" not found in fasta file");
			}
		}

		System.out.println("End "+groupFilename+" : "+neuriteNames.size()+ " : "+groupFasta.length());

		DataWriter.writeStringToFile(outputFile, groupFasta);
		
	}
	
	private static final String PARAM_GROUP_FILE_DIR = "groupFileDir";
	private static final String PARAM_FASTA_FILENAME = "fastaFile";
	private static final String PARAM_OUTPUT_DIR = "outputDir";
	private static final String PARAM_SURROGATE_NUM = "surrogateNum";
	private static final String PARAM_SURROGATE_CONSTRAINT = "constraint";
	private static final String PARAM_NUM_SURROGATES = "numSurrogates";
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			File groupFileDir = new File(config.getString(PARAM_GROUP_FILE_DIR));
			String fastaFilename = config.getString(PARAM_FASTA_FILENAME);
			String outputDir = config.getString(PARAM_OUTPUT_DIR);
			// In case fasta file is actually a surrogate file
			int surrogateNum = config.getInt(PARAM_SURROGATE_NUM);
			int surrogateConstraint = config.getInt(PARAM_SURROGATE_CONSTRAINT);
			int numSurrogates = config.getInt(PARAM_NUM_SURROGATES);
			
			Map<String,String> sequenceMap = createSequenceMap(fastaFilename,surrogateNum,surrogateConstraint,numSurrogates);
			System.out.println(sequenceMap.size());
			
			String surrogateString = "";
			if (fastaFilename.endsWith(".dat")){
				if (surrogateNum == -1){
					surrogateString = "_SurrogateConstraint"+surrogateConstraint+"-RandNum";
				}else{
					surrogateString = "_SurrogateConstraint"+surrogateConstraint+"-"+surrogateNum+"of"+numSurrogates;
				}
			}
			
			if (groupFileDir.isDirectory()){
				File[] groupFiles = groupFileDir.listFiles();
				for (int i = 0; i < groupFiles.length; i++){
					//createGroupFastaFile(groupFiles[i].getAbsolutePath(),fastaFilename,outputDir);
					if (groupFiles[i].isFile()){
						createGroupFastaFile(groupFiles[i].getAbsolutePath(),sequenceMap,outputDir,surrogateString);
					}
				}
			}else{
				createGroupFastaFile(groupFileDir.getAbsolutePath(),sequenceMap,outputDir,surrogateString);
			}
			
			//System.out.println();
		} catch (Exception e){
			e.printStackTrace();
		}
	}

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = new JSAP();
        
        FlaggedOption opt;

        opt = new FlaggedOption(PARAM_GROUP_FILE_DIR)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag('g') 
        .setLongFlag(PARAM_GROUP_FILE_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_OUTPUT_DIR)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag('o') 
        .setLongFlag(PARAM_OUTPUT_DIR);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_FASTA_FILENAME)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault("")
        .setShortFlag('f') 
        .setLongFlag(PARAM_FASTA_FILENAME);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SURROGATE_NUM)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_SURROGATE_NUM);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_SURROGATE_CONSTRAINT)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_SURROGATE_CONSTRAINT);
        jsap.registerParameter(opt);

        opt = new FlaggedOption(PARAM_NUM_SURROGATES)
        .setStringParser(JSAP.INTEGER_PARSER)
        .setRequired(false)
        .setDefault("0")
        .setShortFlag(JSAP.NO_SHORTFLAG) 
        .setLongFlag(PARAM_NUM_SURROGATES);
        jsap.registerParameter(opt);

        JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (@SuppressWarnings("rawtypes")
			java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + CreateGroupFastaFile.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;
	}

}
