package org.cn3.ns.scripts;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.cn3.morpho.data.DataWriter;
import org.cn3.ns.data.NSDataReader;
import org.cn3.ns.motif.generation.MotifConstants;
import org.cn3.ns.resources.NSConstants;

import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;

/**
 * 
 * @author gillette
 * Determines the number of sequences and proportion of total characters 
 * constrained at each k for each set.
 * Also outputs the status in terms of last k/length completed
 * java org.krasnow.cng.working.NMerConstraintStatus -T 0.05 -S 0 -K 0 -o LTS -s 20
 */
public class NMerConstraintStatus {
	
	private static final int maxConstraint = 4;
	private static final int MAX_SEQUENCE_LENGTH = 900;
	
	public static void main(String args[]){
		try{
			JSAPResult config = readParameters(args);
			int sets = config.getInt(MotifConstants.PARAM_SET_NUM);
			int mapping = NSConstants.MAPPING_MODE_TOPO_ACT;
			double threshold = config.getDouble(MotifConstants.PARAM_THRESHOLD);
			int kmerThresholdType = 0, seqThresholdType = config.getInt(MotifConstants.PARAM_SEQ_THRESHOLD_TYPE);
			if (seqThresholdType != MotifConstants.SEQ_THRESHOLD_WHOLE){
				kmerThresholdType = config.getInt(MotifConstants.PARAM_KMER_THRESHOLD_TYPE);
			}

			int method = config.getInt(NSConstants.PARAM_RANDOM_METHOD);
			String order = config.getString(NSConstants.PARAM_ORDER);
			boolean doubleLineBreak = config.getBoolean(MotifConstants.PARAM_DOUBLE_LINEBREAK);

			File file;
			int i, k;
			
			// Get sequence names and map them to unique ids
			List sequenceNames = NSDataReader.loadSequenceList(mapping);
			Map lengthMap = NSDataReader.loadLengthDatabase(mapping);
			Map sequenceIdMap = new HashMap();
			for (int j = 0; j < sequenceNames.size(); j++){
				sequenceIdMap.put(sequenceNames.get(j), new Integer(j));
			}

			int[] lastSize = new int[sets];
			int[] lastK = new int[sets];
			long[][] numCompleted = new long[sets][maxConstraint];
			long[][] amountCompleted = new long[sets][maxConstraint];
			
			for (int setNum = 1; setNum <= sets; setNum++){
				// Loading map keyed by seq length, values are lists of sequence ids of seqs of that length
				
				FileInputStream fis = null;
				DataInputStream dis = null;
				
				file = new File(
						MotifConstants.getConstrainedTreesFilename(
								order, method, setNum, threshold, kmerThresholdType, seqThresholdType));
				System.out.println(file.getAbsolutePath());
				if (file.exists()){
					try{
						fis = new FileInputStream(file);
						dis = new DataInputStream(fis);
			
						String dataString = "";
						Pattern p = Pattern.compile("(\\d+)\\|(\\d)\\-(\\d)");
						Matcher m;
						int id = -1, kmin = 0, kmax;

						dataString = NSDataReader.readLine(dis, doubleLineBreak);
						
						while (dataString != null) {
						    m = p.matcher(dataString);
					        if (m.matches()){
					        	id = Integer.parseInt(m.group(1));
					        	kmin = Integer.parseInt(m.group(2));
					        	kmax = Integer.parseInt(m.group(3));
					        	// Increment number and amount completed for setNum/k
					        	for (k = kmin; k <= kmax && k <= maxConstraint; k++){
					        		if (k > 0){
						        		i = k - 1;
						        		numCompleted[setNum-1][i]++;
						        		amountCompleted[setNum-1][i] +=
						        			((Integer)lengthMap.get(sequenceNames.get(id))).intValue();
					        		}
					        	}

					        	// Clear the line
					        	if (doubleLineBreak){
					        		dis.read(); dis.read();
					        	}
					        	
					        }
					        dataString = NSDataReader.readLine(dis, doubleLineBreak);
					    }
			        	lastK[setNum-1] = kmin;
			        	if (id != -1)
			        		lastSize[setNum-1] = ((Integer)lengthMap.get(sequenceNames.get(id))).intValue();
						
					} catch (IOException ioe){
						throw new Exception(ioe);
					} finally{
						if (dis != null){
							dis.close();
						}
						if (fis != null){
							fis.close();
						}
					}
				}
			}

			// Get total database length
			long totalLength = 0, numSequences = 0;
			for (Iterator it = lengthMap.keySet().iterator(); it.hasNext();){
				int length = ((Integer)lengthMap.get(it.next())).intValue();
				if (length < MAX_SEQUENCE_LENGTH){
					totalLength += length;
					numSequences++;
				}
			}
			
			// Write result files
			StringBuffer numberSb = new StringBuffer(), amountSb = new StringBuffer(), lastSb = new StringBuffer();
			StringBuffer header = new StringBuffer();
			lastSb.append("Set,k,Length\n");
			for (k = 1; k <= maxConstraint; k++){
				header.append(","+k);
			}
			numberSb.append(header+"\n");
			amountSb.append(header+"\n");
			
			for (int setNum = 1; setNum <= sets; setNum++){
				lastSb.append(setNum+","+lastK[setNum-1]+","+lastSize[setNum-1]+"\n");
				numberSb.append(setNum);
				amountSb.append(setNum);
				for (k = 1; k <= maxConstraint; k++){
					numberSb.append(","+((double)numCompleted[setNum-1][k-1]/numSequences));
					amountSb.append(","+((double)amountCompleted[setNum-1][k-1]/totalLength));
				}
				numberSb.append("\n");
				amountSb.append("\n");
			}
			
			String dir = "C:/Documents and Settings/gillette/My Documents/research/MotifAnalysis/";
			DataWriter.writeStringToFile(dir+"NumberCompletedBySet"+order+".csv", numberSb.toString());
			DataWriter.writeStringToFile(dir+"AmountCompletedBySet"+order+".csv", amountSb.toString());
			DataWriter.writeStringToFile(dir+"LastCompletedBySet"+order+".csv", lastSb.toString());
		} catch (Exception e){
			e.printStackTrace();
		}
		
	}
	

	private static JSAPResult readParameters(String[] args) throws Exception{
		JSAP jsap = MotifConstants.getParameters();
        
    	JSAPResult config = jsap.parse(args);
        if (!config.success()) {
        	for (java.util.Iterator errs = config.getErrorMessageIterator();errs.hasNext();) {
        		System.err.println("Error: " + errs.next());
        	}
            System.err.println();
            System.err.println("Usage: java "
                                + NMerConstraintStatus.class.getName());
            System.err.println("                "
                                + jsap.getUsage());
            System.err.println();
            // show full help as well
            System.err.println(jsap.getHelp());
            System.exit(1);
        }
        
        return config;

	}
	
}
